(() => {
    'use strict';

    angular
        .module('easyorder.modals.facebiometrics',[])
        .factory('FaceBiometricsModal', FaceBiometricsModal);

        FaceBiometricsModal.$inject = ['$ionicModal', '$rootScope', '$q', '$ionicPopup', 'AlertService', 'LogService', '$ionicLoading', 'BiometricsService', 'UserService', 'ConfigService', 'CartService'];

    function FaceBiometricsModal($ionicModal, $rootScope, $q, $ionicPopup, AlertService, LogService, $ionicLoading, BiometricsService, UserService, ConfigService, CartService) {
        const HEADER = '[FaceBiometricsModal]'; 
        const $scope = $rootScope.$new(),
            modalOptions = {
                scope: $scope,
                focusFirstInput: true
        };

        $scope.loadModel = function (name) {
            const filePath = "file:///android_asset/www/unico/models/" + name;
            return $q((resolve) => {
                window.resolveLocalFileSystemURL(
                    filePath,
                    function (fileEntry) {
                        fileEntry.file(
                            function (file) {
                                const reader = new FileReader();

                                reader.onloadend = function () {
                                    resolve(this.result);
                                };

                                reader.readAsArrayBuffer(file);
                            },
                            function () {
                                resolve(null);
                            }
                        );
                    },
                    function () {
                        resolve(null);
                    }
                );
            })
        }

        function confirmCustomerPhoto(customerFaceImage) {
            $scope.customerFaceImage = customerFaceImage;
            LogService.info(`${HEADER} Confirmação da foto pelo operador solicitada após a não autorização.`);

            // ambos os casos deve logar e cancelar
            $ionicPopup.confirm({
                templateUrl: 'js/components/biometrics/customer-photo-confirmation.html',
                title: 'O cliente é o mesmo da foto?',
                scope: $scope,
                buttons: [
                    { text: 'Não',
                        onTap: function(event) {
                            LogService.info(`${HEADER} Foto não confirmada pelo operador. Fluxo será abortado pela não autorização!`);
                            const message = createPhotoConfirmationMessageLog(false);
                            LogService.info(message);
                            $scope.cancel();
                        }
                    },
                    {
                        text: '<b>Sim</b>',
                        type: 'button-theme',
                        onTap: function(event) {
                            LogService.info(`${HEADER} Foto confirmada pelo operador mas fluxo será abortado pela não autorização!`);
                            const message = createPhotoConfirmationMessageLog(true);
                            LogService.info(message);
                            $scope.cancel();
                        }
                    }
                ]
            });
        }

        function createPhotoConfirmationMessageLog(photoConfirmed) {
            // foi solicitada uma mensagem de log específica, contendo a mensagem abaixo.
            const storeNumber = ConfigService.config.storeNumber;
            const terminalNumber = ConfigService.config.terminalNumber;
            const operatorCode = UserService.operatorCode();
            const nsu = CartService.cart.nsu;

            return `Store: [${storeNumber}], WorkstationId: [${terminalNumber}], OperatorPartyId: [${operatorCode}], TransactionSequence: [${nsu}] - [biometrics validation] - User reported that the photo displayed is ${photoConfirmed ? 'match' : 'not match'} with customer`;
        }

        function onFaceNotRecognized(errorMessage) {
            const message = `${errorMessage} Deseja efetuar uma nova captura?`;
            LogService.info(`${HEADER} ${message}`);

            return AlertService.confirm(message, 0, null, 'Sim', 'Não')
                .then((okPressed) => {
                    if (okPressed) {
                        $scope.initCamera();
                    } else {
                        LogService.info(`${HEADER} Nova captura recusada.`);
                        const customerFaceImage = BiometricsService.getCustomerFaceImage($scope.customer);
                        if (customerFaceImage) {
                            confirmCustomerPhoto(customerFaceImage);
                        } else {
                            // se não tiver foto do cliente, cancela
                            $scope.cancel();
                        }
                    }
                });
        }

        function onCapture(result) {
            $ionicLoading.hide()
                .then(() => extractBase64Photo(result))
                .then((base64Image) => confirmPhoto(base64Image))
                .catch((err) => LogService.error(err));
        }

        function confirmPhoto(base64Image) {
            $scope.customerFaceImage = base64Image;
            LogService.info(`${HEADER} Confirmação de foto para envio da biometria.`);

            $ionicPopup.confirm({
                templateUrl: 'js/components/biometrics/customer-photo-confirmation.html',
                title: 'Confirma envio da foto para validação da biometria?',
                scope: $scope,
                buttons: [
                    { text: 'Não',
                        onTap: function(event) {
                            LogService.info(`${HEADER} Foto não confirmada para envio, será capturada novamente.`);
                            $scope.initCamera();
                        }
                    },
                    {
                        text: '<b>Sim</b>',
                        type: 'button-theme',
                        onTap: function(event) {
                            LogService.info(`${HEADER} Foto será enviada para validação biométrica.`);
                            onSendCapturedPhoto(base64Image);
                        }
                    },
                ]
            });
        }

        function onCaptureError(error) {
            $ionicLoading.hide()
                .then(() => {
                    const message = "Erro na captura da foto, deseja tentar novamente?";
                    AlertService.confirm(message, 0, null, 'Sim', 'Não')
                        .then((result) => {
                            if (result) {
                                $scope.initCamera();
                            } else {
                                $scope.cancel();
                            }
                        });
                })
            
        }

        function createFaceRecognitionRequest(base64Image) {
            return {
                storeNumber: ConfigService.config.storeNumber,
		        terminalNumber: ConfigService.config.terminalNumber,
		        operatorCode: UserService.operatorCode(),
		        nsu: CartService.cart.nsu,
                image: base64Image,
                id: $scope.customer.biometrics.id,
                operationFlow: $scope.operationFlow,
                document: $scope.customer.document
            };
        }

        function onSendCapturedPhoto(base64Image) {
            LogService.info(`${HEADER} Enviando foto para o serviço de biometria.`);
            const params = createFaceRecognitionRequest(base64Image);
            $ionicLoading.show({
                template: 'Imagem em validação, por favor aguarde...',
                duration: 0
            })
            .then(() => BiometricsService.faceRecognition(params))
            .then((response) => {
                const recognized = response && response.data && response.data.recognized === true;
                if (recognized) {
                    closeCamera()
                        .then(() => $ionicLoading.hide())
                        .then(() => $scope.onFaceRecognized())
                } else {
                    $ionicLoading.hide()
                        .then(() => onFaceNotRecognized('Biometria não identificada.'));
                }
            })
            .catch((error) => {
                $ionicLoading.hide().then(() => {
                    onFaceNotRecognized('Erro ao autorizar biometria.');
                });
            }); 
        }

        $scope.initCamera = () => {
            var configurations = {
                optional: {
                    // FACE_MODE_TYPE.BACK
                    FACE_MODE: 2, 
                    LABEL_DOCUMENT_TYPE_OTHERS: "doc",
                },
                cameraType: {
                    // TYPE_CAMERA.CAMERA_NORMAL
                    code: 1 
                }
            }

            var layout = {
                silhouette: {
                    primaryColor: "#0bbd26",
                    secondaryColor: "#bd0b0b",
                    neutralColor: "#fff",
                },
                buttonCapture: {
                    backgroundColor: "#2980ff",
                    iconColor: "#fff",
                },
                popupLoadingHtml: `<div style="position: absolute; top: 45%; right: 50%; transform: translate(50%, -50%); z-index: 10; text-align: center;">Loading...</div>`,
                boxMessage: {
                    backgroundColor: "#2980ff",
                    fontColor: "#fff"
                },
                boxDocument: {
                    backgroundColor: "#2980ff",
                    fontColor: "#fff"
                }
            }

            var callback = {
                on: {
                    success: function (result) {
                        onCapture(result);
                    },
                    error: function (error) {
                        onCaptureError(error);
                    },
                    support: function (e) {
                        console.log(e)
                    },
                    cameraLoaded: function (e) {
                        $ionicLoading.hide();
                    }
                }
            }
            LogService.info(`${HEADER} Iniciando nova captura da biometria.`);
            return $ionicLoading.show({
                template: 'Iniciando camera...',
                duration: 0
            })
            .then(() => $scope.isCameraOpened = true)
            .then(() => $scope.camera.initCamera(configurations, callback, layout));
        }

        $scope.prepareCamera = async () => {
            $scope.camera = new unico.UnicoCamera();

            let saveModelPromise = null;
            if ($rootScope.isCordova) {
                if (ionic.Platform.isAndroid()) {
                    let modelBuffers = {};
                    let promises = [];
                    let models = ['tiny_face_detector_model-shard1.json', 'face_landmark_68_model-shard1.json','tiny_face_detector_model-weights_manifest.json' ,'face_landmark_68_model-weights_manifest.json' ]
                    models.forEach(modelName => {
                        promises.push($scope.loadModel(modelName).then((buffer) => { modelBuffers[modelName] = buffer } ))
                    });

                    saveModelPromise = await $q.all(promises)
                    .then(() => $scope.camera.savePreLoadedModels(modelBuffers))
                } else {
                    saveModelPromise = AlertService.error('Sistema operacional não comporta carregamento dos arquivos da camera')
                    .then($q.reject);
                }
            } else {
                saveModelPromise = $q.resolve();
            }
            
            return $ionicLoading.show({
                template: 'Configurando camera...',
                duration: 0
            })
            .then(() => saveModelPromise)
            .then(() => $scope.camera.loadModels("unico/models"))
            .then(() => $scope.initCamera())
            .catch(() => {
                $ionicLoading.hide();
            });

        }

        function extractBase64Photo (captureData) {
            let base64Image = '';
            if (captureData && captureData.image) {
                base64Image = captureData.image.replace("data:image/jpeg;base64,", "");
            }
            return base64Image;
        }


        $scope.$on('modal.shown', function() {
            $scope.isCameraOpened = false;
            $scope.prepareCamera();
        });

        function closeCamera() {
            if ($scope.isCameraOpened) {
                try {
                    $scope.camera.closeCamera();
                } catch (error) {
                    LogService.error(`${HEADER} Erro ao fechar a câmera: ${error}`);
                } finally {
                    $scope.isCameraOpened = false;
                    return $q.resolve();
                }
            } else {
                return $q.resolve();
            }
        }

        function open(customer, operationFlow) {
            $scope.customer = customer;
            $scope.operationFlow = operationFlow;
            return new Promise((resolve, reject) => {
                LogService.info(`${HEADER} Solicitando biometria do cliente de documento '${customer.document}'!`);
                $ionicModal.fromTemplateUrl(
                    'js/components/biometrics/face-biometrics-modal.html',
                    modalOptions
                ).then(function (modalInstance) {
                        $scope.onFaceRecognized = function (manual) {
                            LogService.info(`${HEADER} Biometria do cliente de documento '${customer.document}' reconhecida ${manual ? 'de forma manual' : 'automaticamente pelo serviço' }.`);
                            closeAndRemove(modalInstance)
                                .finally(() => resolve());
                        };

                        $scope.cancel = function () {
                            LogService.info(`${HEADER} Biometria do cliente de documento '${customer.document}' cancelada.`);
                            closeAndRemove(modalInstance)
                                .finally(() => reject(`Biometria cancelada!`));
                        };
                        return modalInstance.show();
                });
            });
        }

        function closeAndRemove(modalInstance) {
            return closeCamera()
                .then(() => modalInstance.hide())
                .then(() => modalInstance.remove());
        }

        const instance = {
            open: open
        };

        return instance;
    }
})();