angular.module('fileLogger', ['ngCordova.plugins.file'])
.factory('$fileLogger', ['$q', '$window', '$cordovaFile', '$filter',
    function ($q, $window, $cordovaFile, $filter) {
        'use strict';

        function parseStorageFilename() {
            return `${storageBaseFilename}.${moment().format('YYYYMMDD')}.log`;
        }

        var messageFileWriterPromiseChain = $q.resolve();
        
        var levels = ['DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL', 'RESPONSE ERROR', 'REQUEST ERROR', 'RELOAD'];

        let storageBaseFilename = 'caixarapido';
        var storageFilename = parseStorageFilename();
        var storageLocation = null;

        let dateFormat = 'dd/MM/yyyy HH:mm:ss:sss';
        let dateTimezone = "";

        // last compressed date
        // inicializando o valor apesar de não ter comprimido nada
        let lastCompressed = moment();

        const MAX_ZIP_FILE_SIZE = 10485760;

        let logCtx;

        function isRipple() {
            return $window.parent && $window.parent.ripple;
        }

        function isBrowser() {
            return (!$window.cordova && !$window.PhoneGap && !$window.phonegap) || isRipple();
        }

        function log(level) {
            if (angular.isString(level)) {
                level = level.toUpperCase();

                if (levels.indexOf(level) === -1) {
                    level = 'INFO';
                }
            } else {
                level = 'INFO';
            }

            const now = moment();

            const timestamp = dateFormat ? $filter('date')(now.toDate(), dateFormat, dateTimezone) : now.toJSON();

            var messages = parseLogMessage(Array.prototype.slice.call(arguments, 1));
            messages = messages.join(' ');
            
            if(messages !== undefined) {
                messages = hiddenPrivateInformations(messages);
            }

            const ctxMessage = parseCtxMessage(logCtx);

            writeConsoleMsg(level, [`${timestamp} ${ctxMessage}${messages}`]);
            messageFileWriterPromiseChain = messageFileWriterPromiseChain
            .then(() => {
                if (now.dayOfYear() != lastCompressed.dayOfYear() 
                        || now.year() != lastCompressed.year()) {
                    return checkFileSizeAndZip(storageFilename, now)
                    .then(() => nextStorageFilename());
                }
                return $q.resolve();
            })
            .then(() => writeLog(`${timestamp} ${level} ${ctxMessage}${messages} \n`))
            .catch(error => console.log(error));
        }

        function parseCtxMessage(ctx) {
            if (!ctx) {
                return '';
            }

            let ret = '';

            const append = (key, value) => {
                value = value || ctx[key];
                if (value) {
                    ret = ret.length == 0 ?
                        `${key}=${value}` :
                        `${ret}, ${key}=${value}`;
                }
            };

            append('transactionsIds');
            append('externalTransactionsIds');

            return ret.length == 0 ? '' : `[${ret}] `;
        }

        function compressFile(logfile, now) {
            return getLogfile(logfile)
            .then((content) => {
                const zip = new JSZip();

                zip.file(logfile, content);

                return zip.generateAsync({ type: "blob", compression: "DEFLATE" })
                .then((zipContent) => {
                    return zipFilename(logfile)
                    .then((zipFile) => {
                        return write(zipFile, zipContent, true)
                        .then(() => log("INFO", `Arquivo ${logfile} compactado ${zipFile}`))
                        .catch(error => {
                            log("ERROR", `Não foi possível escrever em arquivo compactado: ${zipFile}`, error);
                            return $q.reject(error);
                        });
                    })
                })
            })
            .then(() => lastCompressed = now)
            .then(() => deleteFile(logfile));
        }

        function write(file, content, creating) {
            if (isBrowser()) {
                $window.localStorage[file] = content;
                return $q.resolve();
            } else {
                if (!$window.cordova || !$window.cordova.file || !storageLocation) {
                    return $q.reject('cordova.file.dataDirectory is not available');
                }
                return $q((resolve, reject) =>
                    $cordovaFile.writeFile(storageLocation, file, content, creating)
                    .then(resolve)
                    .catch(reject)
                );
            }
        }

        function zipFilename(logfile) {
            return lsFiles('.zip')
            .then((files) => {
                let name = `${logfile}.zip`;

                let i = 0;
                while(files.includes(name)) {
                    name = `${logfile}.${++i}.zip`;
                }

                return $q.resolve(name);
            });
        }

        function compressFiles() {
            return lsFiles('.log')
                .then((files) => {
                    const promises = files
                        .filter(name => name != storageFilename)
                        .map(file => checkFileSizeAndZip(file, moment()));
                    return $q.all(promises);
                });
        }

        function storageAvailable() {
            return $window.cordova && $window.cordova.file && storageLocation;
        }

        function checkFileSizeAndZip(logfile, now) {
            return $q((resolve,  reject) => {
                if (isBrowser()) {
                    resolve();
                } else {
                    if (!storageAvailable()) {
                        resolve();
                    } else {
                        console.log('Verificando tamanho do arquivo ' + storageLocation + logfile);
                        window.resolveLocalFileSystemURL(storageLocation + logfile, function(fileEntry) {
                            fileEntry.file(function(fileObj) {
                                // zipa apenas se o tamanho for maior que o definido
                                if (fileObj.size > MAX_ZIP_FILE_SIZE) {
                                    console.log('Tamanho do arquivo '+  fileObj.size + ' excede ' + MAX_ZIP_FILE_SIZE + '. Arquivo não será compactado.');
                                    resolve();
                                } else {
                                    console.log('Arquivo de tamanho '+  fileObj.size + ' será compactado.');
                                    compressFile(logfile, now).then(() =>  {
                                        console.log('Finalizada compressao do arquivo ' + logfile);
                                        resolve();
                                    });
                                    
                                }
                            },
                            function(err){
                                console.log('Erro ao obter dados do do arquivo:' + err);
                                reject();
                            });
                        }, reject);
                    }
                    
                }
            });
        }
     
        function lsFiles(ext) {
            return $q((resolve,  reject) => {
                if (isBrowser()) {
                    resolve(Object.keys($window.localStorage));
                } else {
                    if (!$window.cordova || !$window.cordova.file || !storageLocation) {
                        return $q.reject('cordova.file.dataDirectory is not available');
                    }
                    window.resolveLocalFileSystemURL(storageLocation, (fileSystem) => {
                        const reader = fileSystem.createReader();
                        reader.readEntries((entries) => {
                            resolve(entries
                                .map(entry => entry.name)
                            );
                        }, reject);
                    }, reject);
                }
            }).then((files) => {
                return angular.isString(ext) && ext.length > 0 ?
                    files.filter(files => files.endsWith(ext)) :
                    files;
            });
        }

        function parseLogMessage(messages) {
            var ret = [];
            var text;

            for (var i = 0; i < messages.length; i++) {
                if (angular.isArray(messages[i])) {
                    text = '[Array]';

                    try {
                        text = JSON.stringify(messages[i]);
                    } catch (e) {
                    }
                    ret.push(text);
                } else if (angular.isObject(messages[i])) {
                    text = '[Object]';

                    try {
                        text = JSON.stringify(messages[i]);
                    } catch (e) {
                    }
                    ret.push(text);
                } else {
                    ret.push(messages[i]);
                }
            }
            return ret;
        }
        
        function writeConsoleMsg(level, messages) {
            if (angular.isObject(console) && angular.isFunction(console.log)) {
                switch (level) {
                    case 'DEBUG':
                        if (angular.isFunction(console.debug)) {
                            console.debug.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'INFO':
                        if (angular.isFunction(console.info)) {
                            console.info.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'WARN':
                        if (angular.isFunction(console.warn)) {
                            console.warn.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'ERROR':
                        if (angular.isFunction(console.error)) {
                            console.error.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'FATAL':
                        if (angular.isFunction(console.error)) {
                            console.error.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'RESPONSE ERROR':
                        if (angular.isFunction(console.error)) {
                            console.error.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'REQUEST ERROR':
                        if (angular.isFunction(console.error)) {
                            console.error.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    case 'RELOAD':
                        if (angular.isFunction(console.warn)) {
                            console.warn.apply(console, messages);
                        } else {
                            console.log.apply(console, messages);
                        }
                        break;
                    default:
                        console.log.apply(console, messages);
                }
            }
        }
        
        function writeLog(message) {
            var q = $q.defer();

            if (isBrowser()) {
                if (!$window.localStorage[storageFilename]) {
                    $window.localStorage[storageFilename] = '';
                }
                $window.localStorage[storageFilename] += message;
                q.resolve();
            } else {
                if (!$window.cordova || !$window.cordova.file || !storageLocation) {
                    q.reject('cordova.file.dataDirectory is not available');
                    return q.promise;
                }

                $cordovaFile.checkFile(storageLocation, storageFilename).then(
                    function () {
                        $cordovaFile.writeExistingFile(storageLocation, storageFilename, message).then(
                            function () {
                                q.resolve();
                            },
                            function (error) {
                                q.reject(error);
                            }
                        );
                    },
                    function () {
                        $cordovaFile.writeFile(storageLocation, storageFilename, message, true).then(
                            function () {
                                q.resolve();
                            },
                            function (error) {
                                q.reject(error);
                            }
                        );
                    }
                );
            }
            return q.promise;
        }


        function getLogfile(logfile) {
            var q = $q.defer();

            if (isBrowser()) {
                q.resolve($window.localStorage[logfile]);
            } else {
                if (!$window.cordova || !$window.cordova.file || !storageLocation) {
                    q.reject('cordova.file.dataDirectory is not available');
                    return q.promise;
                }
                $cordovaFile.readAsText(storageLocation, logfile).then(
                    function (result) {
                        q.resolve(result);
                    },
                    function (error) {
                        q.reject(error);
                    }
                );
            }
            return q.promise;
        }

        function deleteFile(logfile) {
            var q = $q.defer();

            if (isBrowser()) {
                $window.localStorage.removeItem(logfile);
                q.resolve();
            } else {
                if (!$window.cordova || !$window.cordova.file || !storageLocation) {
                    q.reject('cordova.file.dataDirectory is not available');
                    return q.promise;
                }
                $cordovaFile.removeFile(storageLocation, logfile).then(
                    function (result) {
                        q.resolve(result);
                    },
                    function (error) {
                        q.reject(error);
                    }
                );
            }
            return q.promise;
        }

        function nextStorageFilename() {
            setStorageFilename(parseStorageFilename());
        }

        function setStorageFilename(filename) {
            if (angular.isString(filename) && filename.length > 0) {
                storageFilename = filename;
                return true;
            } else {
                return false;
            }
        }

        function setStorageMethod(method) {
            if (method == 'EXTERNAL_DIRECTORY') {
                storageLocation = $window.cordova.file.externalDataDirectory;
            } else if (method == 'EXTERNAL_ROOT_DIRECTORY') {
                storageLocation = $window.cordova.file.externalRootDirectory;
            } else if (method == 'DATA_DIRECTORY') {
                storageLocation = $window.cordova.file.dataDirectory;
            } else {
                storageLocation = method;
            }
        }

        function setTimestampFormat(format, timezone) {
            if (!(angular.isUndefined(format) || angular.isString(format))) {
                throw new TypeError('format parameter must be a string or undefined');
            }
            if (!(angular.isUndefined(timezone) || angular.isString(timezone))) {
                throw new TypeError('timezone parameter must be a string or undefined');
            }
            dateFormat = format;
            dateTimezone = timezone;
        }

        function setStorageBaseFilename(filename) {
            if (!angular.isString(filename) && filename.length > 0) {
                throw new Error('Filename is not a string');
            }
            storageBaseFilename = filename;
            nextStorageFilename();
        }

        function checkFile() {
            var q = $q.defer();

            if (isBrowser()) {
                q.resolve({
                    'name': storageFilename,
                    'localURL': 'localStorage://localhost/' + storageFilename,
                    'type': 'text/plain',
                    'size': ($window.localStorage[storageFilename] ? $window.localStorage[storageFilename].length : 0)
                });
            } else {
                if (!$window.cordova || !$window.cordova.file || !storageLocation) {
                    q.reject('cordova.file.dataDirectory is not available');
                    return q.promise;
                }

                $cordovaFile.checkFile(storageLocation, storageFilename).then(function (fileEntry) {
                    fileEntry.file(q.resolve, q.reject);
                }, q.reject);
            }
            return q.promise;
        }

        function debug() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('DEBUG');
            log.apply(undefined, args);
        }

        function fatal() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('FATAL');
            log.apply(undefined, args);
        }

        function responseError() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('RESPONSE ERROR');
            log.apply(undefined, args);
        }

        function requestError() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('REQUEST ERROR');
            log.apply(undefined, args);
        }

        function info() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('INFO');
            log.apply(undefined, args);
        }

        function warn() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('WARN');
            log.apply(undefined, args);
        }

        function error() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('ERROR');
            log.apply(undefined, args);
        }

        function reload() {
            var args = Array.prototype.slice.call(arguments, 0);
            args.unshift('RELOAD');
            log.apply(undefined, args);
        }

        function reconfigure(ctx) {
            logCtx = ctx;
        }

		function hiddenPrivateInformations(messages) {
            if(messages instanceof Array){
                for(var message of messages) {
                    message = applyHidden(message, 'password', '*****');
                    message = applyHidden(message, 'senha', '*****');
                }
            } else {
                messages = applyHidden(messages, 'password', '*****');
                messages = applyHidden(messages, 'senha', '*****');
            }
            return messages;
		}

		function applyHidden(message, keyword, replaceText) {
            var regex = "(?:\\b" + keyword + ")\\W+\\w+";
			message = message.replace(new RegExp(regex), keyword + "\": \"" +replaceText);

			return message;
		}	

        return {
            log: log,
            getLogfile: getLogfile,
            setStorageMethod: setStorageMethod,
            setTimestampFormat: setTimestampFormat,
            checkFile: checkFile,
            debug: debug,
            info: info,
            warn: warn,
            error: error,
            fatal: fatal,
            requestError: requestError,
            responseError: responseError,
            reload: reload,
            reconfigure: reconfigure,
            setStorageBaseFilename: setStorageBaseFilename,
            compressFiles: compressFiles
        };
    }
]);
