(() => {
    'use strict';

    angular
        .module('easyorder.modals.fiscalcustomer',[])
        .factory('FiscalCustomerModal', FiscalCustomerModal);

        FiscalCustomerModal.$inject = ['$q', '$ionicModal', '$rootScope', '$ionicLoading', 'AlertService', 'LogService', 'CustomerService', 'UserService'];

    function FiscalCustomerModal($q, $ionicModal, $rootScope, $ionicLoading, AlertService, LogService, CustomerService, UserService) {
        const HEADER = '[FiscalCustomerModal]'; 
        const $scope = $rootScope.$new(),
            modalOptions = {
                scope: $scope,
                focusFirstInput: true
        };

        function init(customer, mandatoryFiscalFields, isOverLimit, storeState, disableFilledFields) {
            $scope.mandatoryDocument = true;
            $scope.mandatoryName = false;
            $scope.mandatoryAddress = false;
            $scope.disableFilledFields = disableFilledFields;

            // só exibe os campos adicionais se passar do limite obrigatório
            if (isOverLimit) {
                $scope.mandatoryName = mandatoryFiscalFields && mandatoryFiscalFields.includes('NAME');
                $scope.mandatoryAddress = mandatoryFiscalFields && mandatoryFiscalFields.includes('ADDRESS');
            }

            $scope.fiscalCustomer = {};

            if (customer) {
                $scope.customerCopy = JSON.parse(JSON.stringify(customer));
                // não clona todos os campos pois pode mudar o destinatário e seria enviado por exemplo o endereço errado
                $scope.fiscalCustomer.document = customer.document;
                $scope.fiscalCustomer.documentType = customer.documentType;
                if ($scope.fiscalCustomer.documentType == 'CNPJ') {
                    $scope.fiscalCustomer.stateRegistration = customer.stateRegistration;
                }
                $scope.fiscalCustomer.name = $scope.mandatoryName ? customer.name : null;
                $scope.fiscalCustomer.email = customer.email;
                $scope.fiscalCustomer.address = $scope.mandatoryAddress && customer.address ? JSON.parse(JSON.stringify(customer.address)) : {};
            }

            if (!$scope.fiscalCustomer.documentType) {
                $scope.fiscalCustomer.documentType = 'CPF';
            }

            if ($scope.mandatoryAddress && !$scope.fiscalCustomer.address && storeState) {
                // se for obrigatória a identificação do endereço e não tiver sido passado, usa o da UF da loja
                $scope.fiscalCustomer.address = {
                    state: storeState
                };
            }
        }

        $scope.disableCustomerField = (fieldName) => {
            if (!$scope.disableFilledFields) {
                return false;
            }
            if (!$scope.customerCopy) {
                return false;
            }
            return isNotEmpty($scope.customerCopy[fieldName]);
        };

        $scope.disableAddressField = (fieldName) => {
            if (!$scope.disableFilledFields) {
                return false;
            }
            if (!$scope.customerCopy || $scope.customerCopy.address == null) {
                return false;
            }
            return isNotEmpty($scope.customerCopy.address[fieldName]);
        }

        function isNotEmpty(arg) {
            if (arg == null) {
                return false;
            }

            if (typeof arg == 'string') {
                return arg.length > 0;
            }
            return true;
        }

        $scope.fillFiscalAddress = function () {
            let searchPromise = $q.resolve({});

            if (!$scope.mandatoryAddress) {
                return searchPromise; 
            }

            const zipCode = !$scope.fiscalCustomer.address ? null : $scope.fiscalCustomer.address.zipCode;
            const containsZipCode = zipCode != null && zipCode.trim().length > 0;

            if (containsZipCode && isAddressIncomplete($scope.fiscalCustomer.address)) {
                searchPromise = $ionicLoading.show({
					template: 'Consultando endereço...',
					duration: 0
				}).then(() =>
					CustomerService.getAddress(zipCode)
				).finally(() =>
					$ionicLoading.hide()
				);
            }

            return searchPromise.then((addr) => {
				const dAddr = $scope.fiscalCustomer.address;
				dAddr.addressName = chkstr(dAddr.addressName, addr.address, "");
				dAddr.neighborhood = chkstr(dAddr.neighborhood, addr.district, "");
				dAddr.state = chkstr(dAddr.state, addr.state, "");
				dAddr.city = chkstr(dAddr.city, addr.city, "");
				dAddr.cityCode = chkstr(dAddr.cityCode, addr.cityCode, "");
                dAddr.addressNumber = chkstr(dAddr.addressNumber, "", "");
                dAddr.phone = chkstr(dAddr.phone, "", "");
			}).catch((err) =>
                AlertService.restError(err, 'Falha ao consultar CEP')
            );
        }

        function chkstr(str1, str2, str3) {
			return str1 || str2 || str3;
		}
        
        function open(customer, mandatoryFiscalFields, isOverLimit, storeState, disableFilledFields) {
            init(customer, mandatoryFiscalFields, isOverLimit, storeState, disableFilledFields);

            // completa o endereço fiscal do cliente se estiver faltando algum campo
            return $q((resolve, reject) =>
                $scope.fillFiscalAddress()
                .then(() =>
                    $ionicModal.fromTemplateUrl(
                        'js/components/fiscal-customer/fiscal-customer-modal.html',
                        modalOptions
                    )
                ).then((modalInstance) => {
                    $scope.confirm = () => {
                        closeAndRemove(modalInstance)
                        .finally(() => resolve($scope.fiscalCustomer));
                    }
                    $scope.close = function () {
                        closeAndRemove(modalInstance)
                        .finally(() => reject(`Fechado!`));
                    };
                    return modalInstance.show();
                })
            );
        }

        function isUserDoingSelfService(document) {
            const user = UserService.currentUser;
                       
            if (user && user.document && document && user.document ===  document) {
                return true;
            }
            return false;
        }

        $scope.validateDocument = () => {
            if ($scope.fiscalCustomer && $scope.fiscalCustomer.document && $scope.fiscalCustomer.documentType) {
                
                if (isUserDoingSelfService($scope.fiscalCustomer.document)) {
                    AlertService.warn('Proibido auto serviço!').then(() => $scope.fiscalCustomer.document = '');
                } else {
                    const documentType = $scope.fiscalCustomer.documentType;
                    const document = $scope.fiscalCustomer.document;
                    
                    switch (documentType) {
                        case 'CPF':
                            if (document.length === 11 && !validaCPF(document)) {
                                AlertService.warn(documentType + " inválido").then(() => $scope.fiscalCustomer.document = '');
                            }
                            break;
                        case 'CNPJ':
                            if (document.length === 14 && !validaCNPJ(document)) {
                                AlertService.warn(documentType + " inválido").then(() => $scope.fiscalCustomer.document = '');
                            }
                            break;
                        case 'PASSPORT':
                            if (document.length === 9 && !validaPassport(document)) {
                                AlertService.warn("Passaporte inválido").then(() => $scope.fiscalCustomer.document = '');
                            }
                            break;
                        case 'RNE':
                            // TODO qual a validação do RNE?
                            break;
                    }
                }
             }
        }
        function isAddressIncomplete(addr) {
			return addr == null
				|| !addr.addressName
				|| !addr.neighborhood
				|| !addr.state
				|| !addr.city
				|| !addr.cityCode
		}

        function closeAndRemove(modalInstance) {
            return modalInstance.hide()
                .then(() => {
                    return modalInstance.remove();
                });
        }

        const instance = {
            open: open
        };

        return instance;
    }
})();