(() => {
    'use strict';

    angular
        .module('easyorder.modals.fiscalmenu',[])
        .factory('FiscalMenuModal', FiscalMenuModal);

        FiscalMenuModal.$inject = ['$ionicModal', '$rootScope', '$q', '$ionicLoading', 'ConfigService', 'AlertService', 'BarcodeService', 'PAFService'];

    function FiscalMenuModal($ionicModal, $rootScope, $q, $ionicLoading, ConfigService, AlertService, BarcodeService, PAFService) {
        const HEADER = '[FiscalMenuModal]';
        const $scope = $rootScope.$new(),
            modalOptions = {
                scope: $scope,
                focusFirstInput: true
        };

        $scope.SCREEN_PAF_NFCE_IDENTIFICATION = 'paf-nfce-identification';
        $scope.SCREEN_PAF_NFCE_RECORDS = 'paf-nfce-records';
        $scope.SCREEN_IDENTIFIED_SALES = 'identified-sales';
        $scope.SCREEN_EXTERNAL_REQUESTS_REGISTERED = 'external-requests-registered';

        let isScannerActivated = false;

        function open() {
            // por default, armazena localmente
            const localFile = ConfigService.config.business.paf_file_storage_mode !== 'REMOTE';
            // configura o serviço para operar no modo do arquivo local ou remoto
            PAFService.setLocalFile(localFile);
            isScannerActivated = BarcodeService.scanActivated();
            if (isScannerActivated) {
                BarcodeService.cancel();
            }
            $ionicModal.fromTemplateUrl(
                'js/components/fiscal-menu/fiscal-menu-modal.html',
                modalOptions
            ).then(function (modalInstance) {
                    $scope.close = function () {
                        closeAndRemove(modalInstance)
                            .then(() => {
                                if (isScannerActivated) {
                                    BarcodeService.reInitScan();
                                }
                            });
                    };

                    return modalInstance.show();
            });
        }

        /**
         *
         */

        $scope.openNestedModal = (screenName) => {
            $scope.currentScreen = screenName;
            $scope.searchParams = {
                stockType: 'FULL'
            };
            $ionicModal.fromTemplateUrl(
                `js/components/fiscal-menu/${screenName}.html`, {
                    scope: $scope,
                    focusFirstInput: true,
                    animation: 'none'
                }).then(function (modalInstance) {
                    $scope.closeNested = function () {
                        closeAndRemove(modalInstance);
                    };

                    return modalInstance.show();
            });
        }


        $scope.openPafNfceIdentification = () => {
            $ionicLoading.show({
                template : 'Por favor, aguarde...',
                duration : 0
            }).then(() => {
                return PAFService.pafNfceIdentification();
            }).then(result => {
                $ionicLoading.hide().then(() => {
                    if (result.cnpj) {
                        $scope.res = result;
                        $scope.openNestedModal($scope.SCREEN_PAF_NFCE_IDENTIFICATION)
                    } else {
                        AlertService.error("Nenhuma informação encontrada!");
                    }
                });
            }).catch(err => {
                return $ionicLoading.hide().then(() => {
                    $scope.loadingState = false;
                    return AlertService.restError(err, err);
                });
            });
        }

        // 2) Registros do PAF-NFC-e
        function validatePafNfceRecordsParams() {
            return new Promise((resolve, reject) => {
                const initialDate = toMoment($scope.searchParams.initialDate);
                if (!initialDate.isValid()) {
                    reject('Data inicial inválida!');
                } else {
                    const finalDate = toMoment($scope.searchParams.finalDate);
                    if (!finalDate.isValid()) {
                        reject('Data final inválida!');
                    } else {
                        if (initialDate > finalDate) {
                            reject('Data inicial superior à data final!');
                        } else {
                            var now = moment();
                            if (finalDate > now) {
                                reject('Data final superior à data atual!');
                            } else {
                                if ($scope.searchParams.stockType === 'PARTIAL') {
                                    const hasProductCodes = $scope.searchParams.productCodes && $scope.searchParams.productCodes.length;
                                    const hasProductDescription = $scope.searchParams.productDescriptions && $scope.searchParams.productDescriptions.length;
                                    if (hasProductCodes || hasProductDescription) {
                                        resolve();
                                    } else {
                                        reject('Estoque parcial precisa de pelo menos um código ou descrição!');
                                    }
                                } else {
                                    resolve();
                                }

                            }
                        }
                    }
                }
            });
        }

        $scope.changeStockType = () => {
            if ($scope.searchParams.stockType === 'FULL') {
                $scope.searchParams.productCodes = null;
                $scope.searchParams.productCode = '';
                $scope.searchParams.productDescriptions = null;
                $scope.searchParams.productCode = '';
            }
        }

        $scope.addProductCode = () => {
            if (!$scope.searchParams.productCodes) {
                $scope.searchParams.productCodes = [];
            }
            if (!$scope.searchParams.productCode) {
                AlertService.error('Insira um código de produto!');
            } else if ($scope.searchParams.productCodes.includes($scope.searchParams.productCode)) {
                AlertService.warn('Código de produto já adicionado!');
            } else {
                $scope.searchParams.productCodes.push($scope.searchParams.productCode);
                $scope.searchParams.productCode = '';
            }
        }

        $scope.removeProductCode = (productCode) => {
            const productCodes = $scope.searchParams.productCodes;
            if (productCodes) {
                const index = productCodes.indexOf(productCode);
                if (index > -1) {
                    productCodes.splice(index, 1);
                }
            }
        }

        $scope.addProductDescription = () => {
            if (!$scope.searchParams.productDescriptions) {
                $scope.searchParams.productDescriptions = [];
            }
            if (!$scope.searchParams.productDescription) {
                AlertService.error('Insira uma descrição de produto!');
            } else if ($scope.searchParams.productDescriptions.includes($scope.searchParams.productDescription)) {
                AlertService.warn('Descrição de produto já adicionada!');
            } else {
                $scope.searchParams.productDescriptions.push($scope.searchParams.productDescription);
                $scope.searchParams.productDescription = '';
            }
        }

        $scope.removeProductDescription = (productDescription) => {
            const productDescriptions = $scope.searchParams.productDescriptions;
            if (productDescriptions) {
                const index = productDescriptions.indexOf(productDescription);
                if (index > -1) {
                    productDescriptions.splice(index, 1);
                }
            }
        }

        function sendPafNfceRecordsRequest() {
            return PAFService.generatePafNfceRecords(
                ConfigService.config.companyNumber,
                ConfigService.config.storeNumber,
                ConfigService.config.terminalNumber,
                toDate($scope.searchParams.initialDate),
                toDate($scope.searchParams.finalDate),
                $scope.searchParams.stockType == 'FULL'? null : $scope.searchParams.productCodes,
                $scope.searchParams.stockType == 'FULL'? null : $scope.searchParams.productDescriptions
            );
        }

        $scope.generatePafNfceRecords = () => {
            validatePafNfceRecordsParams()
                .then(() => {
                    $ionicLoading.show({
                        template : 'Por favor, aguarde...',
                        duration : 0
                    }).then(() => {
                        return sendPafNfceRecordsRequest();
                    }).then(result => {
                        handlePafResponse(result);
                    }).catch(err => {
                        return $ionicLoading.hide().then(() => {
                            $scope.loadingState = false;
                            return AlertService.error("Erro ao gerar arquivo no servidor!");
                        });
                    });
                }).catch(errorMessage => {
                    AlertService.error(errorMessage);
                })
        }

        // 3) Vendas Identificadas pelo CPF/CNPJ
        function validateDocument(document) {
            // válido se o documento for vazio ou CPF/CNPJ for válido
            var validated = false;
            if (document) {
                if (document.length === 11 && validaCPF(document)) {
                    validated = true;
                } else if (document.length === 14 && validaCNPJ(document)) {
                    validated = true;
                }
            } else {
                validated = true;
            }
            return validated;
        }

        function validateIdentifiedSalesParams() {
            return new Promise((resolve, reject) => {
                const document = $scope.searchParams.document;
                if (!validateDocument(document)) {
                    reject(`CPF/CNPJ inválido!`);
                } else {
                    const date = $scope.searchParams.date;
                    if (!date || date.length !== 6) {
                        reject('Data inválida!');
                    } else {
                        const month = parseInt(date.substr(0,2));
                        if (month < 1 || month > 12) {
                            reject('Mês inválido!');
                        } else {
                            const year = parseInt(date.substr(2,6));
                            const currentTime = new Date();
                            const currentYear = currentTime.getFullYear();
                            const currentMonth = currentTime.getMonth() + 1;
                            if (year > currentYear) {
                                reject('Ano superior ao atual!');
                            } else if (year === currentYear && month > currentMonth) {
                                reject('Mês superior ao atual!');
                            } else {
                                resolve();
                            }
                        }
                    }
                }
            });
        }

        function sendIdentifiedSalesRequest() {
            return PAFService.generateIdentifiedSales(
                ConfigService.config.companyNumber,
                ConfigService.config.storeNumber,
                ConfigService.config.terminalNumber,
                toDate('01' + $scope.searchParams.date), // data só mês e ano
                $scope.searchParams.document,
                {
                    name: $rootScope.APP_NAME_DESC,
                    version: $rootScope.PAF_VERSION
                }
            );
        }

        const extractAndSaveFileContent = (fileName, fileContent) => {
            const defer = $q.defer();
            if (fileContent) {
                try {;
                    $rootScope.dataSourceManager.savePafFile(fileName, fileContent)
                    .then(fileName => {
                        defer.resolve(`Arquivo gerado no dispositivo em '${fileName}'!`);
                    }).catch(error => {
                        console.log(error);
                        defer.reject("Erro ao salvar arquivo!");
                    });

                } catch (err) {
                    console.log(err);
                    defer.reject("Erro ao descompactar arquivo!");
                }

            } else {
                defer.reject("Arquivo não foi retornado pelo servidor!");
            }
            return defer.promise;
        }

        let handlePafResponse = (result) => {
            if (result && result.data) {
                const pafResponse = result.data;
                if (pafResponse instanceof PafLocalFileResponse) {
                    const gzip = pafResponse.gzip;
                    const fileName = pafResponse.fileName;
                    if (gzip) {
                        extractAndSaveFileContent(fileName, gzip).then((message) => {
                            $ionicLoading.hide();
                            AlertService.info(message).then(() => $scope.closeNested());

                        }).catch((error) => {
                            $ionicLoading.hide();
                            AlertService.error(error);
                        })
                    } else {
                        $ionicLoading.hide();
                        AlertService.error("Arquivo não foi retornado pelo servidor!");
                    }

                } else if (pafResponse instanceof PafRemoteFileResponse) {
                    $ionicLoading.hide();
                    const message = pafResponse.message;
                    if (message) {
                        AlertService.info(message).then(() => $scope.closeNested());
                    } else {
                        AlertService.error("Resposta não foi retornado pelo servidor!");
                    }
                } else {
                    $ionicLoading.hide();
                    AlertService.error("Tipo de resposta inválida!");
                }
            } else {
                $ionicLoading.hide();
                AlertService.error("Dados não foram retornados pelo servidor!");
            }
        }
        $scope.generateIdentifiedSales = () => {
            validateIdentifiedSalesParams()
                .then(() => {
                    $ionicLoading.show({
                        template : 'Por favor, aguarde...',
                        duration : 0
                    }).then(() => {
                        return sendIdentifiedSalesRequest();
                    }).then(result => {
                        handlePafResponse(result);
                    }).catch(err => {
                        return $ionicLoading.hide().then(() => {
                            $scope.loadingState = false;
                            return AlertService.error("Erro ao gerar arquivo no servidor!");
                        });
                    });
                }).catch(errorMessage => {
                    AlertService.error(errorMessage);
                })
        }

        // 4) Requisições Externas Registradas
        function validateExternalRequestsRegisteredParams() {
            return new Promise((resolve, reject) => {
                const date = toMoment($scope.searchParams.date);
                if (!date.isValid()) {
                    reject('Data inválida!');
                } else {
                    var now = moment();
                    if (date > now) {
                        reject('Data superior à data atual!');
                    } else {
                        resolve();
                    }


                }
            });
        }

        function sendExternalRequestsRegisteredRequest() {
            return PAFService.generateExternalRequestsRegistered(
                ConfigService.config.companyNumber,
                ConfigService.config.storeNumber,
                toDate($scope.searchParams.date),
                {
                    name: $rootScope.APP_NAME_DESC,
                    version: $rootScope.PAF_VERSION
                }
            );
        }

        $scope.generateExternalRequestsRegistered = () => {
            validateExternalRequestsRegisteredParams()
                .then(() => {
                    $ionicLoading.show({
                        template : 'Por favor, aguarde...',
                        duration : 0
                    }).then(() => {
                        return sendExternalRequestsRegisteredRequest();
                    }).then(result => {
                        handlePafResponse(result);
                    }).catch(err => {
                        return $ionicLoading.hide().then(() => {
                            $scope.loadingState = false;
                            return AlertService.error("Erro ao gerar arquivo no servidor!");
                        });
                    });
                }).catch(errorMessage => {
                    AlertService.error(errorMessage);
                })
        }

        // 5) "Mesas Abertas"
        $scope.showOpenTablesMessage = () => {
            AlertService.warn("Este PAF-NFC-e não executa funções de atendimento de mesas, não podendo ser utilizado por estabelecimento que necessite desse recurso.");
        }

        // 6) "Controle dos DAV"
        function sendControlDavRequest() {
            return PAFService.generateControlDav();
        }

        $scope.generateControlDav = () => {
            $ionicLoading.show({
                    template : 'Por favor, aguarde...',
                    duration : 0
            }).then(() => {
                 return sendControlDavRequest();
            }).then(result => {
                 handlePafResponse(result);
            }).catch(err => {
                 return $ionicLoading.hide().then(() => {
                         $scope.loadingState = false;
                         return AlertService.error("Erro ao gerar arquivo no servidor!");
                 });
            });
        }

        function toMoment(value) {
            if (!value) {
                return;
            }
            return moment(value, "DDMMYYYY");
        }

        function toDate(value) {
            const moment = toMoment(value);
            if (!moment) {
                return;
            }
            return moment.toDate();
        }

        function closeAndRemove(modalInstance) {
            return modalInstance.hide()
                .then(() => {
                    return modalInstance.remove();
                });
        }

        const instance = {
            open: open
        };

        return instance;
    }
})();
