(() => {
    'use strict';

    angular
        .module('easyorder.modals.tagsreader',[])
        .factory('TagsReaderModal', TagsReaderModal);

    TagsReaderModal.$inject = ['$ionicModal', '$rootScope', '$q', '$ionicPopup', 'AlertService', 'LogService', 'BarcodeService', 'UserService'];

    function TagsReaderModal($ionicModal, $rootScope, $q, $ionicPopup, AlertService, LogService, BarcodeService, UserService) {
        const HEADER = '[TagsReaderModal]'; 
        const $scope = $rootScope.$new(),
            modalOptions = {
                scope: $scope,
                focusFirstInput: true
        };

        $scope.STATUS_READ_INTERNAL_TAG = 1;
        $scope.STATUS_READ_KIT_ITEMS = 2;
        $scope.barcodeService = BarcodeService;

        function reset() {
            $scope.totalItemsKit = 0;
            $scope.kitTagsRead = [];
            $scope.externalTag = null;
            $scope.sku = null;
        };

        function handleInternalTag(tag) {
            if (tag.topBottom !== '1') {
                AlertService.error(`Etiqueta lida '${tag.fullTag}' não é etiqueta interna!`)
                    .finally(() => initScanner());
            } else {
                addInternalTag(tag);
            }
        }

        function handleKitTag(tag) {
            // A comparação não é pela etiqueta inteira, mas apenas pelo sku + top bottom
            const tagAlreadyAdded = $scope.kitTagsRead.find(currentTag => currentTag.uniqueId === tag.uniqueId);
            if (tagAlreadyAdded) {
                AlertService.error(`Etiqueta '${tag.fullTag}' já foi lida!`)
                    .finally(() => initScanner());
            } else {
                pushItemKit(tag);
            }
        }

        function pushItemKit(tag) {
            $scope.kitTagsRead.push(tag);
            
            if ($scope.kitTagsRead.length === $scope.totalItemsKit) {
                // OK, retorna
                LogService.info(`Todos os ${$scope.totalItemsKit} ítens do kit foram lidos!`);
                $scope.onTagReadComplete(tag);
            } else {
                 // ainda faltam ítens a ler
                updateTitle();
                AlertService.info(`Por favor, efetue a leitura da etiqueta do próximo ítem!`)
                    .finally(() => initScanner());
            }

        }

        function init(tag, totalItemsKit) {
            reset();
            LogService.info(`Iniciando leitura de kit/interna para a etiqueta '${tag.fullTag}'. Total de ítens no kit: ${totalItemsKit}.`);
            
            $scope.totalItemsKit = totalItemsKit;
            $scope.sku = tag.sku;

            if (tag.isExternal()) {
                // etiqueta externa
                $scope.externalTag = tag;
                setStatus($scope.STATUS_READ_INTERNAL_TAG);
                AlertService.info(`Por favor, efetue a leitura da etiqueta interna do produto!`)
                    .finally(() => initScanner());
            } else {
                setStatus($scope.STATUS_READ_KIT_ITEMS);
                pushItemKit(tag);
            }
        }
        
        function setStatus(status) {
            $scope.status = status;
            updateTitle();
        }

        function updateTitle() {
            $scope.title = $scope.status === $scope.STATUS_READ_INTERNAL_TAG ? 'Etiqueta Interna' : `Leitura de Kit (${$scope.kitTagsRead.length}/${$scope.totalItemsKit})`;
        }

        function isValidTagLength(code) {
            return code && code.length === 18;
        }

        function onScannedCode(scannedCode) {
            if (!isValidTagLength(scannedCode)) {
                AlertService.error(`Código '${scannedCode}' não é uma etiqueta válida!`)
                    .finally(() => initScanner());
            } else {
                const tag = new CeaTag(scannedCode);
                const newSku = tag.sku;
                if (newSku !== $scope.sku) {
                    AlertService.error(`Código do produto da etiqueta '${newSku}' é diferente da anterior '${$scope.sku}'!`)
                        .finally(() => initScanner());
                } else if (tag.isExternal()) {
                    AlertService.error(`Etiqueta lida '${tag.fullTag}' é uma etiqueta externa!`)
                        .finally(() => initScanner());
                } else {
                    if ($scope.status === $scope.STATUS_READ_INTERNAL_TAG) {
                        handleInternalTag(tag);
                    } else {
                        handleKitTag(tag);
                    }
                }
            }
        }

        function initScanner() {
            const inParam = '0/44/0/74';
            BarcodeService.init(function success(session) {
                BarcodeService.cancel();
                const scannedCode = session[0];
                onScannedCode(scannedCode);
            }, { inSizeCam: [inParam, '2/64/2/148'], outSizeCam: ['275/44/0/128', '275/64/0/148'] });
        };

        function addInternalTag(tag) {
            // etiqueta ok, se só tem 1 item no kit, finaliza
            if ($scope.totalItemsKit === 1) {
                $scope.onTagReadComplete(tag);
            } else {
                // se tem mais ítens, inicia a busca dos ítens do kit
                setStatus($scope.STATUS_READ_KIT_ITEMS);
                pushItemKit(tag);
            }
        }

        function onInternalTagTyped(typedCode) {
            let defered  = $q.defer();
            if (!typedCode) {
                AlertService.warn(`Informe um código de etiqueta interna!`)
                    .finally(() => defered.reject());
            } else if (!isValidTagLength(typedCode)) {
                AlertService.error(`Código '${typedCode}' não é uma etiqueta válida!`)
                    .finally(() => defered.reject());
            } else {
                const tag = new CeaTag(typedCode);
                const newSku = tag.sku;
                if (newSku !== $scope.sku) {
                    AlertService.error(`Código do produto da etiqueta '${newSku}' é diferente da anterior '${$scope.sku}'!`)
                        .finally(() => defered.reject());
                } else if (tag.isExternal()) {
                    AlertService.error(`Etiqueta lida '${tag.fullTag}' é uma etiqueta externa!`)
                        .finally(() => defered.reject());
                } else if (tag.topBottom !== '1') {
                    AlertService.error(`Etiqueta lida '${tag.fullTag}' não é etiqueta interna!`)
                        .finally(() => defered.reject());
                } else {
                    defered.resolve();
                }
            }
            return defered.promise;
        }
        $scope.manualTyping = () => {
            LogService.info("Exibindo popup de digitação de etiqueta interna.");
            BarcodeService.cancel();
            $scope.itemPopup = {};
            
			$scope.popupInternalTagCode = $ionicPopup.show({
				templateUrl: 'views/enter-product-code.html',
				title: 'Etiqueta Interna',
				scope: $scope,
				buttons: [
					{
						text: 'Cancelar',
						onTap: function () {
							initScanner();
						}
					},
					{
						text: 'OK',
						onTap: function (e) {
							e.preventDefault();
                            onInternalTagTyped($scope.itemPopup.value).then(() => {
                                $scope.popupInternalTagCode.close();
                                addInternalTag($scope.itemPopup.value);
                            });
						}
					},
				]
			});
        };

        function open(tag, totalItemsKit) {
            return new Promise((resolve, reject) => {
                if (tag) {
                    init(tag, totalItemsKit);
                    $ionicModal.fromTemplateUrl(
                        'js/components/tags-reader/tags-reader-modal.html',
                        modalOptions
                    ).then(function (modalInstance) {
                            $scope.onTagReadComplete = function (tag) {
                                closeAndRemove(modalInstance)
                                    .finally(() => resolve(tag.productCode));
                            };
                            $scope.close = function () {
                                closeAndRemove(modalInstance)
                                    .finally(() => reject(`Leitura de etiquetas cancelada! Produto não será adicionado ao carrinho!`));
                            };
                            
                            return modalInstance.show();
                    });
                } else {
                    reject(`Etiqueta inválida!`);
                }
            });
        }

        function closeAndRemove(modalInstance) {
            return Promise.resolve(BarcodeService.cancel())
                .then(() => modalInstance.hide())
                .then(() => modalInstance.remove());
        }

        $scope.authorizeKit = () => {
			$scope.requestauth = {};
			var permission = UserService.checkPermission(KIT_SALE);

			if (permission) {
	            $scope.onTagReadComplete($scope.kitTagsRead[0]);
			} else {
                BarcodeService.cancel();
				var authPopup = $ionicPopup
					.show({
						templateUrl : 'views/requestauth.html',
						scope : $scope,
						buttons : [
								{
									text : 'Cancelar',
                                    onTap: function(e) {
                                        initScanner();
                                    }
								},
								{
									text : '<b>Autorizar</b>',
									type : 'button-theme',
									onTap : function(event) {
										var sucesso = function(result) {
											if (result.authorized) {
                                                $scope.onTagReadComplete($scope.kitTagsRead[0]);
											} else {
												AlertService.error('Autorização negada!')
                                                    .finally(() => initScanner());
                                                
											}
										};

										var erro = function(error) {
											AlertService.restError(error,
													"Erro ao consultar autorização!")
                                                    .finally(() => initScanner());
										};

										if (!$scope.requestauth.user) {
											AlertService.error('Usuário inválido!')
                                                .finally(() => initScanner());
										} else if (!$scope.requestauth.password) {
											AlertService.error('Senha inválida!')
                                                .finally(() => initScanner());
										} else {
											UserService.checkAuthorization($scope.requestauth.user,
													$scope.requestauth.password,
													[KIT_SALE],
													sucesso, erro);
										}
									}
								}]
					});
			}
		}

        const instance = {
            open: open
        };

        return instance;
    }
})();
