(function () {
    'use strict';
    angular.module('easyorder.controllers.annulment', [])
        .controller('AnnulmentCtrl', function (
            $rootScope,
            $scope,
            $timeout,
            TransactionReturnService,
            ConfigService,
            UtilService,
            $ionicModal,
            AlertService,
            $ionicLoading,
            CartService,
            $stateParams,
            UserService,
            EFTService,
            EWalletService,
            $ionicPopup,
            LogService,
            $q,
            PaymentLinkService) {

            function annulment(reversedPayments) {
                const returnData = $stateParams.returnData;
                const nsu = ConfigService.nsuControll.incrementNSU();
                return TransactionReturnService.return({
                    appInfo: {
                        name: $rootScope.APP_NAME,
                        version: $rootScope.appVersion
                    },
                    sourceTransactionKey: returnData.sourceTransaction.key,
                    transactionKey: {
                        storeInfo: {
                            companyNumber: ConfigService.config.companyNumber,
                            storeNumber: ConfigService.config.storeNumber,
                        },
                        date: new Date(),
                        nsu: nsu,
                        terminal: ConfigService.config.terminalNumber,
                    },
                    returnCouponNumber: nsu,
                    operationType: 'ANNULMENT',
                    returnReason: $stateParams.returnReason,
                    operator: {
                        code: UserService.operatorCode(),
                        name: UserService.operatorName()
                    },
                    authorizer: {
                        code: $stateParams.authorizer && $stateParams.authorizer.id ? $stateParams.authorizer.id : $stateParams.authorizer,
                        name: $stateParams.authorizer && $stateParams.authorizer.name ? $stateParams.authorizer.name : null
                    },
                    customer: CartService.cart.customer,
                    fiscalItems: returnData.fiscalItems,
                    sourceTransactionTotalValue: returnData.sourceTransaction.totalValue,
                    totalValue: returnData.fiscalItems
                        .reduce((acc, item) => new Number((acc + item.totalValue).toFixed(2)), 0),
                    reversedPayments: reversedPayments,
                });
            }


            $scope.goBack = () => {
                UtilService.navigateWithoutHistory("easyorder.operations");
            };

            $scope.saleAnnulment = () => {
                $ionicLoading.show({
                    template : 'Aguarde...',
                    duration: 0,
                })
                .then(() => returnPayments($stateParams.sourceTransaction))
                .then(reversedPayments => annulment(reversedPayments))
                .then(result => confirmTransactions().then(() => result))
                .then(result => {
                    $ionicLoading.hide().then(() => {
                        UtilService.navigateWithoutHistory('easyorder.couponview', {
                            pdf: result.pdf,
                            printContents: result.printContents,
                            documentTitle: 'AnulacaoVenda'
                        });
                    });
                }).catch(err => {
                    if (!err) {
                        return;
                    }
                    return undoTransactions().finally(() => {
                        $ionicLoading.hide().then(() => {
                            $scope.loadingState = false;

                            AlertService.restError(err, err)
                                .then(() => {
                                    AlertService.confirm('Deseja tentar efetuar a anulação novamente?')
                                        .then(confirm => confirm ? $scope.saleAnnulment() : $scope.goBack());
                                });
                        });
                    });
                });
            };

            function returnPayments(srcTrns) {
                const nonAllowedPmnt = srcTrns.payments.some(
                    pmnt => !eftFilter(pmnt)
                        && !qrLinxFilter(pmnt)
                        && !paymentLinkFilter(pmnt)
                );
                if (nonAllowedPmnt) {
                    return $q.reject('Só é possível cancelamento de Crédito, Débito, QR Linx e Link de Pagamentos!');
                }

                $scope.reversedPayments = [];
                // Estornaremos primeiro os TEFS, pois
                // o QRLinx não dá pra desfazer a operação
                // de estorno, TEFS podem existir perguntas
                // onde o usuário pode desistir da operação
                return srcTrns.payments.sort((p1, p2) => {
                    const isP1EFT = eftFilter(p1);
                    const isP2EFT = eftFilter(p2);

                    if (isP1EFT) {
                        return isP2EFT ? 0 : -1;
                    }

                    return isP2EFT ? 1 : 0;
                }).reduce((acc, pmnt) =>
                    acc.then(_ =>
                        returnPayment(pmnt)
                    ),
                    $q.resolve()                        
                ).then(() =>
                    $scope.reversedPayments
                );
            }

            function returnPayment(payment) {
                if (qrLinxFilter(payment)) {
                    return returnQRLinxPayment(payment);
                }
                if (eftFilter(payment)) {
                    return returnEFTPayment(payment);
                }
                if (paymentLinkFilter(payment)) {
                    return returnPaymentLink(payment);
                }
                return $q.reject(`FATAL: Forma de pagamento ${payment.paymentMethod}`
                                 + `não é esperada neste trecho, contate o suporte`);
            }

            function returnQRLinxPayment(payment) {
                return EWalletService.cancelPayment({
                    paymentId: payment.id,
                    remoteTerminalId: payment.remoteTerminalId,
                    walletType: payment.paymentMethod
                }).then(() => {
                    $scope.reversedPayments.push(payment);
                    return payment;
                });
            }

            function returnEFTPayment(payment) {
                const deferred = $q.defer();

                let paymentDate = new Date($rootScope.transactionDate);
                paymentDate.setHours(12,0,0);

                EFTService.executeTransaction(
                    {
                        amount: new Number(payment.value),
                        transactionType: EFTService.transactionType.CANCEL,
                        operatorCode: UserService.operatorCode(),
                        transactionDate: paymentDate.toISOString(),  
                        transactionNSU: payment.id,
                    },
                    {
                        paymentDone: res => {
                            $ionicLoading.show({
                                template : 'Aguarde...',
                                duration: 0,
                            }).finally(() => {
                                const pmnt = {
                                    id: res.NSURede,
                                    paymentMethod: payment.paymentMethod,
                                    date: new Date(),
                                    value: new Number(payment.value),
                                    customerReceipt: res.ComprovanteCliente,
                                    businessReceipt: res.ComprovanteEstabelecimento,
                                    paymentAuthorizationNSU: res.NSU,
                                    cardNumber: res.NumeroCartao,
                                    lastCardDigits: res.UltimosDigitosCartao,
                                    cardExpirationDate: res.DataValidadeCartao
                                };
                                // salvando pagamento que foi estornado
                                $scope.reversedPayments.push(pmnt);
                                deferred.resolve(pmnt);
                            });
                            
                        },
                        paymentError: error => {
                            deferred.reject(error);
                        },
                        displayMessage: message => {
                            if (message) {
                                $timeout(function () {
                                    $scope.message = message;
                                });
                                $ionicLoading.show({
                                    template : message,
                                    duration: 0,
                                });
                            }
                            LogService.info("displayMessage: " + $scope.message);
                        },
                        clearDisplay: () => {
                            $timeout(function () {
                                $scope.message = "";
                            });
                            LogService.info("clearDisplay");
                        },
                        selectOption: (title, options, responseHandler) => {
                            $q((resolve) => {
                                $scope.message = title;
                                processOptions(options, responseHandler);
                                LogService.info("Opção tef: ", title, " Opções: ", options);
                                autoSelectResponse($scope.message, $scope.options, payment.paymentMethod, $stateParams.sourceTransaction.payments)
                                    .then(
                                        () => LogService.info('Opção automática')                                        
                                    )
                                    .catch(() => {
                                        $ionicModal.fromTemplateUrl('views/tef-modal.html', {
                                            scope: $scope,
                                            animation: 'slide-in-up',
                                            hardwareBackButtonClose: false,
                                            backdropClickToClose: false,
                                        })
                                        .then(modal => {
                                            $scope.modal = modal;
                                            $ionicLoading.hide()
                                            .then(() => {
                                                $scope.modal.show();
                                            });
                                        });
                                    })
                                resolve();
                            }).then(() => {});
                        },
                        selectOptionExtended: (title, options, responseHandler) => {
                            $q((resolve) => {
                                $scope.message = title;
                                processOptions(options, responseHandler);
                                LogService.info("Opção estendida: ", title, " Opções: ", options);
                                autoSelectResponse($scope.message, $scope.options, payment.paymentMethod, $stateParams.sourceTransaction.payments)
                                    .then(
                                        () => LogService.info('Opção automática')                                        
                                    )
                                    .catch(() => {
                                        $ionicModal.fromTemplateUrl('views/tef-modal.html', {
                                            scope: $scope,
                                            animation: 'slide-in-up',
                                            hardwareBackButtonClose: false,
                                            backdropClickToClose: false,
                                        })
                                        .then(modal => {
                                            $scope.modal = modal;
                                            $ionicLoading.hide()
                                            .then(() => {
                                                $scope.modal.show();
                                            });
                                        });
                                    })
                                resolve();
                            }).then(() => {});
                        },
                        confirmAction: (message) => {
                            return $ionicLoading.hide().then(() =>
                                $ionicPopup.confirm({
                                    title: 'TEF - Anulação',
                                    template: message,
                                    okType: "button-theme"
                                })
                            ).finally(() =>
                                $ionicLoading.show({
                                    template : 'Aguarde...',
                                    duration : 0
                                })
                            );
                        },
                        alertMessage: (message) => {
                            return $ionicLoading.hide().then(() =>
                                AlertService.show('TEF - Anulação', message)
                            ).finally(() =>
                                $ionicLoading.show({
                                    template : 'Aguarde...',
                                    duration : 0
                                })
                            );
                        }
                    }
                );
                return deferred.promise;
            }

            function autoSelectResponse(title, options, paymentMethod, payments) {
                //  TODO 'Work around' temporário pelo prazo de entrega. Será evoluído o plugin do TEF para suportar de forma automática
                const deferred = $q.defer();
                const translatedPaymentMethod = paymentMethod === 'DEBIT' ? 'DEBITO' : 'CREDITO';
                let optionSelected = null;
                let isSelectingCartType = false;
                if (title && options) {
                    const titleUpper = title.toUpperCase().trim();
                    if (titleUpper === 'SELECIONE O TIPO DE CANCELAMENTO') {
                        isSelectingCartType = true;
                        optionSelected = options.find(opt => opt.label.toUpperCase().contains(translatedPaymentMethod));
                    } else if (titleUpper.contains('SELECIONE O TIPO DO CARTAO')) {
                        optionSelected = options.find(opt => opt.label.toUpperCase().contains('CHIP'));
                    }
                }

                if (optionSelected) {
                    // verifica se existem formas diferentes para só exibir a mensagem se existirem múltiplas formas
                    const multipleEfts = isSelectingCartType && payments.length > 1; 
                    if (multipleEfts) {
                        $ionicLoading.hide().then(() => {
                            AlertService.info(`Por favor, utilize agora o próximo cartão de ${paymentMethod === 'DEBIT' ? 'débito' : 'crédito'}, na mesma ordem efetuada na venda!`)
                                .then(() => {
                                    LogService.info(`Opção do TEF foi selecionada automaticamente: ${optionSelected.label}, valor ${optionSelected.id} para o método ${paymentMethod}.`)
                                    $scope.tickOptionHandler(optionSelected.handler, [optionSelected.id, optionSelected.label, optionSelected.value])
                                    deferred.resolve();
                                });
                        });

                    } else {
                        LogService.info(`Opção do TEF foi selecionada automaticamente: ${optionSelected.label}, valor ${optionSelected.id} para o método ${paymentMethod}.`)
                        $scope.tickOptionHandler(optionSelected.handler, [optionSelected.id, optionSelected.label, optionSelected.value])
                        deferred.resolve();
                    }
                } else {
                    deferred.reject();
                }
                return deferred.promise;
            }

            function confirmTransactions() {
                if ($scope.reversedPayments == null ||
                        $scope.reversedPayments.length == 0) {
                    return $q.resolve();
                }

                let ret = $q.resolve();
                const eftPayments = $scope.reversedPayments.filter(eftFilter);

                if (eftPayments.length > 0) {
                    if (EFTService.isDTEF()) {
                        ret = ret.then(() => eftPayments.reduce(
                            (acc, payment) => acc.then(() => EFTService.confirmTransaction(payment.paymentAuthorizationNSU)),
                            $q.resolve()
                        ));
                    } else {
                        ret = EFTService.confirmTransaction(eftPayments[0].paymentAuthorizationNSU);
                    }
                }
                    
                return ret.then(() => {
                    delete $scope.reversedPayments;
                });
            }

            function undoTransactions() {
                if ($scope.reversedPayments == null ||
                        $scope.reversedPayments.length == 0) {
                    return $q.resolve();
                }

                let ret = $q.resolve();
                const eftPayments = $scope.reversedPayments.filter(eftFilter);

                if (eftPayments.length > 0) {
                    if (EFTService.isDTEF()) {
                        ret = ret.then(() => eftPayments.reduce(
                            (acc, payment) => acc.then(() => EFTService.undoTransaction(payment.paymentAuthorizationNSU)),
                            $q.resolve()
                        ));
                    } else {
                        ret = EFTService.undoTransaction(eftPayments[0].paymentAuthorizationNSU);
                    }
                }

                return ret.then(() => {
                   delete $scope.reversedPayments;
                });
            }

            function returnPaymentLink(payment) {
                const paymentLinkId = findPaymentLinkId(payment);
                const returned = $scope.reversedPayments
                    .filter(paymentLinkFilter)
                    .some(p => paymentLinkId == findPaymentLinkId(p));

                if (returned) {
                    $scope.reversedPayments.push(payment);
                    return $q.resolve(payment);
                }

                return PaymentLinkService
                    .disablePaymentLink(paymentLinkId)
                    .then(() => {
                        $scope.reversedPayments.push(payment);
                        return payment;
                    });
            }

            function findPaymentLinkId(payment) {
                return payment.paymentMethod == 'PAYMENT_LINK' ?
                    payment.id : payment.groupId;
            }

            function eftFilter(payment) {
                const eftPaymentMethods = ['CREDIT', 'DEBIT'];
                return eftPaymentMethods.includes(payment.paymentMethod)
                    && !isLinkProvider(payment.provider)
            }

            function qrLinxFilter(payment) {
                return payment.paymentMethod == 'QR_LINX';
            }

            function paymentLinkFilter(payment) {
                return payment.paymentMethod == 'PAYMENT_LINK'
                    || isLinkProvider(payment.provider);
            }

            function isLinkProvider(provider) {
                if (provider == null) {
                    return false;
                }

                const linkProviders = [
                    'PAGAR_ME',
                    'LINX_PAYHUB_LINK',
                    'E_SITEF_WEB_CHECKOUT'
                ];

                return linkProviders.includes(provider.id);
            }

            function processOptions(options, responseHandler) {
                $scope.showWatermark = false;
                $scope.options = [];

                var splitedOpt = options.replace(/\(|\)/g, ""); //removendo "(" e ")"
                splitedOpt = splitedOpt.split("#"); // Separando as opções.

                for (var i = 0; i < splitedOpt.length; i++) {
                    var split = splitedOpt[i].split(",");
                    $scope.options.push({
                        id:         split[0].trim(),
                        label:      split[1].trim().replace(/\"/g, ""),
                        value:      split[split.length-1],
                        handler:    responseHandler,
                        type:        "single-option"
                    });
                }
                $scope.headerOptions = "Opções";
            };

            $scope.tickNumberInputHandler = function(inputOption) {
                var value = sprintf('%1$' + (inputOption.maxSize ? '0' + inputOption.maxSize : '') + 'd', inputOption.value);

                $scope.tickOptionHandler(inputOption.handler, [value]);
            };

            $scope.tickStringInputHandler = function(inputOption) {
                const value = sprintf('%1$' + (inputOption.maxSize ? '0' + inputOption.maxSize : '') + 's', inputOption.value);

                $scope.tickOptionHandler(inputOption.handler, [value]);
            };

            $scope.tickOptionHandler = function(func, params) {
                $scope.options.length = 0;
                $scope.showWatermark = true;

                if ($scope.modal) {
                    $scope.modal.hide();
                }
                
                $ionicLoading.show();

                func.apply(this, params);
            };

            $scope.saleAnnulment();
        })
})();
