(function() {
    'use strict';

    angular.module('easyorder.controllers.installment-plan-payment', [])

    .controller('InstallmentPlanPaymentCtrl', function (
        $q,
        $timeout,
        $scope,
        CartService,
        PaymentService,
        UtilService,
        LogService,
        AlertService,
        UserService,
        ConfigService,
        InstallmentPlanService,
        $ionicLoading,
    ) {
		const currentPayment = PaymentService.currentPayment();
        const currentPlan = currentPayment.plan;
        const currentPaymentSeq = PaymentService.payments().length;

        const config = ConfigService.config;
        const cart = CartService.cart;        

        let authed = false;
        let stopAuth = false;

        $scope.footerMessage = "Solicitando Crédito...";

        async function auth() {
            try {
                let retryKey = null;

                const downPaymentValue = MathUtils.round(currentPlan.downPaymentValue);
                const totalValue = MathUtils.round(currentPlan.totalValue);

                let retryCount = 0;

                for (;;) {
                    if (stopAuth) {
                        break;
                    }

                    if (retryCount > 0) {
                        await updateFooterMessage(`Processando (${retryCount})`);
                    }

                    const res = await InstallmentPlanService.auth({
                        transactionKey: {
                            companyNumber: config.companyNumber,
                            storeNumber: config.storeNumber,
                            terminal: config.terminalNumber,
                            nsu: cart.nsu,
                            date: new Date(),
                        },
                        customer: cart.customer,
                        operatorCode: UserService.operatorCode(),
                        fiscalItems: cart.fiscalItems,
                        totalValue: totalValue,
                        downPaymentValue: downPaymentValue,
                        chargeValue: currentPlan.charges,
                        insuranceCode: currentPlan.insuranceCode,
                        insuranceTotalValue: currentPlan.insuranceTotalValue,
                        insurancePaymentValue: currentPlan.insurancePaymentValue,
                        insuranceDownPaymentValue: currentPlan.insuranceDownPaymentValue,
                        installments: currentPlan.instalments,
                        retryKey: retryKey,
                    });

                    if (stopAuth) {
                        break;
                    }

                    const resData = res.data;

                    switch (resData.paymentStatus) {
                    case 'AUTHORIZING':
                        retryKey = resData.retryKey;
                        retryCount++;
                        await updateFooterMessage(`Crédito em Análise, aguarde...`);
                        await $timeout(5000);
                        continue;
                    case 'AUTHORIZED':
                        authed = true;
                        $scope.back = () => {};

                        await updateFooterMessage(`Crédito Autorizado`);

                        const iofRate = findRate('IOF', 'MONTHS', resData.rates);

                        cart.additionalCharge = MathUtils.add(cart.additionalCharge, resData.chargeValue);

                        // poderíamos pegar o valor do plano diretamente, porém o PDV
                        // faz a façanha de não subir o pagamento com acréscimos e valor
                        // do seguro, logo fazemos este artifício técnico.
                        let paymentValue = currentPayment.paymentValue;
                        if (ConfigService.config.business.apply_charges_on_payment_value) {
                            paymentValue = MathUtils.add(paymentValue, currentPlan.charges);
                        }
                        if (currentPlan.discount) {
                            paymentValue = MathUtils.subtract(paymentValue, currentPlan.discount);
                            cart.discounts.push({
                                sequenceId: cart.discounts.length,
                                value: currentPlan.discount,
                                source: 'PLAN',
                            });
                        }

                        const payment = {
                            type:                       'SaleInstallmentPlanPayment',
                            paymentCodeMethod:          'INSTALLMENT_PLAN',
                            paymentSequence:            currentPaymentSeq,
                            paymentValue:               paymentValue,
                            paymentCodePlan:            currentPlan.code,
                            paymentCodeEvent:           currentPlan.paymentCodeEvent,
                            paymentPlanExternalCode:    currentPlan.externalCode,
                            paymentPlanProvider:        currentPlan.provider,
                            rates:                      resData.rates,
                            chargeValue:                resData.chargeValue,
                            iofValue:                   iofRate != null ? iofRate.value : 0.00,
                            downPaymentValue:           resData.downPaymentValue,
                            insuranceCode:              resData.insuranceCode,
                            insuranceDrawNumber:        resData.insuranceDrawNumber,
                            insurancePolicyNumber:      resData.insurancePolicyNumber,
                            insuranceTotalValue:        resData.insuranceTotalValue,
                            insurancePaymentValue:      resData.insurancePaymentValue,
                            insuranceDownPaymentValue:  resData.insuranceDownPaymentValue,
                            contractType:               resData.contractType,
                            contractNumber:             resData.contractNumber,
                            paymentDate:                new Date().toJSON(),
                        };

                        if (resData.installments && resData.installments.length > 0) {
                            payment.numberOfInstallments = resData.installments.length;
                            payment.dateTofirstInstalment =
                                resData.installments[resData.installments.length - 1].dueDate;
                            payment.installments = resData.installments.map(i => ({
                                paymentSequence:            i.number,
                                instalmentOriginalValue:    i.instalmentOriginalValue,
                                instalmentDateExpiration:   i.dueDate,
                                installmentValue:           i.instalmentValue,
                                installmentsNumber:         resData.installments.length,
                                insuranceValue:             i.insuranceValue,
                            }));
                        }

                        cart.salePayments.push(payment);
                        PaymentService.currentPaymentDone();

                        // esperando um tempinho pra trocar de tela
                        await $timeout(3000);

                        await $ionicLoading.show({
                            template : 'Por favor, aguarde...',
                            duration : 0
                        });

                        UtilService.navigateAfterReceivedPayment();
                        return;
                    default:
                        await updateFooterMessage('Status desconhecido');
                        AlertService.error(`Status do pagamento desconhecido: ${resData.paymentStatus}`);
                        return;
                    }
                }
            } catch(err) {
                LogService.error(err);
                await updateFooterMessage(err && err.data && err.data.message ? err.data.message : 'Falha durante autorização');
                await AlertService.restError(err, err && err.data && err.data.message ? err.data.message : 'Serviço indisponível');
                $scope.back();
            }
        }

        function findRate(type, chronoUnit, rates) {
            const rate = rates.find(r => r.type == type);

            if (!rate) {
                return 0;
            }

            return rate.values.find(v => v.chronoUnit == chronoUnit) || 0;
        }

        $scope.back = async () => {
            if (authed) {
                return;
            }

            // trocando a instância, para que não ocorrer multiplos cliques
            $scope.back = () => {};

            stopAuth = true;
            
            try {
                await CartService.undoCashbackDiscountPartial();
                await CartService.undoFirtsPurchaseDiscount();
                await undoPaymentCampaigns();
                PaymentService.clearInitializedPaymentData();
            } catch(err) {
                LogService.error(err);
            }

            goToPaymentMethods();
        };
        
        async function undoPaymentCampaigns() {
            const campaigns = PaymentService.currentPayment.campaigns;

            if (campaigns && campaigns.length) {
                campaigns.forEach(campaign => {
                    LogService.info(`Desfazendo desconto de campanha por pagamento: ${campaign.codeCampaign}`);
                    CartService.undoCampaignDiscountsPartial(campaign);
                });
            }
        };

        async function goToPaymentMethods() {
            if (CartService.hasUnprocessedPredefinedPaymentMethod()) {
                return UtilService.navigateWithoutHistory('easyorder.predefinedpaymentmethods');
            } else {
                return UtilService.navigateWithoutHistory('easyorder.paymentmethods');
            }
        }

        async function updateFooterMessage(message) {
            return $q((resolve) => {
                $scope.footerMessage = message;
                resolve();
            });
        }

        auth();
    });
})();
