(function () {
    'use strict';
    angular.module('easyorder.controllers.transactionproductlist', [
        'ionic',
        'ionic.closePopup',
    ]).controller('TransactionProductListCtrl', function (
        $scope,
        $ionicPopup,
        UtilService,
        AlertService,
        ProductService,
        IonicClosePopupService,
        $stateParams,
        ConfigService
    ) {

        $scope.op = $stateParams.op;
        $scope.sourceTransaction = $stateParams.sourceTransaction;
        $scope.productImageEnabled = ConfigService.config.business.product_images_enabled === true;
        $scope.totalValue = 0;
        $scope.selectedCount = 0;
        const isAnnulment = $scope.op === 'ANNULMENT';

        // No caso de anulação (ANNULMENT), não filtra por uuid e já seleciona todos os itens (pois não é parcial)
        $scope.displayItems = $scope.sourceTransaction.availableFiscalItems
            .filter(item => {
                return isAnnulment ? true : item.uuid != null;
            })
            .map(item => {
            return {
                item: {
                    sequenceId: item.sequenceId,
                    code: item.code,
                    automationCode: item.automationCode,
                    unitValue: item.unitValue,
                    totalValue: item.totalValue,
                    quantity: item.quantity,
                    // mantendo o description/compactDescription do legado
                    description: isAnnulment ? item.productDescription : item.description,
                    compactDescription: isAnnulment ? item.productDescription : item.compactDescription,
                    uuid: item.uuid,
                    lineItemTaxes : item.lineItemTaxes,
                    cfopNumber: item.cfopNumber,
                    taxCst    : item.taxCst,
                    pisCst    : item.pisCst,
                    cofinsCst : item.cofinsCst,
                },
                selected: isAnnulment ? true : false,
                quantity: isAnnulment ? item.quantity: 0
            };
        });
        // atualiza os totais
        if (isAnnulment) {
            update();
        }


        $scope.goBack = () => {
            UtilService.navigateWithoutHistory('easyorder.transactionsearch', {
                op: $scope.op,
                authorizer: $stateParams.authorizer,
            });
        };

        $scope.goNext = () => {
            if ($scope.displayItems.some(displayItem => displayItem.selected)) {
                if ($scope.op == 'RETURN') {
                    const srcTrns = $scope.sourceTransaction;
                    const nonAllowedPmnt = srcTrns.payments.some(
                        pmnt => !isPaymentAllowed(pmnt)
                    );
                    if (nonAllowedPmnt) {
                        return AlertService.error(
                            'Só é possível cancelamento de Crédito, Débito, QR Linx e Link de Pagamentos, favor efetuar operação no PDV'
                        );
                    }

                    const rawSrcTrnsDate = srcTrns.key.date;
                    const timeIdx = rawSrcTrnsDate.indexOf('T');
                    const srcTrnsDate = timeIdx != -1 ? moment(rawSrcTrnsDate.substring(0, timeIdx))
                                                      : moment(rawSrcTrnsDate);

                    if (srcTrnsDate.isSame(moment(), 'day') &&
                        isPartialReturn(srcTrns, $scope.displayItems.filter(displayItem => displayItem.selected))) {
                        return AlertService.error(`
                            Não é possível realizar estorno parcial para para pagamento em Cartão, QR Linx e Link de Pagamentos na data atual
                        `);
                    }
                }

                UtilService.navigateWithoutHistory('easyorder.transactionjustification', {
                    op: $scope.op,
                    authorizer: $stateParams.authorizer,
                    data: parseReturnData(),
                });
            } else {
                AlertService.alert("Erro", "Escolha pelo menos um produto da lista.");
            }
        };

        function isPaymentAllowed(pmnt) {
            return eftFilter(pmnt)
                || qrLinxFilter(pmnt)
                || paymentLinkFilter(pmnt);
        }

        function eftFilter(payment) {
            const eftPaymentMethods = ['CREDIT', 'DEBIT'];
            return eftPaymentMethods.includes(payment.paymentMethod)
                && !isLinkProvider(payment.provider)
        }

        function qrLinxFilter(payment) {
            return payment.paymentMethod == 'QR_LINX';
        }

        function paymentLinkFilter(payment) {
            return payment.paymentMethod == 'PAYMENT_LINK'
                || isLinkProvider(payment.provider);
        }

        function isLinkProvider(provider) {
            if (provider == null) {
                return false;
            }

            const linkProviders = [
                'PAGAR_ME',
                'LINX_PAYHUB_LINK',
                'E_SITEF_WEB_CHECKOUT'
            ];

            return linkProviders.includes(provider.id);
        }

        $scope.selectProduct = (displayItem, select) => {
            if (!select) {
                displayItem.quantity = 0;
                update();
                return;
            }

            if (displayItem.quantity > 1) {
                showQuantityPopup(displayItem)
                .then(quantity => {
                    if (quantity) {
                        displayItem.quantity = quantity;
                    } else {
                        displayItem.selected = false;
                    }
                });
            } else {
                displayItem.quantity = 1;
            }
            update();
        };

        function update() {
            $scope.selectedCount = $scope.displayItems.filter(displayItem => displayItem.selected).length;

            if (isAnnulment) {
                // se for anulação, usa o total da transação para evitar problemas com descontos e arredondamentos
                $scope.totalValue = $scope.sourceTransaction.totalValue;
            } else {
                $scope.totalValue = $scope.displayItems.filter(displayItem => displayItem.selected)
                    .map(displayItem => displayItem.item.unitValue * displayItem.item.quantity)
                    .reduce((acc, val) => acc + val, 0);
            }

        }

        function parseReturnData() {
            return {
                sourceTransaction: $scope.sourceTransaction,
                fiscalItems: $scope.displayItems.filter(item => item.selected)
                    .map(displayItem => {
                        const item = displayItem.item;
                        const quantity = displayItem.quantity;
                        return {
                            sequenceId: item.sequenceId,
                            code: item.code,
                            automationCode: item.automationCode,
                            marketingCode: item.marketingCode,
                            uuid: item.uuid,
                            unitValue: item.unitValue,
                            totalValue: item.unitValue * quantity,
                            quantity: quantity,
                            description: item.description,
                            compactDescription: item.compactDescription,
                            lineItemTaxes : item.lineItemTaxes,
                            cfopNumber: item.cfopNumber,
                            taxCst    : item.taxCst,
                            pisCst    : item.pisCst,
                            cofinsCst : item.cofinsCst,
                        };
                    }),
            };
        };

        function showQuantityPopup(item) {
            $scope.quantityPopupData = {
                item: item,
                quantity: item.quantitySold,
                max: item.quantitySold,
                min: 1,
            };

            $scope.decrement = () => {
                const min = $scope.quantityPopupData.min;
                const quantity = --$scope.quantityPopupData.quantity;
                $scope.quantityPopupData.quantity = Math.max(min, quantity);
            };

            $scope.increment = () => {
                const max = $scope.quantityPopupData.min;
                const quantity = ++$scope.quantityPopupData.quantity;
                $scope.quantityPopupData.quantity = Math.min(max, quantity);
            };

            $scope.clamp = () => {
                const min = $scope.quantityPopupData.min;
                const max = $scope.quantityPopupData.min;
                $scope.quantityPopupData.quantity = Math.min(Math.max(parseInt(quantity), min), max);
            };

            const popup = $ionicPopup.show({
                template: `
                <div class="modal-popup">
                    <div class="count-content">
                        <div class="count">
                            <button ng-class="{true: 'minus', false: ''}[(quantityPopupData.quantity > quantityPopupData.min)]"
                            ng-click="decrement()" class="button-count">
                                <i class="fa fa-minus"></i>
                            </button>
                            <input ng-model="quantityInput.count" ng-change="clamp()" type="tel" class="count-value" disabled>
                            <button class="button-count plus" ng-click="increment()">
                                <i class="fa fa-plus"></i>
                            </button>
                        </div>
                    </div>
                </div>
                `,
                title: 'Indique a quantidade de produtos de entrada.',
                scope: $scope,
                cssClass: 'popupCount',
                buttons: [{
                    text: '<b>Salvar</b>',
                    type: 'button-theme',
                    onTap: function (e) {
                        return $scope.quantityPopupData.quantity;
                    }
                }]
            });
            IonicClosePopupService.register(popup);

            return popup;
        };

        $scope.imageUrl = function(item) {
            return ProductService.getProductImageUrl({
                productCode: item.code,
                automationCode: item.automationCode,
            });
        };


        function isPartialReturn(srcTrns, returnedFiscalItems) {
            const availableCount = srcTrns.availableFiscalItems.filter(item => item.uuid).reduce(quantityAccumulator, 0);
            const returnedCount = returnedFiscalItems.reduce(quantityAccumulator, 0);

            if (availableCount > returnedCount) {
                return true;
            }

            return false;
        }

        function quantityAccumulator(acc, item) {
            return acc + new Number(item.quantity);
        }

        const _MS_PER_DAY = 1000 * 60 * 60 * 24;
        function dateDiffInDays(a, b) {
            // Discard the time and time-zone information.
            const utc1 = Date.UTC(a.getFullYear(), a.getMonth(), a.getDate());
            const utc2 = Date.UTC(b.getFullYear(), b.getMonth(), b.getDate());

            return Math.floor((utc2 - utc1) / _MS_PER_DAY);
        }
    })
})();
