(function() {
	'use strict';

	angular.module('easyorder.services.eft', [
        'easyorder.services.config',
        'easyorder.services.log'
    ])

	.factory('EFTService', function($ionicPopup, $q, $injector, ConfigService, AlertService, CartService, LogService, $rootScope) {
        const self = this;

        self.transactionType = EFTServiceBase.transactionType;

        self.activate = (req) => {
            return self.handler.activate(req);
        };

        self.setActivation = (activation) => {
            self.handler.setActivation(activation);
        };

        self.getActivation = () => {
            return self.handler.getActivation();
        }

        self.executeTransaction = (funcParameters, callbackResolver) => {
            self.handler.executeTransaction(funcParameters, callbackResolver);
        }

        self.undoTransaction = transactionNSU => {
            return self.handler.undoTransaction(transactionNSU);
        }

        self.confirmTransaction = transactionNSU => {
            return self.handler.confirmTransaction(transactionNSU);
        }

        self.finalizeTransaction = () => {
            return self.handler.finalizeTransaction();
        }

        self.isDTEF = () => ConfigService.config.tefserver.provider
            && ConfigService.config.tefserver.provider == 'Linx';

        self.init = () => {
            if (ConfigService.device.lio) {
                self.handler = new LioEFTService(
                    $q,
                    ConfigService,
                    $injector.get('PaymentService'),
                    AlertService,
                    CartService,
                    LogService
                );
            } else {
                self.handler = new DefaultEFTService(
                    $q,
                    ConfigService,
                    $ionicPopup,
                    LogService,
                    $injector.get('PaymentProcessor'),
                    CartService,
                    $rootScope.dataSourceManager
                );
            }
        }
        return self;
    });
})();

class EFTServiceBase {
    constructor($q, ConfigService, LogService) {
        this.activation = null;
        this.$q = $q;
        this.LogService = LogService;
        this.ConfigService = ConfigService;
        this.transactionType = EFTServiceBase.transactionType;
    }

    executeTransaction(funcParameters, callbackResolver) {
        throw new Error('executeTransaction unimplemented');
    }

    undoTransaction(transactionNSU) {
        throw new Error('undoTransaction unimplemented');
    }

    confirmTransaction(transactionNSU) {
        throw new Error('confirmTransaction unimplemented');
    }

    activate(req) {
        return this.$q.resolve();
    }

    finalizeTransaction() {
        // fazendo nada até descobrir se CTF suporta esta chamada
        return this.$q.resolve();
    }

    setActivation(activation) {
        this.activation = activation;
    }

    getActivation() {
        return this.activation;
    }

    isActivationDirty() {
        const activation = this.getActivation();

        if (activation == null) {
            return true;
        }
        
        const storexTerminal = activation.storexTerminal;
        const eftTerminal = activation.eftTerminal;

        const config = this.ConfigService.config;
        const tefserver = config.tefserver;

        if (activation.deviceUuid != device.uuid) {
            return true;
        }

        if (activation.provider != tefserver.provider) {
            return true;
        }

        // comparando dados do storex da ativação realizada
        if (!this.numericEquals(storexTerminal.companyNumber, config.companyNumber)) {
            return true;
        }
        if (!this.numericEquals(storexTerminal.storeNumber, config.storeNumber)) {
            return true;
        }
        if (!this.numericEquals(storexTerminal.terminalNumber, config.terminalNumber)) {
            return true;
        }
        if (!this.numericEquals(storexTerminal.storeDocument, config.store.cnpj)) {
            return true;
        }

        const eftCompanyNumber = tefserver.establishmentCode;
        const eftStoreNumber = this.nonZeroFb(tefserver.storeNumber, config.eftStoreNumber);
        const eftTerminalNumber = config.terminalNumber;
        const eftStoreDocument = this.nonZeroFb(tefserver.storeCnpj, config.store.cnpj);

        if (this.nonZero(eftCompanyNumber) && !this.numericEquals(eftTerminal.companyNumber, eftCompanyNumber)) {
            return true;
        }
        if (this.nonZero(eftStoreNumber) && !this.numericEquals(eftTerminal.storeNumber, eftStoreNumber)) {
            return true;
        }
        if (!this.numericEquals(eftTerminal.terminalNumber, eftTerminalNumber)) {
            return true;
        }
        if (this.nonZero(eftStoreDocument) && !this.numericEquals(eftTerminal.storeDocument, eftStoreDocument)) {
            return true;
        }

        return false;
    }

    numericEquals(num1, num2) {
        return parseInt(num1) == parseInt(num2);
    }

    nonZero(num) {
        if (num != null) {
            const numType = (typeof num);
            if (numType == "string") {
                return num.length > 0 && parseInt(num) > 0;
            } else if (numType == "number" || num instanceof Number) {
                return num > 0;
            }
        }
        return false;
    }

    nonZeroFb(num, fallback) {
        return this.nonZero(num) ? num : fallback;
    }
}

EFTServiceBase.transactionType = {
    CREDIT: 1,
    DEBIT: 2,
    CREDIT_WITH_INSTALMENTS: 3,
    MOBILE_RECHARGE: 13,
    TRAIL_READ: 38,
    RENNER_CBR: 39,
    CHECK_PHONE_NUMBER: 131,
    DEBIT_INVOICE: 216,
    CONDUCTOR_CREDIT_BOOK: 391,
    CONDUCTOR_CARD_STATEMENT: 392,
    CANCEL: 4,
    /* 
     *
     * Apenas para D-TEF, número do terminal utilizado será retornado de forma dinâmica
     * pelo servidor do D-TEF.
     *
     * */
    ACTIVATE_TERMINAL: 17,
    /* 
     *
     * Apenas para D-TEF, diferente da 17 esta utilizará o número do terminal fornecido
     * para a ativação.
     *
     * */
    ACTIVATE_PROVIDED_TERMINAL: 100,
	SITEF_QR: 40,
    RENNER_BILLING_PAYMENT: 41
}

class DefaultEFTService extends EFTServiceBase {
    constructor($q, ConfigService, $ionicPopup, LogService, PaymentProcessor, CartService, DataSourceProviderManager) {
        super($q, ConfigService, LogService);
        this.$ionicPopup = $ionicPopup;
        this.PaymentProcessor = PaymentProcessor;
        this.processing = false;
        this.cancel = false;
        this.waitingCallbackResponse = false;
        this.cancelDeferred = $q.resolve();
        this.autoConfirmation = false;
        this.CartService = CartService;
        this.dataSourceProviderManager = DataSourceProviderManager;
    }

    executeTransaction(funcParameters, callbackResolver) {
        if (this.provider == null) {
            callbackResolver.paymentError('Nenhum provedor TEF configurado');
            return;
        }
        if (!this.isProviderSupported()) {
            callbackResolver.paymentError('Provedor TEF não suportado para esta versão do plugin');
            return;
        }
        if (this.processing) {
            callbackResolver.paymentError('Uma transação já está em processamento, por favor aguarde.');
            return;
        }
        if (this.cancel) {
            callbackResolver.paymentError('Uma operação no pinpad está em cancelamento, por favor aguarde.');
            return;
        }
        if (this.PaymentProcessor.containsPendingEFT()) {
            callbackResolver.paymentError('Este terminal possui transação TEF pendente, por favor aguarde.');
            return;
        }
        if (!funcParameters) {
            callbackResolver.paymentError('Parâmetros inválidos.');
            return;
        }
        if (!funcParameters.transactionType) {
            callbackResolver.paymentError('Tipo de transação não informado.');
            return;
        }
        if (!this.ConfigService.config.tefserver || !this.ConfigService.config.tefserver.host) {
            callbackResolver.paymentError('Parametrização de Provedor TEF não está completa. (Provedor/IP)');
            return;
        } else {
            // se for usado um endpoint https, não é necessário parametrizar a porta
            if (!this.ConfigService.config.tefserver.port && !this.ConfigService.config.tefserver.host.startsWith("https://")) {
                callbackResolver.paymentError('Parametrização de Provedor TEF não está completa. (Porta)');
                return;
            }
        }

        this.processing = true;

        var parameters = "";
        
        if (funcParameters.amount) {
            if (funcParameters.amount.toFixed) {
                funcParameters.amount = funcParameters.amount.toFixed(2);
            }
            funcParameters.amount = funcParameters.amount.replace('.', '');
            funcParameters.amount = funcParameters.amount.replace(',', '');
        }
        var codigoTransacao = "";

        switch (funcParameters.transactionType) {
            case this.transactionType.CREDIT:
            case this.transactionType.CREDIT_WITH_INSTALMENTS:
                codigoTransacao = this.transactionType.CREDIT;
                break;
            case this.transactionType.CONDUCTOR_CREDIT_BOOK:
            case this.transactionType.CONDUCTOR_CARD_STATEMENT:
                codigoTransacao = this.transactionType.RENNER_CBR;
                break;
            case this.transactionType.CANCEL:
                codigoTransacao = 3;
                break;
            default:
                codigoTransacao = funcParameters.transactionType;
                break;                    
        }

        parameters = this.defaultParameters(funcParameters);
        parameters += "CodigoTransacao=" + codigoTransacao + ";";
        parameters += "ValorTransacao=" + funcParameters.amount + ";";
        parameters += "CodigoOperador=" + funcParameters.operatorCode + ";";

        if (funcParameters.transactionType == this.transactionType.CANCEL
             && funcParameters.originTransactionType) {
            parameters += (
                "TipoTransacaoCancelamento="
                + (funcParameters.originTransactionType == this.transactionType.DEBIT ? 2 : 1)
            ) + ";";
        }

        if (funcParameters.DDD && funcParameters.phoneNumber) {
            parameters += "DDDTelefone=" + funcParameters.DDD + ";";
            parameters += "NumeroTelefone=" + funcParameters.phoneNumber + ";";
            parameters += "CodigoOperadora=" + funcParameters.mobileOperatorCode + ";";
        }
            
        if (funcParameters.instalmentsNumber) {
            parameters += "NumeroParcelas=" + funcParameters.instalmentsNumber + ";"
    
            if (funcParameters.transactionType === this.transactionType.DEBIT ||
                funcParameters.transactionType === this.transactionType.DEBIT_INVOICE) {
                parameters += "TipoFinanciamento=0;";
            } else if (funcParameters.transactionType === this.transactionType.CREDIT || 
                       funcParameters.transactionType === this.transactionType.CREDIT_WITH_INSTALMENTS ||
                       funcParameters.transactionType === this.transactionType.CONDUCTOR_CREDIT_BOOK ||
                       funcParameters.transactionType === this.transactionType.CONDUCTOR_CARD_STATEMENT || 
					   funcParameters.transactionType === this.transactionType.SITEF_QR || 	
                       funcParameters.transactionType === this.transactionType.TRAIL_READ) {
                parameters += "TipoFinanciamento=" + (funcParameters.instalmentsNumber <= 1 ? "0;" : 
                    (funcParameters.interestType && funcParameters.interestType === 'AUTHORIZER' ? "1;" : "2;")); // "1" Financiamento Administradora  // Financiamento Loja
            }
        } else {
            if (funcParameters.transactionType === this.transactionType.DEBIT ||
                funcParameters.transactionType === this.transactionType.DEBIT_INVOICE) {
                parameters += "TipoFinanciamento=0;";
            } else if (funcParameters.transactionType === this.transactionType.CREDIT ) {
                parameters += "TipoFinanciamento=0;"; // "0" À vista
            } else if (funcParameters.transactionType === this.transactionType.CREDIT_WITH_INSTALMENTS ) {
                parameters += "TipoFinanciamento=2;"; // Financiamento Loja
            }
        }
        
        if (funcParameters.transactionType === this.transactionType.CONDUCTOR_CREDIT_BOOK ||
            funcParameters.transactionType === this.transactionType.CONDUCTOR_CARD_STATEMENT) {
            
        	if (funcParameters.transactionType === this.transactionType.CONDUCTOR_CREDIT_BOOK) {
                parameters += 'TipoTransacaoRenner=1;'; //Carnê
            } else {
                parameters += 'TipoTransacaoRenner=2;'; //Fatura
            }
            
            if (funcParameters.hasInsurance) {
                parameters += 'ContrataSeguro=S;';
            } else {
                parameters += 'ContrataSeguro=N;';
            }
            
            if (funcParameters.charges && funcParameters.charges > 0) {
                parameters += "TipoParcelamento=3;";
            } else {
                parameters += "TipoParcelamento=4;";
            }
            if (funcParameters.processorCode && funcParameters.processorCode != 0) {
                parameters += `CodigoProcessadora=${funcParameters.processorCode};`;
            }
        }

        if (funcParameters.transactionDate) {
            parameters += `DataTransacao=${funcParameters.transactionDate};`;
        }

        if (funcParameters.transactionNSU) {
            parameters += `NSUTransacao=${funcParameters.transactionNSU.trim()};`;
            if (funcParameters.transactionType === this.transactionType.CANCEL ) {
                parameters += `NumeroControle=${funcParameters.transactionNSU.trim()};`;
            }
        }
        
        if (funcParameters.returnQRCodeString) {
            parameters += "DevolveStringQRCode=1;";
        }
        
        if (funcParameters.customerDocument) {
            parameters += `CPFColetado=${funcParameters.customerDocument.replace(/\D/g, "")};`
        }
        
        if (funcParameters.listNotAcceptedVouchers && funcParameters.listNotAcceptedVouchers.length) {
            parameters += `ListaVouchersNaoAceitos=${funcParameters.listNotAcceptedVouchers.join(",")};`
        }

        if (funcParameters.typedCardNumber) {
            // mantendo o label 'NumeroCartaDigtado' errado pelo legado, a remover posteriormente
            parameters += `NumeroCartaDigtado=${funcParameters.typedCardNumber};NumeroCartaoDigitado=${funcParameters.typedCardNumber};`;
        }

        if (funcParameters.cardExpirationDate) {
            parameters += `DataVencimentoCartao=${funcParameters.cardExpirationDate};`;
        }

        if (funcParameters.cardProcessor) {
            parameters += `RedeDestino=${funcParameters.cardProcessor};`;

            // para não afetar o legado, só enviamos o plano de pagamento externo caso seja conductor
            if (funcParameters.cardProcessor === 'CONDUCTOR' && funcParameters.paymentPlanCode) {
                parameters += `CodigoPlanoPagamento=${funcParameters.paymentPlanCode};`;
            }
        }

        if (typeof funcParameters.activationOpCode == 'number') {
            parameters += `OperacaoSerialPDV=${funcParameters.activationOpCode}`;
        }

        this.LogService.info("executeTransaction :: parameters = ", parameters);
        
        window.dtefMobile.executeTransaction(
            results => {     
                if (results.action === "InputCallback") {
                    this.printLog(`[TEF Callback] Action: ${results.action} - Mensagem: ${results.mensagem} - Label: ${results.label}`);
                    this.inputCallback(results, callbackResolver);
                } else {
                    this.processing = false;

                    /* 
                     * caso tenha sido requisitado cancelar operação e tenha chegado neste ponto
                     * devemos rejeitar a tentativa de cancelamento de operação
                     */
                    if (this.cancel == true) {
                        this.cancel = false;
                        this.cancelDeferred.reject();
                    }
                    callbackResolver.paymentDone(results);
                }
            }, error => {
                this.processing = false;
                this.printLog('D-TEF Error: ', JSON.stringify(error));
                if (this.cancel == false) {
                    callbackResolver.paymentError(error.MensagemErro, error.CodigoRetornoFuncao);
                } else {
                    /* o plugin retorna o cancelamento da operação como erro */
                    this.cancel = false;
                    this.cancelDeferred.resolve();
                }
            }, { parameters: parameters }
        );
    }
    
    printLog(label, value){
        this.LogService.info(label, value);
    }

    defaultParameters(args) {
        args = args == null ? {} : args;
        const config = this.ConfigService.config;

        const params = {
            CupomFiscal: ("000000" + this.ConfigService.nsuControll.currentCartNSU()).slice(-6),
            /* 000 - desabilitado, 111 - habilitado. Utilizado no D-TEF, caso habilitado irá desfazer a transação anterior automaticamente */
            TratarDesfazimento: "000",
            Batch: !this.isDTEF(), /* este parametro foi criado para o D-TEF */
            TipoServidorTEF: this.tefServerCode,
            TipoInterface: '1',
            MensagemDisplayPinPad: 'LINX',
            PinPadAtivo: args.usePinpad == null || args.usePinpad ? '1' : '0',
            // Fixa o tamanho do número no formato 'NNxxxxxx'.
            NumeroTerminalSiTef: "SE" + ("000000" + this.terminalNumber).slice(-6),
            // no plugin CNPJAutomacao apesar do nome é o cnpj da loja
            CNPJAutomacao: this.isDTEF() ? this.storeCNPJ : "34016602040614",
            // Não sabemos se este campo é oficial
            CNPJESTAB: this.storeCNPJ,
            // Não sabemos se nossa versão do plugin utiliza este campo,
            // porém está especificado no documento de versão 5.9
            CNPJSoftwareHouse: "34016602040614",
            NUMPDV: this.terminalNumber,
            NUMESTAB: this.isDTEF() ? this.toLength(this.companyNumber, 4, "0") : this.companyNumber,
            // Verificar se este campo foi criado por nós
            NUMESTABTEF: this.ConfigService.config.tefserver.establishmentCode || this.companyNumber,
            // Fixa o tamanho do número no formato 'xxxxxxxx'.
            CodigoEmpresaSiTef: ("00000000" + this.storeNumber).slice(-8),
            NUMLOJA: this.storeNumber,
            OperadorasTelefonia: this.mobileRechargeOperators,
            ComunicacaoSSL: config.tefserver.secure ? "1" : "0",
            // note that we don't use terminalNumber getter here to avoid dups
            IDTerminal: `${config.storeNumber}${config.terminalNumber}`,
            SerialDispositivo: device.uuid
        };

        if (config.tefserver.certificatePassword != null && config.tefserver.certificatePassword.length > 0) {
            params.SenhaCertificado = config.tefserver.certificatePassword;
        }

        // configura a comunicação via HTTPS ou TCP/IP
        if (this.ConfigService.config.tefserver.host.startsWith('https')) {
            params.HTTPServidor = this.ConfigService.config.tefserver.host;
            params.SenhaHTTPServidor = this.ConfigService.config.tefserver.httpsPassword;
            params.SenhaHTTPHexa = this.ConfigService.config.tefserver.httpsPasswordType == "HEXADECIMAL" ? "1" : "0";

            // O plugin do DTEF tem um bug que não remove alguns parâmetros enviados de sua persistência.
            // Por isso é necessário enviar no outro formato esperado (IPServidor com a URL e a PortaServidor contando a senha)
            // As linhas abaixo poderão ser removidas após correção no plugin.
            params.IPServidor = this.ConfigService.config.tefserver.host;
            params.PortaServidor = this.ConfigService.config.tefserver.httpsPassword;
        } else {
            params.IPServidor = this.ConfigService.config.tefserver.host;
            params.IPServidorSiTef = this.ConfigService.config.tefserver.host;
            params.PortaServidorSiTef = this.ConfigService.config.tefserver.port;
            params.PortaServidor = this.ConfigService.config.tefserver.port;
            params.IPServidorSecundario = this.ConfigService.config.tefserver.secondaryHost;
            params.PortaServidorSecundario = this.ConfigService.config.tefserver.secondaryport;
        }

        if (config.tefserver.certificateActivationKey != null) {
            params.ChaveAtivacaoCertificado = config.tefserver.certificateActivationKey;
        }

        if (config.tefserver.sslProvider) {
            params.ProvedorSSL = config.tefserver.sslProvider;
            // o SiTef usa os certificados da Mozilla pro SSL. Se for usado posteriormente outra lista, será adicionado um parâmetro. 
            if (config.tefserver.sslProvider === 'SOFTWARE_EXPRESS') {
                params.CaminhoCertificado = this.dataSourceProviderManager.getCertificateAbsolutePath(MOZILLA_PEM_FILE);
            }            
        }

        return this.parseParameters(params);
    }

    parseParameters(parameters) {
        let str = '';
        for (const key in parameters) {
            str = `${str}${key}=${parameters[key]};`;
        }
        return str;
    }

    isDTEF() {
        return this.ConfigService.config.tefserver.provider &&
            this.ConfigService.config.tefserver.provider == 'Linx';
    }

    get tefServerCode() {
        switch (this.ConfigService.config.tefserver.provider) {
            case 'Auttar':
                return '3';
            case 'Linx':
                return '0';
            case 'SoftwareExpress':
                return '1';
            default:
                return null;
        }
    }

    get provider() {
        return this.ConfigService.config.tefserver.provider;
    }

    isProviderSupported() {
        return window.dtefMobile.isProviderSupported(this.provider);
    }

    get terminalNumber() {
        const activation = this.getActivation();

        if (activation != null) {
            return activation.eftTerminal.terminalNumber;
        }

        return this.ConfigService.config.terminalNumber;
    }

    get storeCNPJ() {
        const activation = this.getActivation();

        if (activation != null) {
            return activation.eftTerminal.storeDocument;
        }

        return this.nonZero(this.ConfigService.config.tefserver.storeCnpj)
                ? this.ConfigService.config.tefserver.storeCnpj
                : this.ConfigService.config.store.cnpj;
    }

    get companyNumber() {
        const activation = this.getActivation();

        if (activation != null) {
            return activation.eftTerminal.companyNumber;
        }

        return this.nonZero(this.ConfigService.config.tefserver.establishmentCode)
                ? this.ConfigService.config.tefserver.establishmentCode
                : this.ConfigService.config.companyNumber;
    }

    toLength(str, len, pad) {
        if (str == null) {
            return "".padStart(len, pad);
        } else if (str.length > len) {
            return str.substring(str.length - len, str.length);
        } else if (str.length < len) {
            return str.padStart(len, pad);
        } else {
            return str;
        }
    }

    notEmpty(num) {
        if (num != null) {
            const numType = (typeof num);
            if (numType == "string") {
                return num.length != 0;
            } else if (numType == "number" || num instanceof Number) {
                return true;
            }
        }
        return false;
    }
 
    /* priority: config-server > file > store */
    get storeNumber() {
        const activation = this.getActivation();

        if (activation != null) {
            return activation.eftTerminal.storeNumber;
        }

        if (this.notEmpty(this.ConfigService.config.tefserver.storeNumber)) {
            return this.ConfigService.config.tefserver.storeNumber;
        }
        return this.nonZero(this.ConfigService.config.eftStoreNumber)
                ? this.ConfigService.config.eftStoreNumber
                : this.ConfigService.config.storeNumber;
    }

    get mobileRechargeOperators() {
        return this.ConfigService.config.business.mobile_recharge_operators &&
            this.ConfigService.config.business.mobile_recharge_operators.length > 3 ?
                this.ConfigService.config.business.mobile_recharge_operators : 'TIM#LD'
    }

    undoTransaction(transactionNSU) {
        if (this.provider == null) {
            return this.$q.reject('Nenhum provedor TEF configurado');
        }
        if (!this.isProviderSupported()) {
            return this.$q.reject('Provedor TEF não suportado para esta versão do plugin');
        }
        if (this.processing) {
            return this.$q.reject('Uma transação já está em processamento, por favor aguarde.');
        }
        if (this.cancel) {
            return this.$q.reject('Uma operação no pinpad está em cancelamento, por favor aguarde.');
        }
        if (transactionNSU == null && this.tefServerCode != '1') {
            const msg = 'Apenas sitef suporta desfazimento sem nsu de pagamento ignorando undoTransaction'
            this.printLog(msg);
            return this.$q.resolve(msg);;
        }
        if (this.ConfigService.nsuControll.currentCartNSU() == null) {
            var msg = 'Não há registros de transações anteriores no terminal, ignorando undoTransaction';
            this.printLog(msg);
            return this.$q.resolve(msg);
        }
        if (transactionNSU == null && this.tefServerCode != '1') {
            var _msg = 'Apenas sitef suporta desfazimento sem nsu de pagamento ignorando undoTransaction';
            this.printLog(_msg);
            return this.$q.resolve(_msg);
        }
        this.processing = true;
        return this.$q((resolve, _) => {
            window.dtefMobile.undoTransaction(results => {
                this.printLog('undoTransaction - results.action = ' + results.action + ' function = ' + results.function);
                if (results.action === "InputCallback") {
                    this.inputCallback(results);
                } else {
                    this.processing = false;
                    this.printLog(`Transação TEF desfeita NSUTransacao: ${transactionNSU}, CUPOM: ${this.ConfigService.nsuControll.currentCartNSU()}`);
                    resolve();
                }
            }, error => {
                this.processing = false;
                this.printLog('Erro no desfazimento da transação TEF: ', JSON.stringify(error));
                resolve();
            },
            {
                parameters: (transactionNSU != null ? `NSUTransacao=${transactionNSU};` : '') + this.defaultParameters()
            });
        });
    }

    confirmTransaction(transactionNSU) {
        if (this.provider == null) {
            return this.$q.reject('Nenhum provedor TEF configurado');
        }
        if (!this.isProviderSupported()) {
            return this.$q.reject('Provedor TEF não suportado para esta versão do plugin');
        }
        if (this.processing) {
            return this.$q.reject('Uma transação já está em processamento, por favor aguarde.');
        }
        if (this.cancel) {
            return this.$q.reject('Uma operação no pinpad está em cancelamento, por favor aguarde.');
        }
        this.processing = true;
        return this.$q((resolve, reject) => {
            var parameters = "NSUTransacao=" + transactionNSU + ";";
            parameters = parameters + this.defaultParameters();

            window.dtefMobile.confirmTransaction(results => {
                this.printLog('confirmTransaction - results.action = ' + results.action + ' function = ' + results.function);
                if (results.action === "InputCallback") {
                    this.inputCallback(results);
                } else {
                    this.processing = false;
                    this.printLog('Transação TEF confirmada NSU: ', transactionNSU);
                    resolve('Transação TEF confirmada NSU: ', transactionNSU);
                }
            }, error => {
                this.processing = false;
                this.printLog('Erro na confirmação da transação TEF: ', JSON.stringify(error));
                reject('Erro na confirmação da transação TEF');
            }, { parameters: parameters });
        });
    }

    finalizeTransaction() {
        if (!ionic.Platform.isAndroid() || this.provider == "Auttar") {
            return this.$q.resolve();
        }
        if (this.provider == null) {
            return this.$q.reject('Nenhum provedor TEF configurado');
        }
        if (!this.isProviderSupported()) {
            return this.$q.reject('Provedor TEF não suportado para esta versão do plugin');
        }
        if (!this.processing && !this.cancel) {
            return this.$q.resolve();
        }
        this.cancelDeferred = this.$q.defer();
        this.cancel = true;
        
        if (this.waitingCallbackResponse) {
            window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": "0", "intRetorno": "0", "stringRetorno": "" } });
        }
        return this.cancelDeferred.promise;
    }

    inputCallback(results, callbackResolver) {
        var res;

        var sizeRange = this.resultsInputSizeRange(results);
        this.waitingCallbackResponse = false;

        this.printLog("Funcao inputcallback: ", results.function);
    
        if (callbackResolver) {
            if (results.function === "beep") {
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "limpaDisplayTerminal") {
                callbackResolver.clearDisplay();
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "displayTerminal") {
                callbackResolver.displayMessage(results.mensagem);
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "displayErro") {
                callbackResolver.displayMessage(results.mensagem);
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "mensagem") {
                callbackResolver.displayMessage(results.mensagem);
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "solicitaConfirmacao") {
                // estará true quando passar em "entraCartao"
                if (this.autoConfirmation) {
                    window.dtefMobile.sendCallbackResponse(null, null,
                        { parameters: { "resultado": "0", "intRetorno": "0", "stringRetorno": "0" } });
                    this.autoConfirmation = false;
                } else {
                    if (callbackResolver.confirmAction) {
                        callbackResolver.confirmAction(
                            results.mensagem
                        ).then((res) => {
                            if (res) {
                                window.dtefMobile.sendCallbackResponse(null, null,
                                    { parameters: { "resultado": "0", "intRetorno": "0", "stringRetorno": "0" } });
                            } else {
                                window.dtefMobile.sendCallbackResponse(null, null,
                                    { parameters: { "resultado": "1", "intRetorno": "1", "stringRetorno": "1" } });
                            }
                        });
                    } else {
                        this.$ionicPopup.confirm({
                            title: 'Transação TEF',
                            template: results.mensagem,
                            okType: "button-theme"
                        }).then(res => {
                            if (res) {
                                window.dtefMobile.sendCallbackResponse(null, null,
                                    { parameters: { "resultado": "0", "intRetorno": "0", "stringRetorno": "0" } });
                            } else {
                                window.dtefMobile.sendCallbackResponse(null, null,
                                    { parameters: { "resultado": "1", "intRetorno": "1", "stringRetorno": "1" } });
                            }
                        });
                    }
                }
            } else if (results.function  === "mensagemAlerta") {
                if (callbackResolver.alertMessage) {
                    callbackResolver.alertMessage(
                        results.mensagem
                    ).then(() => {
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": "0", "intRetorno": "0", "stringRetorno": "" } });
                    });
                } else {
                    this.$ionicPopup.alert({
                        title: 'Transação TEF',
                        template: results.mensagem,
                        okType: "button-theme"
                    }).then(() => {
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": "0", "intRetorno": "0", "stringRetorno": "" } });
                    });
                }
            } else if (results.function  === "entraCartao") {
                window.dtefMobile.sendCallbackResponse(null, null,
                    { parameters: { "resultado": -1, "intRetorno": -1, "stringRetorno": '-1' } });
                this.autoConfirmation = true;
            } else if (results.function  === "entraDataValidade") {
                this.waitingCallbackResponse = true;
                
                callbackResolver.numberInput(
                        results.label,
                        number => {
                            window.dtefMobile.sendCallbackResponse(null, null,
                                { parameters: { "resultado": 0, "intRetorno": 0, "stringRetorno": number } });
                        }
                );
            } else if (results.function  === "entraData") {
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function  === "entraCodigoSeguranca") {
                this.waitingCallbackResponse = true;
                
                callbackResolver.numberInput(
                    results.label,
                    number => {
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": 0, "intRetorno": 0, "stringRetorno": number } });
                    },
                    sizeRange
                );
            } else if (results.function  === "selecionaOpcao") {
                this.waitingCallbackResponse = true;
                
                callbackResolver.selectOption(
                    results.label,
                    results.opcoes,
                    result => {
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": "0", "intRetorno": result, "stringRetorno": result } });
                    }
                );
            } else if (results.function  === "selecionaOpcaoEx") {
                this.waitingCallbackResponse = true;
                
                callbackResolver.selectOptionExtended(
                    results.label,
                    results.opcoes,
                    (resIndex, resText, resValue) => {
                        var strRetorno = resValue ? resValue : resText;
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": "0", "intRetorno": resIndex, "stringRetorno": strRetorno } });
                    }
                );
            } else if (results.function  === "entraValor") {// Já passado na transação
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function  === "entraNumero") {
                this.waitingCallbackResponse = true;
                
                callbackResolver.numberInput(
                    results.label,
                    number => {
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": 0, "intRetorno": 0, "stringRetorno": number } });
                    },
                    sizeRange
                );
            } else if (results.function  === "entraString") {
                this.waitingCallbackResponse = true;
                
                callbackResolver.stringInput(
                    results.label,
                    str => {
                        window.dtefMobile.sendCallbackResponse(null, null,
                            { parameters: { "resultado": 0, "intRetorno": 0, "stringRetorno": str } });
                    },
                    sizeRange
                );
            } else if (results.function  === "entraMascara") {
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function  === "operacaoCancelada") {
                res = { "resultado": this.cancel == true ? "1" : "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function  === "setaOperacaoCancelada") {
                res = { "resultado": this.cancel == true ? "1" : "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "exibirQRCode") {
                callbackResolver.exibirQRCode(results.qrCode);
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else if (results.function === "removerQRCode") {
                callbackResolver.removerQRCode();
                res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
            } else {
                return;
            }    
        } else {
            res = { "resultado": "0", "intRetorno": "0", "stringRetorno": "" };
        }
    
        if (res != null) {
            window.dtefMobile.sendCallbackResponse(null, null, { parameters: res });
        }
    }


    resultsInputSizeRange(results) {
        var sizeRange;
    
        if (results.function  === "entraNumero") {
            var iDigitosExatos = results.iDigitosExatos;
            var iMinimoDigitos = results.iMinimoDigitos;
            var iMaximoDigitos = results.iMaximoDigitos;
    
            sizeRange = (iDigitosExatos > 0 ?
                    { minSize: iDigitosExatos, maxSize: iDigitosExatos } :
                    { minSize: (iMinimoDigitos > 0 ? iMinimoDigitos : undefined), maxSize: (iMaximoDigitos > 0 ? iMaximoDigitos : undefined) });
        } else if (results.function  === "entraCodigoSeguranca") {
            var tamanhoMax = results.tamanhoMax;
    
            sizeRange = { minSize: tamanhoMax, maxSize: tamanhoMax };
        } else if (results.function === "entraString") {
            var minimoDigitos = results.minimoDigitos;
            var maximoDigitos = results.maximoDigitos;
            
            sizeRange = { minSize: (minimoDigitos > 0 ? minimoDigitos : undefined), maxSize: (maximoDigitos > 0 ? maximoDigitos : undefined) };
        }
        return sizeRange;
    }

    activate(req) {
        if (window.cordova == null) {
            this.setActivation(null);
            return this.$q.resolve();
        }

        if (!this.isProviderSupported()) {
            return this.$q.reject('Provedor TEF não suportado para esta versão do plugin');
        }

        // apenas dtef no momento suporta ativação
        if (!this.isDTEF()) {
            this.setActivation(null);
            return this.$q.resolve();
        }

        req = req || { force: false, activationOpCode: 0 };
        req.activationOpCode = req.activationOpCode || 0;

        if (!req.force && !this.isActivationDirty()) {
            return this.$q.resolve();
        }

        this.setActivation(null);

        return this.$q((resolve, reject) => {
            this.executeTransaction(
                {
                    transactionType: this.transactionType.ACTIVATE_PROVIDED_TERMINAL,
                    activationOpCode: req.activationOpCode,
                    usePinpad: false
                },
                {
                    paymentDone: res => {
                        this.LogService.info(`
                            Terminal TEF após ativação:
                            companyNumber '${res.NumeroEmpresa}'
                            storeNumber '${res.NumeroLoja}'
                            terminalNumber '${res.NumeroPDV}'
                        `);
                        const config = this.ConfigService.config;
                        this.setActivation({
                            storexTerminal: {
                                companyNumber: config.companyNumber,
                                storeNumber: config.storeNumber,
                                terminalNumber: config.terminalNumber,
                                storeDocument: config.store.cnpj
                            },
                            eftTerminal: {
                                companyNumber: res.NumeroEmpresa,
                                storeNumber: res.NumeroLoja,
                                terminalNumber: res.NumeroPDV,
                                storeDocument: this.storeCNPJ
                            },
                            date: new Date().toISOString(),
                            deviceUuid: device.uuid,
                            provider: this.provider
                        });
                        resolve();
                    },
                    paymentError: (error, code) => {
                        this.LogService.error(`EFTService: ${code} - ${error}`);

                        // caso tenha ocorrido algum erro na retentativa da ativação
                        // alertamos o erro
                        if (req.activationOpCode > 0) {
                            reject(`TEF: ${error}`);
                            return;
                        }

                        // O D-TEF hoje não retorna o código do erro,
                        // logo fazemos parse na mensagem
                        if (error.includes("T037")) {
                            this.activate({
                                ...req,
                                activationOpCode: 1
                             }).then(resolve, reject);
                        } else if (error.includes("T038")) {
                            this.activate({
                                ...req,
                                activationOpCode: 2
                             }).then(resolve, reject);
                        } else {
                            reject(`TEF: ${error}`);
                        }
                    },
                    displayMessage: msg => {
                        this.LogService.info(`EFTService: ${msg}`);
                    },
                    clearDisplay: () => {
                        this.LogService.info('##################### CLEAR DISPLAY #####################');
                    }
                }
            );
        });
    }
}

class LioEFTService extends EFTServiceBase {
    constructor($q, ConfigService, PaymentService, AlertService, CartService, LogService) {
        super($q, ConfigService, LogService);
        this.PaymentService = PaymentService;
        this.AlertService = AlertService;
        this.CartService = CartService;
        this.LogService = LogService;
        this.lio = window.plugins.lio;
        this.lio.init('URmGbHdim2V3L9P1CoFEaXzf4g1PlLRYQolAGnaefQ5GhMx5L3',
                    'wSmTJ4ct5ewwFM7xN11Y27ItWeT0q0Q9AVjvZeSu4ZGHaD0tkB', LogService.log(), AlertService.error);
    }

    executeTransaction(funcParameters, callbackResolver) {      
        const checkoutReq = {
            nsu: this.ConfigService.nsuControll.currentNSU(),
            amount: funcParameters.amount * 100,
            installments: funcParameters.instalmentsNumber,
            paymentMethod: this.parsePaymentMethod(funcParameters.transactionType, funcParameters.instalmentsNumber)
        };

        return this.placeOrder()
        .then(() => {
            return this.$q((resolve, reject) => {
                this.lio.checkout(checkoutReq, resolve, reject);
            })
        })
        .then(callbackResolver.paymentDone)
        .catch(callbackResolver.paymentError);
    }

    parsePaymentMethod(paymentMethod, installmentsNumber) {
        switch (paymentMethod) {
            case EFTServiceBase.transactionType.CREDIT:
                return  installmentsNumber == 1 ? this.lio.PMNT_METHOD_CREDIT : this.lio.PMNT_METHOD_CREDIT_INSTALLMENTS_STORE;
            case EFTServiceBase.transactionType.CREDIT_WITH_INSTALMENTS:
                return this.lio.PMNT_METHOD_CREDIT_INSTALLMENTS_STORE;
            case EFTServiceBase.transactionType.DEBIT:
                return this.lio.PMNT_METHOD_DEBIT;
            default:
                throw new Error('Payment method not supported');
        }
    }

    placeOrder() {
        return this.bind()
        .then(() => {
            return this.$q((resolve, reject) => {
                if (this.PaymentService.totalPayments() == 0) {
                    const order = {
                        nsu: this.ConfigService.nsuControll.currentNSU(),
                        items: this.PaymentService.currentPayment().fiscalItems.map(fiscalItem => {
                            return {
                                sku: fiscalItem.productCode,
                                name: this.CartService.cart.fiscalItems
                                        .find(item => item.productCode == fiscalItem.productCode).productDescription,
                                value: fiscalItem.unitValue * 100,
                                quantity: fiscalItem.quantitySold
                            }
                        })
                    };
                    this.lio.placeOrder(order, resolve, reject);
                } else {
                    resolve();
                }
            });
        });
    }

    bind() {
        return this.$q((resolve, reject) => {
            this.lio.isBound(resolve, reject); // rejects if not bound
        })
        .catch(() => {
            return this.$q((resolve, reject) => {
                this.lio.bind(resolve, reject);
            });
        });
    }

    undoTransaction() {
        return this.$q((resolve, reject) => {
            this.lio.cancelOrder({
                nsu: this.ConfigService.nsuControll.currentNSU()
            }, resolve, reject);
        })
        .then(() => {
            return this.AlertService.info('Efetuar o(s) cancelamento(s) do pagamento(s) em ' +
                    'Minha LIO > Cancelamento de vendas.');
        })
        .catch(() => {
            return this.AlertService.info('Efetuar o cancelamento(s) do pagamento(s) em ' +
                    'Minha LIO > Cancelamento de vendas.');
        });
    }

    confirmTransaction() {
        this.lio.confirmOrder({nsu: this.ConfigService.nsuControll.currentNSU()}, LogService.log(), LogService.log());
    }
}

