(function() {
    'use strict';

    angular.module('easyorder.services.alert', [
		'ngResource',
		'easyorder.services.config',
		'easyorder.services.log'
	])

    .factory('AlertService', function($ionicPopup, $q, $timeout, LogService) {
        
		function info(message, okText) {
			LogService.info(message);
			return showPopup('Info', message, okText);
		}
		
		function warn(message) {
			LogService.warn(message);
			return showPopup('Aviso', message);
		}
		
		function error(message) {
			LogService.error(message);
			return showPopup('Erro', message);
		}

		function alert(title, message) {
			LogService.info(message);
			return showPopup(title, message);
		}

		function restError(httpResponse, defaultMessage) {
			var message = (httpResponse && httpResponse.data && httpResponse.data.message) ? httpResponse.data.message : defaultMessage;
			if (httpResponse && httpResponse.message){
				message  = httpResponse.message;
			}
			
			LogService.error(message);
			return showPopup('Erro', message);
		}
		
		function showPopup(title, message, okText, timeout) {
		    var ret;
		    var popupPromise = ret = $ionicPopup.alert({
				title: title,
				template: message,
				subTitle: timeout ? `(expira em ${timeout/1000} seg.)` : undefined,
				okType: "button-theme",
				okText: okText ? okText : 'OK'
			});
		    
		    if (timeout) {
                var timeoutPromise = $timeout(() => {}, timeout);
                
                ret = $q((resolve, reject) => {
                    popupPromise.then((result) => {
                        $timeout.cancel(timeoutPromise);
                        resolve(result);
                    });
                    timeoutPromise.then(() => {
                        resolve(true);
                        popupPromise.close(true);
                    });
                });
            }
		    return ret;
		}
		
		/**
		 * Exibe mensagem para confirmação. O opção a ser confirmada será rejeitada
		 * se o usuário do app não respondê-la no tempo passado ou, caso esse tempo
		 * não seja informado, após 40 segundos.
		 * Para que o popup aguarde indefinidamente é necessário passar um timeout 
		 * zero ou negativo.
		 */
		function confirm(message, timeout, title, okText, cancelText) {
		    var timeoutPromise, popupPromise;
		    
            title = title || 'Confirmação';
			const hasTimeout = timeout == null || timeout > 0;
			if (hasTimeout) {
				timeout = timeout || 40000;
				timeoutPromise = $timeout(() => {}, timeout);
			}
		    
		    var popupPromise =  $ionicPopup.confirm({
				title: title,
				subTitle: hasTimeout ? `(expira em ${timeout/1000} seg.)` : null,
				template: message,
				okType: "button-theme",
                cancelText: cancelText ? cancelText : 'Cancelar',
                okText: okText ? okText : 'OK'
			});
	        
            return $q((resolve, reject) => {
                popupPromise.then((result) => {
					if (hasTimeout) {
						$timeout.cancel(timeoutPromise);
					}
                    resolve(result);
                });
				if (hasTimeout) {
					timeoutPromise.then(() => {
						resolve(false);
						
						popupPromise.close(false);
						LogService.warn(`A confirmação '${message}' foi rejeitada por limite de tempo.`)
					});
				}
            });
		}

		function expirablePopup(templateUrl, scope, timeout) {
		    let timeoutPromise, popupPromise;
		    
		    timeout = timeout || 3000;
		    timeoutPromise = $timeout(() => {}, timeout);
		    
			let popupData = {};
			popupData.templateUrl = templateUrl;
			popupData.scope = scope;
			popupData.buttons = (scope.buttons) ? scope.buttons : [];
			popupData.cssClass = (scope.cssClass) ? scope.cssClass : undefined;
			
		    popupPromise =  $ionicPopup.confirm(popupData);
            return $q((resolve, reject) => {
                timeoutPromise.then(() => {
                    resolve();
                    popupPromise.close();
                });
            });
		}
		
		function prompt(title, placeholder, inputType, subTitle, defaultText, okText, cancelText) {
			return $ionicPopup.prompt({
	              title: title,
	              inputPlaceholder: placeholder,
	              inputType: inputType?inputType:'text',
				  okType: "button-theme",
				  subTitle: subTitle,
				  defaultText: defaultText,
				  cancelText: cancelText?cancelText:'Cancelar',
				  okText: okText?okText:'OK'
		   });
		}

		function showNumericPopup(scope, title, subTitle) {
			scope.numericPopup = { number: undefined };
			return $ionicPopup.show({
				template: '<input type="tel" ng-model="numericPopup.number" ui-mask="9?9?9">',
				title: title,
				subTitle: subTitle,
				scope: scope,
				buttons: [
					{ text: 'Cancel' },
					{ text: '<b>OK</b>',
					  type: 'button-theme',
					  onTap: function(e) {
						if (!scope.numericPopup.number) {
							e.preventDefault();
						} else {
							return scope.numericPopup.number;
						}
					}},
				]
			}).finally(() => scope.numericPopup = undefined);
		}

		/**
		 * Função mostra a view genérica de listagem, obs: O icone de remover só aparecerar 
		 * se for passado a 'funcRemove' o 'fieldRemove'.
		 * @param {*} scope 
		 * @param {*} list Lista de objetos.
		 * @param {*} lines Lista linhas com o titulo e campo a ser exibido exp: [{'title': '<Titulo>', 'field': '<campo>'}]
		 * @param {*} funcRemove Nome da função que será chamada ao clickar no icone de remover.
		 * @param {*} fieldRemove Campo chave do objeto para ser excluido.
		 */
		function showAlertList(scope, title, list, lines, funcRemove) {
			scope.listGeneric = list;
            scope.funcRemoveGeneric = funcRemove;
			scope.rowsGeneric = lines;
			scope.titleGeneric = title;
            return $ionicPopup.show({
				scope: scope,
				title: title,
                templateUrl: 'views/list-generic.html',
                cssClass: 'listGeneric',
                buttons: [{
                    text: 'OK',
                    type: 'button-theme'
                }]
            });
		}

		return {
			info: info,
			warn: warn,
			confirm: confirm,
			error: error,
			alert: alert,
			prompt: prompt,
			restError: restError,
			show: showPopup,
			showAlertList: showAlertList,
			expirablePopup: expirablePopup,
			showNumericPopup: showNumericPopup
		};
    })
})();