(function() {
    'use strict';
    angular.module('easyorder.services.log', [
        "ngResource", 
        "ngStorage", 
        "ngCordova"
    ])
    .factory('LogService', function($q, $loggers, $fileLogger, $elasticLogger, $rootScope) {
        const service = {};
        const protectedKeys = [
            'companyNumber',
            'storeNumber',
            'terminalNumber',
            'appName',
            'appVersion'
        ];
        let logCtx;

        /**
         * Inicializa as configurações do arquivo de log criando a estrutura de pastas
         */
        service.initialize = function(){
            if (!$rootScope.isCordova) {
                return $q.resolve();
            }
            const deferred = $q.defer();
 
            //Caso os diretórios da aplicação não existam, são criados
            window.resolveLocalFileSystemURL($rootScope.rootFilePath, (fileSystem) => {
                //Cria o diretório Linx
                fileSystem.getDirectory('Linx', { create: true }, (subFile) => {
                    //Cria o diretório Linx/CaixaRapido
                    subFile.getDirectory('CaixaRapido', { create: true }, (subFileLog) => {
                        //Cria o diretório Linx/CaixaRapido/logs
                        subFileLog.getDirectory('logs', { create: true }, (_) => {
                            deferred.resolve();
                        }, function (error) {
                            deferred.reject(error);
                        });
                    }, function(error) {
                        deferred.reject(error);
                    });
                }, function(error) {
                    deferred.reject(error);
                });
            }, function(error) {
                deferred.reject(error);
            });
           
            return deferred.promise;
        };

        /**
         * Configura os parâmetros para o arquivo de Log
         */ 
        service.configuration = function(config) {
            $fileLogger.setStorageMethod($rootScope.fullRootFilePathLogs);
            $loggers.addLogger($fileLogger);
            return $fileLogger.compressFiles();
        };

        service.info = function(msg, parameters){
            if (parameters){
                $loggers.log('INFO', msg, parameters);
            } else {
                $loggers.info(msg);
            }
        };

        service.warn = function(msg, parameters){
            if (parameters) {
                $loggers.log('WARN', msg, parameters);
            } else {
                $loggers.warn(msg);
            }
        };

        service.error = function(msg, parameters){
            if (parameters) {
                $loggers.log('ERROR', msg, parameters);
            } else {
                $loggers.error(msg);
            }
        };

        service.debug = function(msg, parameters){
            if (parameters) {
                $loggers.log('DEBUG', msg, parameters);
            } else {
                $loggers.debug(msg);
            }
        };

        service.fatal = function(msg) {
            $loggers.fatal(msg);
        };

        service.responseError = function(msg) {
            $loggers.responseError(msg);
        };

        service.log = function(msg){ 
            $loggers.log(msg);
        };

        /**
         * Print no arquivo de log quando a aplicação é recarregada, ou solicitada outra configuração
         */
        service.logStatus = function(terminalNumber, storeNumber, companyNumber, currentNSU, host, port) {
            $loggers.info('Configurações do Caixa Rápido: Versão: [' + $rootScope.appVersion + '],' 
                    + ' Data: [' + new Date().toLocaleDateString() +'], ' 
                    + ' Nsu: [' + currentNSU + '],' 
                    + ' Loja: [' + storeNumber + '],' 
                    + ' Terminal: [' + terminalNumber + '],' 
                    + ' Estabelecimento: [' + companyNumber + '] ');
            $loggers.info("Webservice: [" + host + "], " 
                    + "porta: [" + port + "]");
        };

        service.putKey = (key, value, mode) => {
            if (!key) {
                $loggers.warn("[LogService] Foi chamado putKey sem 'key'");
                return;
            }
            if (!value) {
                $loggers.warn("[LogService] Foi chamado putKey sem 'value'");
                return;
            }

            if (!logCtx) {
                $loggers.warn("[LogService] Foi chamado putKey antes da inicialização");
                return;
            }

            // REPLACE, APPEND
            mode = mode || 'REPLACE';

            if (!logCtx[key] || mode == 'REPLACE') {
                logCtx[key] = value;
            } else if (mode == 'APPEND') {
                logCtx[key] = `${logCtx[key]},${value}`;
            } else {
                $loggers.warn(`[LogService] Foi chamado putKey com mode desconhecido: ${mode}`);
            }
        };

        service.getKey = (key) => {
            if (!logCtx) {
                return null;
            }
            return logCtx[key];
        };

        service.clearKey = (key) => {
            if (!logCtx) {
                $loggers.warn(`[LogService] Foi chamado clearKey porém o contexto de log ainda não foi inicializado.`);
                return;
            }

            if (!logCtx.key) {
                return;
            }

            if (!key) {
                $loggers.warn(`[LogService] Foi chamado clearKey com chave nula`);
                return;
            }

            if (protectedKeys.includes(key)) {
                $loggers.warn(`[LogService] Foi chamado clearKey com valor de chave protegida, não será removido: ${key}`);
                return;
            }

            const containsKey = Object.keys(logCtx)
                .filter(k => !protectedKeys.includes(k))
                .some(k => k == key);
            
            if (containsKey) {
                delete logCtx[key];
            }
        };

        service.clearKeys = () => {
            if (!logCtx) {
                $loggers.warn(`[LogService] Foi chamado clearKeys porém o contexto de log ainda não foi inicializado.`);
                return;
            }

            Object.keys(logCtx)
                .filter(k => !protectedKeys.includes(k))
                .forEach(k => {
                    delete logCtx[k];
                });
        };
        
        service.reconfigure = function(config) {
            $loggers.removeLogger($elasticLogger);

            logCtx = {
                companyNumber: `${config.companyNumber}`.padStart(5, '0'),
                storeNumber: `${config.storeNumber}`.padStart(4, '0'),
                terminalNumber: `${config.terminalNumber}`.padStart(3, '0'),
                appName: 'storex-mobile',
                appVersion: $rootScope.appVersion,
            };

            $fileLogger.reconfigure(logCtx);

            // se tiver habilitado o Elastic
            if (config.business.elastic_enabled) {
                if (config.business.elastic_servers) {
                    var servers = config.business.elastic_servers.split(';');
                    $elasticLogger.reconfigure(servers, logCtx);
                    $loggers.addLogger($elasticLogger);

                    $loggers.debug(`Servidores do Elastic: ${servers.join(', ')}`);
                } else {
                    $loggers.error('Envio de logs via Elastic habilitado mas nenhum servidor foi informado. Elastic não será utilizado!');
                }
            }
        };

        return service;
    });
})();
