(function() {
    'use strict';
    angular.module('easyorder.services.messaging', [
        'ngResource',
        'easyorder.services.config',
        'easyorder.services.log'
    ])

    .factory('MessagingService', function($q, ConfigService, $resource, MailService,
                            LogService) {
        
        const PHONE_TABLE = {
            BRA: {
                countryCode: '55',
                countryCodeLength: 2,
                minPhoneNumberLength: 8,
                maxPhoneNumberLength: 9,
                areaCodeLength: 2,
                minLocalPhoneNumberLength: 10,
                maxLocalPhoneNumberLength: 11,
                minIntPhoneNumberLength: 12,
                maxIntPhoneNumberLength: 13,
                phoneNumberIdx: 0,
                localPhoneNumberIdx: 2,
                intPhoneNumberIdx: 4,
                localAreaCodeIdx: 0,
                intAreaCodeIdx: 2,
                intCountryCodeIdx: 0,
                areaCodes: [
                    '61', , '62', '64', '65', '66', '67',
                    '82', '71', '73', '74', '75', '77',
                    '85', '88', '98', '99', '83', '81', '87',
                    '86', '89', '84', '79', '68', '96', '92', '97',
                    '91', '93', '94', '69', '95', '63', '27', '28',
                    '31', '32', '33', '34', '35', '37', '38',
                    '21', '22', '24', '11', '12', '13', '14', '15',
                    '16', '17', '18', '19', '41', '42', '43', '44',
                    '45', '46', '51', '53', '54', '55', '47', '48',
                    '49',
                ],
            }
        };

        var service = {
            url : ''
        };
        
        /**
         * 
         */
        service.loadSettings = function() {
            service.url = ConfigService.config.linxwebservices.secure ? 'https://' : 'http://';
            service.url += ConfigService.config.linxwebservices.host + ':';
            service.url += ConfigService.config.linxwebservices.port;
            service.url += '/linx-webservices/v1/messaging';
        };
        
        /**
         * Envio de mensagem por SMS
         */
        service.sendSms = function(from, to, msg, contentType, contentParameters) {
            var params = {
                from : from,
                to : parsePhoneNumber(to.replace(/\D/g, "")),
                contents: [parseContent(contentType, msg, contentParameters)]
            };

            return $resource(service.url + '/sms', {}, {
                send : {
                    method : 'POST'
                }
            }).send({}, params).$promise;
        };

        /**
         * Envio de mensagem por SMS
         */
        service.sendWhatsApp = function(from, to, msg, contentType, contentParameters) {
            var params = {
                from: from,
                to: parsePhoneNumber(to.replace(/\D/g, "")),
                contents: [parseContent(contentType, msg, contentParameters)]
            };

            return $resource(service.url + '/whatsapp', {}, {
                send : {
                    method : 'POST'
                }
            }).send({}, params).$promise;
        };

        function parseContent(contentType, content, contentParameters) {
            contentType = contentType || 'text';

            const ret = {
                type: contentType,
                parameters: contentParameters
            };

            if (contentType == 'text') {
                ret.text = content;
            }

            if (contentType == 'external_template') {
                ret.templateId = content;
            }

            return ret;
        }

        service.isPhoneNumberValid = function(number, country) {
            let ret = false;
            // NUMBER, LOCAL, INTERNATIONAL
            let phoneType;
            let phoneIdx;
            let areaCodeIdx;
            let countryCodeIdx;

            country = country || 'BRA';
            const table = PHONE_TABLE[country];

            if (!table) {
                LogService.warn(`Country '${country}' não mapeado para tratamento de telefones`);
                return false;
            }

            const sequences =
                ConfigService.config.business.unallowed_phone_numbers || [];

            if (number) {
                number = number.replace(/\D/g, '');
                const length = number.length;

                if (isInRange(length,
                              table.minPhoneNumberLength,
                              table.maxPhoneNumberLength)) {
                    ret = true;
                    phoneType = 'NUMBER';
                    phoneIdx = table.phoneIdx;
                    areaCodeIdx = -1;
                    countryCodeIdx = -1;
                } else if (isInRange(length,
                                     table.minLocalPhoneNumberLength,
                                     table.maxLocalPhoneNumberLength)) {
                    ret = true;
                    phoneType = 'LOCAL';
                    phoneIdx = table.localPhoneNumberIdx;
                    areaCodeIdx = table.localAreaCodeIdx;
                    countryCodeIdx = -1;
                } else if (isInRange(length,
                                     table.minIntPhoneNumberLength,
                                     table.maxIntPhoneNumberLength)) {
                    ret = true;
                    phoneType = 'INTERNATIONAL';
                    phoneIdx = table.intPhoneNumberIdx;
                    areaCodeIdx = table.intAreaCodeIdx;
                    countryCodeIdx = table.intCountryCodeIdx;
                } else {
                    return false;
                }
            }

            if (ret) {
                if (countryCodeIdx >= 0) {
                    const countryCode = number.substring(
                        countryCodeIdx,
                        table.countryCodeLength
                    );
                    if (!countryCode == table.countryCode) {
                        return false;
                    }
                }
                if (areaCodeIdx >= 0) {
                    const areaCode = number.substring(
                        areaCodeIdx,
                        table.areaCodeLength,
                    );
                    if (!table.areaCodes.includes(areaCode)) {
                        return false
                    }
                }
                number = number.substring(phoneIdx);
                for (const seq of sequences) {
                    if (number.startsWith(seq)) {
                        return false;
                    }
                }
            }

            return ret;
        };

        function isInRange(number, min, max) {
            if (number == null)
                return false;
            return number >= min
                && number <= max;
        }

        service.sendMessageToClient = (customer, message, subject, contentType, contentParameters) => {
            var rejectedPromises = [];
            const methods = ConfigService.config.business.payment_link_send_methods;
            
            methods.forEach(method => {
                if (method === 'WHATSAPP') {
                    var rejectedZapPromise = $q.defer();
                    
                    rejectedPromises.push(rejectedZapPromise.promise);
                    
                    $q.when(customer.address && customer.address.phone ? $q.resolve() : $q.reject("WHATSAPP: Cliente sem informação de telefone. Link não pode ser enviado."))
                    .then(() => service.sendWhatsApp(undefined, customer.address.phone, message, contentType, contentParameters))
                    .then(rejectedZapPromise.reject)
                    .catch(error => {
                        LogService.error(error);
                        rejectedZapPromise.resolve();
                    });
                } else if (method === 'SMS') {
                    var rejectedSMSPromise = $q.defer();
                    
                    rejectedPromises.push(rejectedSMSPromise.promise);
                    
                    $q.when(customer.address && customer.address.phone ? $q.resolve() : $q.reject("SMS: Cliente sem informação de telefone. Link não pode ser enviado."))
                    .then(() => service.sendSms(undefined, customer.address.phone, message, contentType, contentParameters))
                    .then(rejectedSMSPromise.reject)
                    .catch(error => {
                        LogService.error(error);
                        rejectedSMSPromise.resolve();
                    });
                } else if (method === 'E-MAIL') {
                    var rejectedEmailPromise = $q.defer();
                    
                    rejectedPromises.push(rejectedEmailPromise.promise);
                    
                     $q.when(customer.email ? $q.resolve() : $q.reject("E-Mail: Cliente sem informação de e-mail. Link não pode ser enviado."))
                    .then(() => MailService.send({subject: subject || 'Olá!', type: 'generic', destinations: [customer.email], bodyData: {text: message}}))
                    .then(rejectedEmailPromise.reject)
                    .catch(error => {
                        LogService.error(error);
                        rejectedEmailPromise.resolve();
                    });
                }
            });
            var ret = $q.defer();
            
            $q.all(rejectedPromises) //se todas falharem
            .then(ret.reject)
            .catch(ret.resolve);
            
            return ret.promise;
        }

 		/**
         * 
         */
        var parsePhoneNumber = (number) => {
            var ret;
            
            if (number) {
                if (number.startsWith('55')) {
                    ret = number;
                } else {
                    ret = "55" + number;
                }
            }
            return ret;
        };

        service.loadSettings();

        return service;
    });
})();
