(function() {
    'use strict';

    angular.module('easyorder.services.order', [
                                                'ngResource',
                                                'easyorder.services.config',
                                                'easyorder.services.user',
                                                'easyorder.services.cart'
                                            ])

    .factory('OrderService', function($resource,
                                        ConfigService,
                                        UserService,
                                        CartService) {

        const SYSTEM_NAME = 'CAIXA_RAPIDO';

        var service = {
            url: '',
            orderStatus: {
                NEW: 'NEW',
                CANCELED: 'CANCELED',
                CHECKOUT: 'CHECKOUT',
                ON_HOLD: 'ON_HOLD',
                PROCESSING: 'PROCESSING',
                PAID: 'PAID',
                DELIVERED: 'DELIVERED',
				READY_TO_PACK: 'READY_TO_PACK',
				PAYMENT: 'PAYMENT',
                CANCEL_REQUESTED: 'CANCEL_REQUESTED'
            }
        };

        service.loadSettings = function() {
            service.url  = ConfigService.config.linxwebservices.secure ? 'https://' : 'http://';
            service.url += ConfigService.config.linxwebservices.host + ':';
            service.url += ConfigService.config.linxwebservices.port;
            service.url += '/linx-webservices/v1/orders';
        };

        service.get = function(params, success, error) {
            service.loadSettings();

            return $resource(service.url + '/show/:companyNumber/:storeNumber/:orderNumber', {
                companyNumber: params.companyNumber,
                storeNumber: params.storeNumber,
                orderNumber: params.orderNumber
            }).get({}, success, error);
        };

        service.getExternal = function(params, success, error) {
            service.loadSettings();

            return $resource(service.url + '/external-order/show/:storeNumber/:terminalNumber/:externalOrderNumber').get({
                storeNumber: params.storeNumber,
                terminalNumber: ConfigService.config.terminalNumber,
                externalOrderNumber: params.externalOrderNumber
            }, success, error);
        };

        service.listExternal = function(params, success, error) {
            service.loadSettings();

            return $resource(service.url + '/external-order/list/:companyNumber/:storeNumber/:externalOrderNumber').get({
            	companyNumber: params.companyNumber,
                storeNumber: params.storeNumber,
                externalOrderNumber: params.externalOrderNumber
            }, success, error);
        };

        service.list = function(params, success, error) {
            service.loadSettings();
            return $resource(service.url + '/list').get(params, success, error).$promise;
        };

        service.status = function() {
            service.loadSettings();
            return $resource(service.url + '/status').get();
        };

        service.create = function(order, success, error) {
            service.loadSettings();
            order.originSystem = SYSTEM_NAME;

            order.companyNumber  = ConfigService.config.companyNumber;
            order.storeNumber    = ConfigService.config.storeNumber;
            order.terminalNumber = ConfigService.config.terminalNumber;
            order.operatorCode   = UserService.operatorCode();

            delete order.cashChange;

            return $resource(service.url + '/create', {}, { create: { method: 'POST' } }).create({}, order, success, error).$promise;
        };

        /**
         * Realiza a requisição de mesclagem de mútiplos pedidos
         *
         * @param {Requisição de mesclagem de pedidos} request
         * @param {Mesclagem realizada com sucesso} success
         * @param {Falha na mesclagem} error
         * @returns
         */
        service.merge = function(orders, customer, success, error) {
            service.loadSettings();
            var request = {};
            request.originSystem   = SYSTEM_NAME;
            request.nsu            = ConfigService.nsuControll.currentNSU();
            request.date           = new Date();
            request.companyNumber  = ConfigService.config.companyNumber;
            request.storeNumber    = ConfigService.config.storeNumber;
            request.terminalNumber = ConfigService.config.terminalNumber;
            request.operatorCode   = UserService.operatorCode();
            request.customer       = customer
            request.orders         = orders;

            return $resource(service.url + '/merge', {}, { merge: { method: 'POST' } }).merge({}, request, success, error).$promise;
        };

        service.reserveAndCreate = function(order, freightInfo, success, error) {
        	service.loadSettings();
        	order.originSystem = SYSTEM_NAME;

        	order.companyNumber  = ConfigService.config.companyNumber;
        	order.storeNumber    = ConfigService.config.storeNumber;
        	order.terminalNumber = ConfigService.config.terminalNumber;
            order.operatorCode   = UserService.operatorCode();

        	var params = {
        		order: order,
        		freightInfo: freightInfo
        	};

        	return $resource(service.url + '/reserve-create', {}, { reserveAndCreate: { method: 'POST' } }).reserveAndCreate({}, params, success, error);
        };


        service.confirmInventoryAndCreate = function(order, inventoryInfo, success, error) {
        	service.loadSettings();
        	order.originSystem = SYSTEM_NAME;

        	order.companyNumber  = ConfigService.config.companyNumber;
        	order.storeNumber    = ConfigService.config.storeNumber;
        	order.terminalNumber = ConfigService.config.terminalNumber;
            order.operatorCode   = UserService.operatorCode();

        	var params = {
        		order: order,
        		inventoryInfo: inventoryInfo
        	};

        	return $resource(service.url + '/confirm-create', {}, { confirmInventoryAndCreate: { method: 'POST' } }).confirmInventoryAndCreate({}, params, success, error);
        };

        service.update = function(order, success, error) {
            service.loadSettings();

            if (order.originSystem == '') {
                order.originSystem = SYSTEM_NAME;
            }
            order.operatorCode   = UserService.operatorCode();

            return $resource(service.url + '/update', {}, { 'update': { method: 'POST' } }).update({}, order, success, error).$promise;
        };

        service.updateAsync = function(order, success, error) {
            service.loadSettings();

            if (order.originSystem == '') {
                order.originSystem = SYSTEM_NAME;
            }

            return $resource(service.url + '/update/async', {}, { updateAsync: { method: 'POST' } }).updateAsync({}, order, success, error).$promise;
        };

        service.statusUpdate = function(companyNumber, storeNumber, orderNumber, status, success, error) {
            service.loadSettings();

            var updateParams = {
        		companyNumber: companyNumber,
        		storeNumber: storeNumber,
        		orderNumber: orderNumber,
        		status: status
            };
            return $resource(service.url + '/update/status', {}, { statusUpdate: { method: 'POST' } }).statusUpdate({}, updateParams, success, error).$promise;
        };

        service.cartToOrderConverter = function(cart) {
            // O objeto de venda / pedido é representado pelo carrinho hoje. É preciso corrigir essa distorção.
            var order = undefined;

            if (cart) {
                order = JSON.parse(JSON.stringify(cart));
                order.orderRetrieved = undefined;
                order.currentOrder = undefined;
                delete order.paymentLinkSettings;
                delete order.orderStatus;
                delete order.loyaltyData;

                // no caso do Order, o atributo de pagamento é 'orderPayments' ao invés de 'salePayments'.
                if (order.salePayments) {
                    order.orderPayments = order.salePayments;
                    order.salePayments = undefined;
                }

                if (order.orderPayments) {
                    var index;

                    for (index in order.orderPayments) {
                        // Remove campos inexistentes no OrderPayment.
                        const orderPayment = order.orderPayments[index];

                        orderPayment.customer = orderPayment.paymentCustomer;
                        service.toOrderCustomer(orderPayment.customer);
                        delete orderPayment.paymentCustomer;

                        // campos de ewallet
                        orderPayment.eWalletDocument = orderPayment.walletDocument;
                        orderPayment.eWalletName = orderPayment.walletName;
                        orderPayment.eWalletCode = orderPayment.walletId;

                        if (!orderPayment.billedPaymentMethod && orderPayment.walletProductType) {
                            orderPayment.billedPaymentMethod = orderPayment.walletProductType;
                        }
                        if (!orderPayment.customer && orderPayment.walletPaymentDocument) {
                            orderPayment.customer = {
                                // FIXME: criar lógica para detectar tipo de documento
                                documentType: 'CPF',
                                document: orderPayment.walletPaymentDocument,
                            }
                        }
                        if (!orderPayment.paymentTransactionId && orderPayment.transactionId) {
                            orderPayment.paymentTransactionId = orderPayment.transactionId;
                        }
                        if (!orderPayment.paymentTerminalId && orderPayment.remoteTerminalId) {
                            orderPayment.paymentTerminalId = orderPayment.remoteTerminalId;
                        }

                        // deletando campos de ewallet não existentes no orderpayment
                        delete orderPayment.walletDocument;
                        delete orderPayment.walletName;
                        delete orderPayment.walletId;
                        delete orderPayment.transactionId;
                        delete orderPayment.walletProductType;
                        delete orderPayment.walletPaymentDocument;
                        delete orderPayment.remoteTerminalId;

                        if (!orderPayment.paymentAuthorizerName && orderPayment.acquirerName) {
                            orderPayment.paymentAuthorizerName = orderPayment.acquirerName;
                        }

                        delete orderPayment.type;
                        delete orderPayment.providerName;
                        delete orderPayment.paymentDescriptionPlan;
                        delete orderPayment.storeNumber;
                        delete orderPayment.voucherNumber; // Usando cardNumber.
                        delete orderPayment.voucherOwner;
                        delete orderPayment.acquirerName;
                    }
                }
                // Campos inexistentes no 'Order'
                order.campaignPayment = undefined;
                order.giftsQuantity = undefined;
                order.totalDiscount = undefined;
                order.paymentCampaignProcess = undefined;
                order.campaignPaymentTotalDiscount = undefined;
                order.campaignsPayment = undefined;
                order.promotions = undefined;
                order.deliveryOption = undefined;
                order.serviceItems = undefined;
                order.showroomingInfo = undefined;
                order.prescriptions = undefined;
                order.suggestions = undefined;
                order.giftBenefits = undefined;
                order.coupons = undefined;

                if (order.fiscalItems) {
                    var index;

                    for (index in order.fiscalItems) {
                        // Remove campos inexistentes no FiscalItem.
                        var fiscalItem = order.fiscalItems[index];
                        //fiscalItem.productType = undefined;
                        fiscalItem.groupCode = undefined;
                        fiscalItem.totalExhib = undefined;
                        fiscalItem.unitValueType = undefined;
                        fiscalItem.visibleInSearch = undefined;
                        fiscalItem.originalPrice = undefined;
                        fiscalItem.priceChangeReason        = undefined;
                        // TODO remover ou adicionar no FiscalItem do order?
                        fiscalItem.imei = (fiscalItem.itemUuid && fiscalItem.itemUuid.type ? fiscalItem.itemUuid.value : undefined);
                        fiscalItem.itemUuid = undefined;
                        fiscalItem.uuidInfo = undefined;
                        fiscalItem.serviceItemsAvailable = undefined;
                        fiscalItem.nonFiscalItems = undefined;
                        fiscalItem.hasParent = undefined;
                        fiscalItem.standardTypeVerifierDigitCalculation = undefined;
                        fiscalItem.seller = undefined;
                        fiscalItem.batchCodeRequired = undefined;

                        if (fiscalItem.discounts) {
                            var i;

                            for (i in fiscalItem.discounts) {
                                var discount = fiscalItem.discounts[i];
                                discount.factor = undefined;
                                discount.indentifierCode = undefined;
                            }
                        }

                        if (fiscalItem.lineItemTaxes) {
                        	var i;

                        	for(i in fiscalItem.lineItemTaxes){
                        		var lineItemTax = fiscalItem.lineItemTaxes[i];
                        		lineItemTax.effectiveICMSCalculationBasisReductionValue = undefined;
                        		lineItemTax.effectiveICMSValue = undefined;
                        	}
                        }
                    }
                }

                if (order.discounts) {
                    var index;

                    for (index in order.discounts) {
                        var discount = order.discounts[index];
                        discount.factor = undefined;
                    }
                }

                service.toOrderCustomer(order.customer);
                service.toOrderCustomer(order.fiscalCustomer);
            }
            return order;
        };

        service.toOrderCustomer = (customer) => {
            if (!customer) {
                return;
            }

            customer.avaliableCredit = undefined;
            customer.avaliablePaymentMethods = undefined;
            customer.campaignCode = undefined;
            customer.cardCode = undefined;
            customer.closeBillingDay = undefined;
            customer.custumerCode = undefined;
            customer.dueDate = undefined;
            customer.offerCustomerInsurance = undefined;
            customer.processor = undefined;
            customer.contractNumber = undefined;
            customer.contractDigit = undefined;
            customer.insuranceValue = undefined;
            customer.contractDigit = undefined;
            customer.employeeDiscountPercentage = undefined;
            customer.avaliableCreditFinancialItems = undefined;
            customer.avaliableFastCashWithdraw = undefined;
            customer.deliveryAddresses = undefined;
            customer.groupCode = undefined;
            delete customer.cardNumber;
            delete customer.customerWarning;
            delete customer.customerWarnings;
            delete customer.firstPurchaseDiscountPercentage;
            delete customer.loyaltyParticipant;
            delete customer.faceBiometryStatus;

            return customer;
        }

		service.validateOrder = function(order, currentStoreNumber, successCallback, errorCallback, allowedStatuses) {
			var errorMessage = null;

            allowedStatuses = allowedStatuses || ['CHECKOUT', 'ON_HOLD', 'READY_TO_PACK'];

			if (order) {
				if (order.status) {
					// verifica se o tipo de pedido permite resgate
					if (!allowedStatuses.includes(order.status)) {
						errorMessage = `O status do pedido não permite alteração no ${APP_NAME_DESC}!`;
					} else if (order.destinationStoreNumber != null && order.destinationStoreNumber !== currentStoreNumber) {
						errorMessage = "Pedido não pode ser resgatado! Pedido destino loja '" + order.destinationStoreNumber + "'. Loja atual: '" + currentStoreNumber + "'!";
					} else if (order.orderPayments && order.orderPayments.length && order.status != 'READY_TO_PACK') {
					    for (var index in order.orderPayments) {
					        var payment = order.orderPayments[index];

					        if ((payment.paymentAuthorizationNSU || payment.paymentAuthorizationCode)
                                    && (!payment.mandatoryMethod || payment.paymentState != 'CONFIRMED')) {
					            errorMessage = "Pedido contém pagamentos já efetuados. Tipo de pedido não suportado pela aplicação!";
					            break;
					        }
					    }
					}

                    if (order.originSystem == 'OMS'
                            && order.orderPayments != null
                            && order.orderPayments.length > 0
                            && order.orderPayments.some(p => p.paymentState != 'CONFIRMED')) {
					    errorMessage = "Pedido contém pagamentos já efetuados. Tipo de pedido não suportado pela aplicação!";
                    }

                    if (!errorMessage && CartService.cart.fiscalItems) {
                        for (var index in CartService.cart.fiscalItems) {
                            if(!CartService.isAValidMarketingCode(CartService.cart.fiscalItems[index], true)) {
                                errorMessage = "O pedido não pode ser resgatado por já existirem itens no carrinho que não são permitidos para esta operação";
                            }
                        }
                    }
                    if(!errorMessage && CartService.cart.nonFiscalItems) {
                        for (var index in CartService.cart.nonFiscalItems) {
                            if(!CartService.isAValidMarketingCode(CartService.cart.nonFiscalItems[index], true)) {
                                errorMessage = "O pedido não pode ser resgatado por já existirem itens no carrinho que não são permitidos para esta operação";
                            }
                        }
                    }
                    if(!errorMessage && CartService.cart.serviceItems) {
                        for (var index in CartService.cart.serviceItems) {
                            if(!CartService.isAValidMarketingCode(CartService.cart.serviceItems[index], true)) {
                                errorMessage = "O pedido não pode ser resgatado por já existirem itens no carrinho que não são permitidos para esta operação";
                            }
                        }
                    }
                    if (!errorMessage && order.customer && order.customer.address && order.customer.address.zipCode) {
                        let zipCode = order.customer.address.zipCode;
                        // Utilizei replace simples devido à má estética de soluções com regex
                        zipCode = zipCode.replace('.', ''); // Remove pontos do CEP
                        zipCode = zipCode.replace('-', ''); // Remove traços do CEP
                        zipCode = zipCode.replace(' ', ''); // Remove espaços em branco

                        if (zipCode.length != 8) {
                            errorMessage = "O Pedido não pode ser resgatado! CEP Inválido";
                        }
                    }

					if (!errorMessage) {
					    if (order.fiscalItems) {
				    	 	var SOLD_IN_STORE_Itens = order.fiscalItems.filter(function(fiscalItem) {
					            return !fiscalItem.deliveryOption || fiscalItem.deliveryOption === 'SOLD_IN_STORE';
					        });

							if (!SOLD_IN_STORE_Itens.length && order.type !== 'SHOWROOMING') {
					            var lastDeliveryOption = null;
						        var currentDeliveryOption = null;
					            var index;

					            for (index in order.fiscalItems) {
					                var fiscalItem = order.fiscalItems[index];
							        	if (fiscalItem.deliveryOption) {
							        		currentDeliveryOption = fiscalItem.deliveryOption;
							        	} else {
							        		// se o DeliveryOption for nulo, considera como 'SOLD_IN_STORE'
							        		currentDeliveryOption = 'SOLD_IN_STORE';
							        	}
					                if (lastDeliveryOption == null) {
							        		// primeira checagem
							        		lastDeliveryOption = currentDeliveryOption;
					                } else {
						        		// valida se o tipo é diferent, bloqueando o resgate.
						        		if (currentDeliveryOption !== lastDeliveryOption && !ConfigService.config.business.order_with_delivery_of_merged_items_allowed) {
						        			errorMessage = 'Tipo do delivery do item de descrição ' + fiscalItem.productDescription + ' ('+  currentDeliveryOption +') difere do item anterior ('+ lastDeliveryOption +')! Pedido não pode ser resgatado!';;
						        		}
					                }
					            }
							}
					    } else {
							errorMessage = "Pedido não contém itens!";
						}
					}
				} else {
					errorMessage = 'Status do pedido nulo!';
				}

				if (errorMessage == null) {
					successCallback(order);
				} else {
					const error = {
						message: errorMessage
					};
					errorCallback(error);
				}
			} else {
                const error = {
                    message: "Pedido nulo!"
                };
                errorCallback(error);
			}
		}
        return service;
    });
})();
