(function () {
    'use strict';

    angular.module('easyorder.services.paymentlink', [
    ]).service('PaymentLinkService', function(
        ConfigService,
		CartService,
        LogService,
        $http
    ) {
        const HTTP = 'http://';
        const HTTPS = 'https://';
        const ACCEPTED_METHOD_TYPES = [
            'CREDIT_CARD',
            'DEBIT_CARD',
            'PRIVATE_LABEL',
            'VOUCHER',
            'BOLETO',
            'BANK_TRANFER',
            'SAFETYPAY',
            'CASH',
            'CHECKOUT',
            'PIX'
        ];

        const SERVICE = '/linx-webservices/v1/paymentlink';

        const CHECKOUT = '/checkout';
        const TRANSACTION = '/transaction'
        const CANCEL_CHECKOUT_AND_PAYMENTS = '/undo-checkout';
        const DISABLE_CHECKOUT = '/disable-checkout';
        const BUILD_MESSAGE = '/build-message';

        function url(action) {
            const lws = ConfigService.config.linxwebservices;
            return `${protocol()}${lws.host}:${lws.port}${SERVICE}${action}`;
        }

        function protocol() {
            const lws = ConfigService.config.linxwebservices;
            return lws.secure ? HTTPS : HTTP;
        }

        function serviceHeaders() {
            const providerName = ConfigService.config.business.paymentlink_service_provider;
            if(providerName) {
                return { headers : { 
                    'Service-Provider': providerName  
                }};
            } else {
                return undefined;
            }
        }

        const service = {
            checkout: (checkoutRequest) => {
                if (checkoutRequest.customer) {
                    checkoutRequest.customer.customerId = checkoutRequest.customer.customerId || '0000';
                    
                    if (checkoutRequest.customer.address) {
                        checkoutRequest.customer.address.country = checkoutRequest.customer.address.country || 'BR'; 
                    }
                }
                return $http.post(url(CHECKOUT), checkoutRequest, serviceHeaders());
            },
            cancelCheckoutPayments: (cancelPaymentRequest) => {
                return $http.post(url(CANCEL_CHECKOUT_AND_PAYMENTS), cancelPaymentRequest, serviceHeaders());
            },
            disablePaymentLink: (orderNumber) => {
                return $http.post(url(DISABLE_CHECKOUT), { orderNumber: orderNumber }, serviceHeaders());
            },
            verifyPaymentLink: (orderNumber) => {
                return $http.get(url(`${CHECKOUT}?orderNumber=${orderNumber}`), serviceHeaders());
            },
            verifyTransaction: (orderNumber) => {
				return $http.get(url(`${TRANSACTION}?orderNumber=${orderNumber}` ), serviceHeaders());
            },
            buildMessage: (buildMessageRequest) => {
                return $http.post(url(BUILD_MESSAGE), buildMessageRequest, {}, serviceHeaders());
            }
        };

		service.getCustomer = () => {
			var customer = CartService.cart.customer;
			
            let ret = {
                customerId: 	customer.customerId || '0000',
                name: 			customer.name,
                document: 		customer.document,
                documentType: 	customer.documentType,
                email: customer.email
            }
            
            if (customer.address) {
                ret.address = {
                    addressName:        customer.address.addressName,
                    addressNumber:      customer.address.addressNumber,
                    addressComplement:  customer.address.addressComplement,
                    country:            "BR",
                    state:              customer.address.state,
                    city:               customer.address.city,
                    zipCode:            customer.address.zipCode,
                    neighborhood:       customer.address.neighborhood,
                    phone:              customer.address.phone
                }
            }
            return ret;
        }
        
        /**
         * Gera o número do pedido para criação de link de pagamento
         */
         service.getOrderNumber = () => {
            const year = new Date().getFullYear();
            const month = new Date().getMonth();
            const day = new Date().getDate();
            
            let sumOfDate = year + month + day;
            sumOfDate = ("0000" + sumOfDate).slice(-4);
            
            const hours = new Date().getHours();
            const minutes = new Date().getMinutes();
            const seconds = new Date().getSeconds();
            
            let sumOfTime = hours + minutes + seconds;
            sumOfTime = ("000" + sumOfTime).slice(-3);
            
            let storeNumber = ConfigService.config.storeNumber;
            storeNumber = ("0000" + storeNumber).slice(-4);
            
            let terminalNumber = ConfigService.config.terminalNumber + "";
            terminalNumber = ('0000'+ terminalNumber).slice(-4);
            
            let nsu = ConfigService.nsuControll.currentNSU() + "";
            nsu = ('0000'+ nsu).slice(-4);
            
            return sumOfDate + sumOfTime + storeNumber + terminalNumber + nsu; //SOMA DA DATA + SOMA DO TEMPO + Terminal + Loja + NSU
        };

        /**
         * Esta separação é necessária devido à forma que a API da Pagarme recebe as composições de formas de pagamentos
         */
        function getAvailablePaymentsMethods() {
            let acceptedMethod = [];
            let acceptedMethodsCombination =[];
            const methods = ConfigService.config.business.paymentlink_available_payment_methods;

            methods.forEach(item => {
                // Formas de multimeios da Pagar.me possuem uma especificidade da própria API, por este motivo faz-se necessário esta separação.
                switch(item) {
                    case 'MULTI_CREDIT':
                        acceptedMethodsCombination.push(['CREDIT_CARD', 'CREDIT_CARD']);
                        break;
                    case 'MULTI_DEBIT':
                        acceptedMethodsCombination.push(['DEBIT_CARD', 'DEBIT_CARD']);
                        break;
                    case "MULTI_CREDIT_DEBIT":
                        acceptedMethodsCombination.push(['CREDIT_CARD', 'DEBIT_CARD']);
                        break;
                    default:
                        if(ACCEPTED_METHOD_TYPES.includes(item)) {
                            acceptedMethod.push(item);
                        } else {
                            LogService.warn('Forma de pagamento ', item, ' não mapeada, ela não será adicionada ao link.');
                        }
                        break;
                }
            });
            return {
                acceptedMethod,
                acceptedMethodsCombination
            };
        }
        
        /**
         *
         */
        service.getPlansLinx = (paymentValue) => {
            const availablePaymentsMethods = getAvailablePaymentsMethods();
            return {
                acceptedMethod: availablePaymentsMethods.acceptedMethod,
                acceptedMethodsCombination: availablePaymentsMethods.acceptedMethodsCombination,
                acceptedBrand: [
                    "AMX",
                    "AVI",
                    "BNC",
                    "CAB",
                    "CDZ",
                    "CUP",
                    "DCI",
                    "ELO",
                    "HIP",
                    "JCB",
                    "MCI",
                    "SIC",
                    "SOR",
                    "VIS"
                ],
                installments: service.getInstallmentsLinx(paymentValue)
            }
        };
        
        /**
         * Cria as parcelas
         */
        service.getInstallmentsLinx = (paymentValue) => {
            var installments = [];

            for (var i = 1; i <= 12; i++) {
                installments.push(
                    {
                        "installment": i,
                        "amount": paymentValue / i,
                        "withInterest": false
                    }
                )
            }
            return installments;
        }
        
        return service;
    });
})();
