(function() {
    'use strict';

    angular
        .module('easyorder.services.payment.processor', [])
        .service('PaymentProcessor', PaymentProcessor);

    PaymentProcessor.status = {
        NONE: 0x00,
        MUST_UNDO: 0x01,
        MUST_CONFIRM: 0x02
    };
    PaymentProcessor.$inject = [
        '$rootScope',
        '$q',
        'LogService',
        'ConfigService',
        'CartService',
        'UserService',
        'EFTService',
        'CustomerService',
        'GiftCardService',
        'CovenantService',
        'VoucherService',
        'EWalletService',
        'PaymentLinkService',
        'PrivateLabelService'
    ];
    function PaymentProcessor(
        $rootScope,
        $q,
        LogService,
        ConfigService,
        CartService,
        UserService,
        EFTService,
        CustomerService,
        GiftCardService,
        CovenantService,
        VoucherService,
        EWalletService,
        PaymentLinkService,
        PrivateLabelService
    ) {
        ///// EXPOSED FUNCTIONS /////////////////////
        this.initialize = initialize;
        this.confirm = confirm;
        this.undo = undo;
        this.undoPayment = undoPayment;
        this.pushSalePayment = pushSalePayment;
        this.updateSalePayment = updateSalePayment;
        this.containsPayment = containsPayment;
        this.containsPendingEFT = containsPendingEFT;
        this.pushMobileRecharge = pushMobileRecharge;
        /////////////////////////////////////////////
        ///// INTERNAL ONLY ////////////////////////
        const self = this;

        let dirty = false;
        let processing = false;

        const DB_NAME = 'payments';
        const COLLECTION_NAME = 'payments';
        const COLLECTION_INDICES = ['nsu', 'status', 'date', 'priority', 'type'];

        const SUPPORTED_PAYMENT_TYPES = [
            'EFT',
            'MOBILE_RECHARGE',
            'PRIVATE_LABEL',
            'GIFTCARD',
            'COVENANT',
            'VOUCHER_STOREX',
            'VOUCHER_GIVEX',
            'VOUCHER_OMNI',
            'VOUCHER_SAFE',
            'E_WALLET',
            'PAYMENT_LINK',
        ];

        const UNIQUE_PAYMENT_TYPES = [
            'EFT',
            'MOBILE_RECHARGE',
            'PAYMENT_LINK'
        ];

        const PRIORITY = {
            EFT: 0,
            MOBILE_RECHARGE: 0,
            PRIVATE_LABEL: 1,
            GIFTCARD: 1,
            COVENANT: 1,
            VOUCHER_STOREX: 1,
            VOUCHER_GIVEX: 1,
            VOUCHER_OMNI: 1,
            VOUCHER_SAFE: 1,
            E_WALLET: 1,
            PAYMENT_LINK: 1,
        };

        const PROCESSING_INTERVAL = 10000; // in ms

        // every handle must return a promise
        const paymentHandler = {
            EFT: {
                confirm: (confirmData) => {
                    return EFTService.confirmTransaction(confirmData);
                },
                undo: (undoData) => {
                    return EFTService.undoTransaction(undoData);
                }
            },
            MOBILE_RECHARGE: {
                confirm: (confirmData) => {
                    return EFTService.confirmTransaction(confirmData);
                },
                undo: (undoData) => {
                    return EFTService.undoTransaction(undoData);
                }
            },
            PRIVATE_LABEL: {
                confirm: (confirmData, provider) => {
                    if (provider && provider.id) {
                        return PrivateLabelService.confirmPayment(confirmData);
                    }
                    return CustomerService.confirmPrivateLabelPayment(confirmData);
                },
                undo: (undoData, provider) => {
                    if (provider && provider.id) {
                        return PrivateLabelService.undoPayment(undoData);
                    }
                    return CustomerService.undoPrivateLabelAuthorization(undoData);
                }
            },
            GIFTCARD: {
                confirm: (confirmData) => {
                    return GiftCardService.confirmGiftCardPayment(confirmData);
                },
                undo: (undoData) => {
                    return GiftCardService.undoGiftCardPayment(undoData)
                    .catch(err => {
                        // tratamento de eventos
                        if (undoData.contextUid
                            && ConfigService.config.business.giftcard_service_provider == 'PROMO') {
                            const ignoredStatuses = [400, 402, 403, 409];
                            if (ignoredStatuses.includes(err.status)) {
                                LogService.info(
                                    'Desfazimento do'
                                    + ` CP '${undoData.cardNumber}'`
                                    + ` AMOUNT '${undoData.amount}'`
                                    + ` CONTEXT_UID '${undoData.contextUid}'`
                                   + ` não será executado devido ao Status HTTP '${err.status}'`
                                );
                                return $q.resolve();
                            }
                        }
                        return $q.reject(err);
                    });;
                }
            },
            COVENANT: {
                confirm: (confirmData) => {
                    return CovenantService.covenantSaleAuthConfirmation(confirmData);
                },
                undo: (undoData) => {
                    return CovenantService.undoCovenantSaleAuth(undoData);
                }
            },
            VOUCHER_STOREX: {
                confirm: (confirmData) => {
                    return VoucherService.confirm(
                        confirmData.nsu,
                        confirmData.voucherNumber,
                        confirmData.storeNumber,
                        confirmData.voucherType,
                        confirmData.contextUid,
                        null,
                        confirmData.terminalNumber,
                        confirmData.companyNumber,
                        new Date().toJSON()
                    );
                },
                undo: (undoData) => {
                    return VoucherService.cancel(
                        undoData.nsu,
                        undoData.voucherNumber,
                        undoData.storeNumber,
                        undoData.voucherType,
                        undoData.companyNumber,
                        undoData.terminalNumber,
                        null,
                        null,
                        undoData.contextUid,
                        new Date().toJSON()
                    ).catch(err => {
                        // tratamento de eventos
                        if (undoData.contextUid) {
                            const ignoredStatuses = [400, 402, 403, 409];
                            if (ignoredStatuses.includes(err.status)) {
                                LogService.info(
                                    'Desfazimento do'
                                    + ` VOUCHER '${undoData.voucherNumber}'`
                                    + ` VALUE '${undoData.value}'`
                                    + ` CONTEXT_UID '${undoData.contextUid}'`
                                   + ` não será enfileirado novamente devido ao Status HTTP '${err.status}'`
                                );
                                return $q.resolve();
                            }
                        }
                        return $q.reject(err);
                    });
                },
            },
            VOUCHER_GIVEX: {
                confirm: (confirmData) => {
                    return VoucherService.confirm(
                        confirmData.nsu,
                        confirmData.voucherNumber,
                        confirmData.storeNumber,
                        confirmData.voucherType,
                        confirmData.contextUid,
                        null,
                        null,
                        null
                    );
                },
                undo: (undoData) => {
                    return VoucherService.cancel(
                        undoData.nsu,
                        undoData.voucherNumber,
                        undoData.storeNumber,
                        undoData.voucherType,
                        undoData.companyNumber,
                        undoData.terminalNumber,
                        undoData.authorizationCode,
                        undoData.value,
                        undoData.contextUid,
                    ).catch(err => {
                        // tratamento de eventos
                        if (undoData.contextUid) {
                            const ignoredStatuses = [400, 402, 403, 409];
                            if (ignoredStatuses.includes(err.status)) {
                                LogService.info(
                                    'Desfazimento do'
                                    + ` VT '${undoData.voucherNumber}'`
                                    + ` AMOUNT '${undoData.value}'`
                                    + ` CONTEXT_UID '${undoData.contextUid}'`
                                   + ` não será executado devido ao Status HTTP '${err.status}'`
                                );
                                return $q.resolve();
                            }
                        }
                        return $q.reject(err);
                    });
                }
            },
            VOUCHER_SAFE: {
                confirm: (confirmData) => {
                    return VoucherService.confirm(
                        confirmData.nsu,
                        confirmData.voucherNumber,
                        confirmData.storeNumber,
                        confirmData.voucherType,
                        confirmData.contextUid,
                        confirmData.value,
                        confirmData.terminalNumber,
                        confirmData.companyNumber
                    );
                },
                undo: (undoData) => {
                    return VoucherService.cancel(
                        undoData.nsu,
                        undoData.voucherNumber,
                        undoData.storeNumber,
                        undoData.voucherType,
                        undoData.companyNumber,
                        undoData.terminalNumber,
                        undoData.authorizationCode,
                        undoData.value,
                        undoData.contextUid
                    ).catch(err => {
                        // tratamento de eventos
                        if (undoData.contextUid) {
                            const ignoredStatuses = [400, 402, 403, 409];
                            if (ignoredStatuses.includes(err.status)) {
                                LogService.info(
                                    'Desfazimento do'
                                    + ` VT '${undoData.voucherNumber}'`
                                    + ` AMOUNT '${undoData.value}'`
                                    + ` CONTEXT_UID '${undoData.contextUid}'`
                                   + ` não será executado devido ao Status HTTP '${err.status}'`
                                );
                                return $q.resolve();
                            }
                        }
                        return $q.reject(err);
                    });
                }
            },
            VOUCHER_OMNI: {
                confirm: (confirmData) => {
                    return VoucherService.confirm(
                        confirmData.nsu,
                        confirmData.voucherNumber,
                        confirmData.storeNumber,
                        confirmData.voucherType,
                        confirmData.contextUid,
                    );
                },
                undo: (undoData) => {
                    // Se o cliente foi previamente identificado, atualiza o saldo do vale
                    if (CartService.cart.customer &&
                        CartService.cart.customer.document === undoData.voucherNumber &&
                        ConfigService.config.business.promo_voucher_balance_enabled) {
                        CartService.cart.customer.vouchersBalance += undoData.value;
                    }
                    return VoucherService.create(
                        undoData.nsu,
                        undoData.voucherNumber,
                        undoData.companyNumber,
                        undoData.storeNumber,
                        undoData.terminalNumber,
                        undoData.value,
                        undoData.voucherType,
                    ).catch(err => {
                        // tratamento de eventos
                        if (undoData.contextUid) {
                            const ignoredStatuses = [400, 402, 403, 409];
                            if (ignoredStatuses.includes(err.status)) {
                                LogService.info(
                                    'Desfazimento do'
                                    + ` VT '${undoData.voucherNumber}'`
                                    + ` AMOUNT '${undoData.value}'`
                                    + ` CONTEXT_UID '${undoData.contextUid}'`
                                   + ` não será executado devido ao Status HTTP '${err.status}'`
                                );
                                return $q.resolve();
                            }
                        }
                        return $q.reject(err);
                    });
                }
            },
            E_WALLET: {
                confirm: (confirmData) => EWalletService.confirmPayment(confirmData),
                undo: (undoData) => EWalletService.undoPayment(undoData),
            },
            PAYMENT_LINK: {
                confirm: (confirmData) => PaymentLinkService.verifyTransaction(confirmData.orderNumber),
                undo: (undoData) => PaymentLinkService.cancelCheckoutPayments({ orderNumber: undoData.orderNumber }),
            }
        };

        const _db = () => {
            return new loki(DB_NAME, {
                adapter: _adapter(),
                autoload: true,
                autoloadCallback: _onDbLoaded
            });
        };

        const _adapter = () => {
            if (window.cordova && ionic.Platform.isAndroid()) {
                return new lokiCordovaFsAdapter({
                    prefix: 'loki',
                    dataDirectory: `${$rootScope.fullRootFilePathDb}`
                });
            } else {
                return new loki.LokiLocalStorageAdapter();
            }
        };

        const _onDbLoaded = () => {
            self.coll = _collection();
            _markIncompletePayments();
            dirty = _isCollectionDirty();

            if (!_containsEFT() && window.cordova) {
                EFTService.undoTransaction();
            }

            _startTick();
        };

        const _collection = () => {
            // if collection doesn't exists we create it
            let coll = self.db.getCollection(COLLECTION_NAME);
            if (_isNull(self.coll)) {
                coll = _createCollection();
            }
            return coll;
        };

        const _createCollection = () => {
            return self.db.addCollection(COLLECTION_NAME, {
                indices: COLLECTION_INDICES
            });
        };

        const _markIncompletePayments = () => {
            self.coll.findAndUpdate({ status: PaymentProcessor.status.NONE }, (payment) => {
                payment.status = PaymentProcessor.status.MUST_UNDO;
                return payment;
            });
            _saveDb();
        };

        const _isCollectionDirty = () => {
            if (_isCollectionEmpty()) {
                return false;
            }
            const payments = self.coll.find({ status: { $ne: PaymentProcessor.status.NONE } });
            if (_isNull(payments)) {
                return false;
            }
            return payments.length !== 0;
        };

        const _isCollectionEmpty = () => {
            return self.coll.data.length === 0;
        };

        const _startTick = () => {
            if (!_isNull(self.intervalHandle)) {
                return;
            }
            self.intervalHandle = setInterval(_tickCb, PROCESSING_INTERVAL);
        };

        const _stopTick = () => {
            if (_isNull(self.intervalHandle)) {
                return;
            }
            clearInterval(self.intervalHandle);
            self.intervalHandle = null;
        };

        const _tickCb = () => {
            if (dirty === false || processing === true || !UserService.isAuthenticated()) {
                return;
            }
            processing = true;
            _processPayments().finally(() => {
                dirty = _isCollectionDirty();
                processing = false;
            });
        };

        const _isNull = (arg) => {
            return arg === undefined || arg === null;
        };

        const _updateStatus = (nsu, status) => {
            const parsedNsu = parseInt(nsu);

            if (!_containsPayment(parsedNsu)) {
                LogService.info(`ignoring nsu = ${parsedNsu}, not found on payment processor`);
                return;
            }

            LogService.info(`updating status of nsu = ${parsedNsu} payments to status = ${status}`);

            self.coll.findAndUpdate({ nsu: parsedNsu }, (payment) => {
                // hack, a data deve ser verificada antes do update
                if (!_dateFilter(payment)) {
                    return payment;
                }

                // também deve ser filtrado antes
                if (payment.status != PaymentProcessor.status.NONE) {
                    return payment;
                }

                payment.status = status;
                if (payment.priority === 0) {
                    _processPaymentPromise(payment);
                }
                dirty = true;
                return payment;
            });
            _saveDb();
        };

        const _processPayments = () => {
            const payments = self.coll.chain()
                .find({ status: { $ne: PaymentProcessor.status.NONE } })
                .sort((pA, pB) => {
                    const pDiff = pA.priority - pB.priority;
                    if (pDiff === 0) {
                        return pB.date - pA.date; // descending date
                    }
                    return pDiff; // ascending priority
                }).data();

            return payments.reduce((acc, payment) => {
                return acc.then(() => _processPaymentPromise(payment))
            }, $q.resolve());
        };

        const _processPaymentPromise = (payment) => {
            return _processPayment(payment)
                .then(() => self.coll.remove(payment))
                .then(_saveDb)
                .catch(LogService.error)
                .then(() => {
                    payment.processing = false;
                });
        };

        const _processPayment = (payment) => {
            LogService.info(`processing payment, type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu}`);
            if (payment.processing === true) {
                return $q.reject(`payment type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu} is already processing`);
            }
            payment.processing = true;
            const handler = paymentHandler[payment.type];
            if (_isNull(handler)) {
                payment.processing = false;
                throw new Error(`no handler for payment type = ${payment.type} registered`);
            }
            switch (payment.status) {
                case PaymentProcessor.status.MUST_UNDO:
                    LogService.info(`undoing payment, type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu}`);
                    return _paymentProcessingWrap(handler.undo(payment.data.undo, payment.provider), payment);
                case PaymentProcessor.status.MUST_CONFIRM:
                    LogService.info(`confirming payment, type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu}`);
                    return _paymentProcessingWrap(handler.confirm(payment.data.confirm, payment.provider), payment);
                default:
                    throw new Error(
                        `cannot proccess payment unknown status, nsu = ${payment.nsu}, type = ${payment.type}, value = ${payment.value}, status = ${payment.status}`
                    );
            }
        };

        const _paymentProcessingWrap = (promise, payment) => {
            return promise.then(res => {
                LogService.info(`successfully processed payment, type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu}`);
                return res;
            }).catch(err => {
                LogService.info(`failed to process payment, type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu}`);
                return $q.reject(err);
            });
        };

        const _containsPayment = (nsu) => {
            let payment = self.coll.chain()
                .find({ nsu: nsu })
                .where(_dateFilter)
                .data();
            payment = _first(payment);
            return !_isNull(payment);
        };

        /*
         * Refatorar esta função em favor do _containsPayment
         * hoje o _containsPayment não é dinâmico o suficiente
         */
        const _containsPendingEFT = () => {
            const payments = self.coll.find({
                type: 'EFT',
                status: { $ne: PaymentProcessor.status.NONE }
            });
            return !_isNull(payments) && payments.length > 0;
        };

        const _containsEFT = () => {
            const payments = self.coll.find({
                type: 'EFT'
            });
            return !_isNull(payments) && payments.length > 0;
        };

        const _saveDb = () => {
            self.db.save();
        };

        const _init = () => {
            self.db = _db();
            CartService.initCartObservers.add(() => _eventify(CartService.cart.salePayments, _autoPushSalePayment));
        };

        const _eventify = (arr, callback) => {
            arr.push = function(e) {
                Array.prototype.push.call(arr, e);
                callback(e, arr);
            };
        };

        const _autoPushSalePayment = (salePayment) => {
            if (!_mustAutoPush(salePayment)) {
                return;
            }
            _pushSalePayment(salePayment);
        };

        const _pushSalePayment = (salePayment) => {
            _push(_parsePayment(salePayment));
        };

        const _mustAutoPush = (salePayment) => {
            // pagamentos pré confirmados não são processados
            // atualmente pelo CXR
            if (salePayment.paymentOrigin != null
                    && salePayment.paymentOrigin.type == 'ORDER'
                    && salePayment.paymentState == 'CONFIRMED') {
                return false;
            }

            switch (salePayment.type) {
                case SalePaymentTypes.SALE_TEF_PAYMENT:
                    // Controle criado para transações com Link de Pagamento (Pagar.me)
                    return salePayment.paymentProvider == null ||
                        (salePayment.paymentProvider.id != 'PAGAR_ME' &&
                        salePayment.paymentProvider.id != 'LINX_PAYHUB_LINK');
                case SalePaymentTypes.SALE_PRIVATE_LABEL_PAYMENT:
                case SalePaymentTypes.SALE_COVENANT_PAYMENT:
                case SalePaymentTypes.SALE_VOUCHER_STOREX_PAYMENT:
                case SalePaymentTypes.SALE_VOUCHER_SAFE_PAYMENT:
                case SalePaymentTypes.SALE_VOUCHER_PAYMENT:
                    return true;
                case SalePaymentTypes.SALE_GIFTCARD_PAYMENT:
                default:
                    return false;
            }
        };

        const _parsePayment = (salePayment) => {
            const type = _parsePaymentType(salePayment.type);
            return {
                type: type,
                nsu: parseInt(CartService.cart.nsu),
                status: PaymentProcessor.status.NONE,
                date: new Date(),
                value: new Number(salePayment.paymentValue),
                data: _parsePaymentData(salePayment),
                unique: _isPaymentTypeUnique(type),
                priority: _getPriority(type),
                provider: salePayment.paymentProvider
            };
        };

        const _parsePaymentType = (type) => {
            switch (type) {
                case SalePaymentTypes.SALE_TEF_PAYMENT:
                case SalePaymentTypes.CARD_STATEMENT_PAYMENT:
                    return 'EFT';
                case SalePaymentTypes.SALE_PRIVATE_LABEL_PAYMENT:
                    return 'PRIVATE_LABEL';
                case SalePaymentTypes.SALE_GIFTCARD_PAYMENT:
                    return 'GIFTCARD';
                case SalePaymentTypes.SALE_COVENANT_PAYMENT:
                    return 'COVENANT';
                case SalePaymentTypes.SALE_VOUCHER_STOREX_PAYMENT:
                case SalePaymentTypes.SALE_VOUCHER_PAYMENT:
                    return 'VOUCHER_STOREX';
                case SalePaymentTypes.SALE_VOUCHER_GIVEX_PAYMENT:
                    return 'VOUCHER_GIVEX';
                case SalePaymentTypes.SALE_VOUCHER_OMNI_PAYMENT:
                    return 'VOUCHER_OMNI';
                case SalePaymentTypes.SALE_VOUCHER_SAFE_PAYMENT:
                    return 'VOUCHER_SAFE';
                case SalePaymentTypes.SALE_E_WALLET_PAYMENT:
                    return 'E_WALLET';
                case SalePaymentTypes.SALE_PAYMENT_LINK:
                    return 'PAYMENT_LINK'
                case 'MOBILE_RECHARGE':
                    return type;
                default:
                    throw new Error(`payment type = ${type} not supported`);
            }
        };

        const _parsePaymentData = (salePayment) => {
            switch (salePayment.type) {
                case 'MOBILE_RECHARGE':
                case SalePaymentTypes.SALE_TEF_PAYMENT:
                case SalePaymentTypes.CARD_STATEMENT_PAYMENT:
                    return {
                        confirm: salePayment.paymentAuthorizationNSU,
                        undo: salePayment.paymentAuthorizationNSU
                    };
                case SalePaymentTypes.SALE_PRIVATE_LABEL_PAYMENT:
                    let paymentData = (salePayment.paymentProvider) ?  salePayment : _parsePrivateLabelPaymentData(salePayment);
                    return {
                        confirm: paymentData,
                        undo: paymentData
                    };
                case SalePaymentTypes.SALE_GIFTCARD_PAYMENT:
                    return {
                        confirm: salePayment,
                        undo: salePayment
                    };
                case SalePaymentTypes.SALE_COVENANT_PAYMENT:
                    return {
                        confirm: _parseCovenantPaymentConfirmationData(salePayment),
                        undo: salePayment.paymentAuthorizationNSU
                    };
                case SalePaymentTypes.SALE_VOUCHER_STOREX_PAYMENT:
                case SalePaymentTypes.SALE_VOUCHER_PAYMENT:
                    return {
                        confirm: {
                            nsu: ConfigService.nsuControll.incrementNSU(),
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: 'VOUCHER_STOREX',
                            companyNumber: ConfigService.config.companyNumber,
                            terminalNumber: ConfigService.config.terminalNumber,
                            contextUid: salePayment.contextUid
                        },
                        undo: {
                            nsu: ConfigService.nsuControll.incrementNSU(),
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: 'VOUCHER_STOREX',
                            companyNumber: ConfigService.config.companyNumber,
                            terminalNumber: ConfigService.config.terminalNumber,
                            contextUid: salePayment.contextUid
                        },
                    };
                case SalePaymentTypes.SALE_VOUCHER_GIVEX_PAYMENT:
                    return {
                        confirm: {
                            nsu: ConfigService.nsuControll.incrementNSU(),
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: 'VOUCHER_GIVEX',
                            contextUid: salePayment.contextUid,
                        },
                        undo: {
                            nsu: ConfigService.nsuControll.incrementNSU(),
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: 'VOUCHER_GIVEX',
                            authorizationCode: salePayment.paymentAuthorizationCode,
                            companyNumber: ConfigService.config.companyNumber,
                            terminalNumber: ConfigService.config.terminalNumber,
                            value: salePayment.paymentValue,
                            contextUid: salePayment.contextUid,
                        },
                    };
                case SalePaymentTypes.SALE_VOUCHER_OMNI_PAYMENT:
                    return {
                        confirm: {
                            nsu: ConfigService.nsuControll.incrementNSU(),
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: ConfigService.config.business.giftcard_service_provider,
                            contextUid: salePayment.contextUid,
                        },
                        undo: {
                            nsu: ConfigService.nsuControll.incrementNSU(),
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: ConfigService.config.business.giftcard_service_provider,
                            authorizationCode: salePayment.paymentAuthorizationCode,
                            companyNumber: ConfigService.config.companyNumber,
                            terminalNumber: ConfigService.config.terminalNumber,
                            value: salePayment.paymentValue,
                            contextUid: salePayment.contextUid,
                        },
                    };
                case SalePaymentTypes.SALE_VOUCHER_SAFE_PAYMENT:
                    return {
                        confirm: {
                            nsu: salePayment.safeNsu,
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: 'VOUCHER_SAFE',
                            contextUid: salePayment.contextUid,
                            value: salePayment.paymentValue,
                            terminalNumber: ConfigService.config.terminalNumber,
                            companyNumber: ConfigService.config.companyNumber
                        },
                        undo: {
                            nsu: salePayment.safeNsu,
                            voucherNumber: salePayment.voucherNumber,
                            storeNumber: salePayment.storeNumber,
                            voucherType: 'VOUCHER_SAFE',
                            authorizationCode: salePayment.paymentAuthorizationCode,
                            companyNumber: ConfigService.config.companyNumber,
                            terminalNumber: ConfigService.config.terminalNumber,
                            value: salePayment.paymentValue,
                            contextUid: salePayment.contextUid
                        },
                    };
                case SalePaymentTypes.SALE_E_WALLET_PAYMENT:
                    return {
                        confirm: {
                            paymentId: salePayment.paymentId,
                            paymentValue: salePayment.paymentValue,
                            remoteTerminalId: salePayment.remoteTerminalId,
                            discountValue: salePayment.discountValue,
                            walletType: salePayment.walletType,
                        },
                        undo: {
                            paymentId: salePayment.paymentId,
                            remoteTerminalId: salePayment.remoteTerminalId,
                            transactionId: salePayment.transactionId,
                            walletType: salePayment.walletType,
                        }
                    };
                    case SalePaymentTypes.SALE_PAYMENT_LINK:
                        return {
                            confirm: {
                                orderNumber: salePayment.paymentAuthorizationNSU
                            },
                            undo: {
                                orderNumber: salePayment.paymentAuthorizationNSU
                            }
                        }
                default:
                    throw new Error(`payment type = ${type} not supported`);
            }
        };

        const _parsePrivateLabelPaymentData = (_) => {
            const clonedCart = JSON.parse(JSON.stringify(CartService.cart));
            clonedCart.storeNumber = ConfigService.config.storeNumber;

            return clonedCart;
        };

        const _parseCovenantPaymentConfirmationData = (salePayment) => {
            const clonedCart = JSON.parse(JSON.stringify(CartService.cart));
            return {
                nsu: salePayment.paymentAuthorizationNSU,
                saleNSU: clonedCart.nsu,
                terminalNumber: clonedCart.terminalNumber,
                payments: clonedCart.salePayments.map(p => {
                    let ret = {};
                    ret.type = p.type;
                    return ret;
                })
            };
        };

        const _isPaymentTypeUnique = (type) => {
            if (UNIQUE_PAYMENT_TYPES.includes(type)) {
                if (type == 'EFT' && ConfigService.config.tefserver.provider == 'Linx') {
                    return false;
                }
                return true;
            }
            return false;
        };

        const _getPriority = (type) => {
            return PRIORITY[type];
        };

        const _push = (payment) => {
            if (_isNull(payment)) {
                throw new Error('payment must be provided');
            }
            _validatePaymentFields(payment);

            if (payment.unique === true) {
                let uniquePayment = self.coll.chain()
                    .find({ $and: [
                        { nsu: payment.nsu },
                        { unique: true },
                        { type: payment.type },
                        { status: PaymentProcessor.status.NONE }
                    ] }).where(_dateFilter).data();

                uniquePayment = _first(uniquePayment);

                if (!_isNull(uniquePayment)) {
                    LogService.info(`nsu = ${payment.nsu} already contains a payment of type = ${payment.type}, updating value...`);
                    uniquePayment.value += payment.value;
                    self.coll.update(uniquePayment);
                    _saveDb();
                    return;
                }
            }
            LogService.info(`added payment of value = ${payment.value} type = ${payment.type} nsu = ${payment.nsu}`);
            self.coll.insert(payment);
            _saveDb();
        };

        const _validatePaymentFields = (payment) => {
            if (_isNull(payment.nsu)) {
                throw new Error('payment nsu must be provided');
            }
            if (_isNull(payment.type)) {
                throw new Error('payment type must be provided');
            }
            if (!_isPaymentTypeSupported(payment.type)) {
                throw new Error(`payment of type = ${payment.type} not supported`);
            }
            if (_isNull(payment.data)) {
                throw new Error('payment data must be provided');
            }
            if (_isNull(payment.data.confirm)) {
                throw new Error('payment confirmation data must be provided');
            }
            if (_isNull(payment.data.undo)) {
                throw new Error('payment undo data must be provided');
            }
        };

        const _isPaymentTypeSupported = (type) => {
            return SUPPORTED_PAYMENT_TYPES.includes(type);
        };

        const _updateStatusPromise = (nsu, status) => {
            return $q((resolve, reject) => {
                try {
                    _updateStatus(nsu, status);
                    resolve();
                } catch(err) {
                    reject(err);
                }
            });
        };

        const _dateFilter = (pmnt) => {
            const today = new Date();
            today.setHours(0, 0, 0, 0);
            const pmntDate = new Date(pmnt.date);
            pmntDate.setHours(0, 0, 0, 0);

            return today.getTime() == pmntDate.getTime();
        }

        const _first = (arr) => {
            return _isNull(arr) || arr.length == 0 ? null : arr[0];
        }
        /////////////////////////////////////////////

        ///// EXPOSED FUNCTIONS IMPLEMENTATION //////
        function initialize() {
            LogService.info('initializing payment processor');

            if (window.cordova && ionic.Platform.isAndroid()) {
                window.resolveLocalFileSystemURL($rootScope.rootFilePath, rootEntry => {
                    rootEntry.getDirectory('Linx', { create: true }, linxEntry => {
                        linxEntry.getDirectory('CaixaRapido', { create: true }, cxrEntry => {
                            cxrEntry.getDirectory('db', { create: true }, _ => {
                                _init();
                            });
                        });
                    });
                });
            } else {
                _init();
            }
        }

        function containsPayment(nsu) {
            return _containsPayment(parseInt(nsu));
        }

        function containsPendingEFT() {
            return _containsPendingEFT();
        }

        async function confirm(nsu) {
            const payments = self.coll.chain()
                .find({ nsu: nsu, status: PaymentProcessor.status.NONE })
                .sort((pA, pB) => {
                    const pDiff = pB.priority - pA.priority;
                    if (pDiff === 0) {
                        return pA.date - pB.date; // data crescente
                    }
                    return pDiff; // prioridade decrescente
                }).data();

            try {
                for (const payment of payments) {
                    // TEF executado por último (pela prioridade)
                    await _confirmSync(payment);
                }
            } catch (error) {
                // status não modificado pois pode ser solicitado a retentativa.
                LogService.error(`error confirming payment!`);
                LogService.error(error);
                throw new Error(`error confirming payment!`);
            }
        }

        async function _confirmSync(payment) {
            const handler = paymentHandler[payment.type];
            if (_isNull(handler)) {
                throw new Error(`no handler for payment type = ${payment.type} registered`);
            } else {
                LogService.info(`confirming payment, type = ${payment.type}, value = ${payment.value}, nsu = ${payment.nsu}`);
                await _paymentProcessingWrap(handler.confirm(payment.data.confirm, payment.provider), payment);
                self.coll.remove(payment);
                _saveDb();
            }
        }

        function undo(nsu) {
            return _updateStatusPromise(nsu, PaymentProcessor.status.MUST_UNDO);
        }

        // só utilizar caso a forma de pagamento não seja "auto-push", ver _mustAutoPush
        function pushSalePayment(salePayment) {
            return _pushSalePayment(salePayment);
        }

        function updateSalePayment(filter, salePayment) {
            const payment = _first(self.coll.where(filter));

            if (_isNull(payment)) {
                throw new Error('payment with the given filter doesn\'t exists');
            }

            payment.data = _parsePaymentData(salePayment);

            self.coll.update(payment);
            _saveDb();
        }

        // só utilizar caso seja realmente necessário
        function undoPayment(filter) {
            const payment = _first(self.coll.where(filter));

            if (_isNull(payment)) {
                throw new Error('payment with the given filter doesn\'t exists');
            }

            payment.status = PaymentProcessor.status.MUST_UNDO;
            self.coll.update(payment);
            _saveDb();
            dirty = true;
        }

        // esta função só existe pois uma recarga não é um pagamento, e sim um serviço
        // mesmo assim precisamos confirmar ou desfazer. Caso não existam pagamentos EFT,
        // sem esta chamada a recarga nunca será confirmada ou desfeita
        // propriedades necessárias [paymentValue, paymentAuthorizationNSU]
        function pushMobileRecharge(data) {
            if (_containsEFT()) {
                return;
            }
            _pushSalePayment(Object.assign({
                type: 'MOBILE_RECHARGE',
            }, data));
        }
    }
})();
