/**
 * Este factory serve para lançar erro em caso de falha de conexão do lado do cliente (caixa rápido),
 * pois o javascript de algumas versões do android não dispara o erro com readyState = 4 e status = 0
 */
(function () {
    'use strict';
    angular.module('easyorder.xhrfactory', [])
        .factory('$xhrFactory', function(LogService) {
            const HTTP = "http";
            return function createXhr(method, url) {
                return new window.XMLHttpRequest();
            };
        });
})();

(function () {
    'use strict';

    const lowercase = angular.lowercase;
    const noop = angular.noop;
    const forEach = angular.forEach;
    const isDefined = angular.isDefined;
    const isUndefined = angular.isUndefined;
    const isFunction = angular.isFunction;
    const isPromiseLike = (obj) => obj && isFunction(obj.then);

    angular.module('easyorder.xhrfactory', [])
        .factory('$httpBackend', function(LogService, $browser, $window, $document, $xhrFactory) {
            function doRequest(method, url, post, callback, headers, timeout, withCredentials, responseType, eventHandlers, uploadEventHandlers, retryCount) {
                $browser.$$incOutstandingRequestCount();
                url = url || $browser.url();

                const callbacks = $window.angular.callbacks;

                if (lowercase(method) == 'jsonp') {
                    var callbackId = '_' + (callbacks.counter++).toString(36);
                    callbacks[callbackId] = function(data) {
                        callbacks[callbackId].data = data;
                        callbacks[callbackId].called = true;
                    };

                    var jsonpDone = jsonpReq(url.replace('JSON_CALLBACK', 'angular.callbacks.' + callbackId),
                        callbackId, function(status, text) {
                            completeRequest(callback, status, callbacks[callbackId].data, "", text);
                            callbacks[callbackId] = noop;
                        });
                } else {
                    var xhr = $xhrFactory(method, url);

                    const dtimeout = timeout > 1 ? Date.now() + timeout : null;
                    retryCount = retryCount || 0;

                    xhr.open(method, url, true);

                    forEach(headers, function(value, key) {
                        if (isDefined(value)) {
                            xhr.setRequestHeader(key, value);
                        }
                    });

                    xhr.onload = function requestLoaded(e) {
                        var statusText = xhr.statusText || '';

                        // responseText is the old-school way of retrieving response (supported by IE9)
                        // response/responseType properties were introduced in XHR Level2 spec (supported by IE10)
                        var response = ('response' in xhr) ? xhr.response : xhr.responseText;

                        // normalize IE9 bug (http://bugs.jquery.com/ticket/1450)
                        var status = xhr.status === 1223 ? 204 : xhr.status;

                        // fix status code when it is 0 (0 status is undocumented).
                        // Occurs when accessing file resources or on Android 4.1 stock browser
                        // while retrieving files from application cache.
                        if (status === 0) {
                            status = response ? 200 : urlResolve(url).protocol == 'file' ? 404 : 0;
                        }

                        completeRequest(callback,
                            status,
                            response,
                            xhr.getAllResponseHeaders(),
                            statusText);
                    };

                    const requestError = function() {
                        // The response is always empty
                        // See https://xhr.spec.whatwg.org/#request-error-steps and https://fetch.spec.whatwg.org/#concept-network-error
                        completeRequest(callback, -1, null, null, '');
                    };

                    xhr.onerror = (e) => {
                        // Apenas trataremos retries quando a requisição for http e o servidor
                        // não retornar status.
                        // O ideal seria além destas condições só enviarmos o retry se o request
                        // não for totalmente enviado para o servidor, porém o upload handler
                        // do xhr é bugado, e não notifica quantos bytes da requisição já foram enviados...
                        if (xhr.status === 0 && url.startsWith("http")) {
                            if (retryCount <= 3) {
                                LogService.info(`Failed HTTP request, retrying... retryCount: ${retryCount}`);

                                // Ajustando o timeout de acordo com o delta..
                                timeout = dtimeout ? dtimeout - Date.now() : timeout;
                                if (timeout <= 0) {
                                    timeoutRequest();
                                    return;
                                }

                                // limpando request
                                cleanup();
                                // Marcando que request está completo, apesar do erro
                                // não será retornado para o angular, uma vez que a função
                                // complete não chamará o callback
                                complete();

                                // Estamos criando um novo request, logo devemos cancelar o antigo timeout, caso existente
                                doRequest(method, url, post, callback, headers, timeout, withCredentials, responseType, eventHandlers, uploadEventHandlers, ++retryCount);
                                return;
                            }
                        }
                        requestError();
                    };
                    xhr.onabort = requestError;
                    xhr.ontimeout = requestError;
                    
                    forEach(eventHandlers, function(value, key) {
                        xhr.addEventListener(key, value);
                    });

                    forEach(uploadEventHandlers, function(value, key) {
                        xhr.upload.addEventListener(key, value);
                    });

                    if (withCredentials) {
                        xhr.withCredentials = true;
                    }

                    if (responseType) {
                        try {
                            xhr.responseType = responseType;
                        } catch (e) {
                            // WebKit added support for the json responseType value on 09/03/2013
                            // https://bugs.webkit.org/show_bug.cgi?id=73648. Versions of Safari prior to 7 are
                            // known to throw when setting the value "json" as the response type. Other older
                            // browsers implementing the responseType
                            //
                            // The json response type can be ignored if not supported, because JSON payloads are
                            // parsed on the client-side regardless.
                            if (responseType !== 'json') {
                                throw e;
                            }
                        }
                    }

                    xhr.send(isUndefined(post) ? null : post);
                }

                if (timeout > 0) {
                    var timeoutId = $browser.defer(timeoutRequest, timeout);
                } else if (isPromiseLike(timeout)) {
                    timeout.then(timeoutRequest);
                }


                function timeoutRequest() {
                    jsonpDone && jsonpDone();
                    xhr && xhr.abort();
                }

                function completeRequest(callback, status, response, headersString, statusText) {
                    cleanup();
                    callback(status, response, headersString, statusText);
                    complete();
                }

                function cleanup() {
                    // cancel timeout and subsequent timeout promise resolution
                    if (isDefined(timeoutId)) {
                        $browser.defer.cancel(timeoutId);
                    }
                    jsonpDone = xhr = null;
                }

                function complete() {
                    $browser.$$completeOutstandingRequest(noop);
                }
            };

            function jsonpReq(url, callbackId, done) {
                // we can't use jQuery/jqLite here because jQuery does crazy stuff with script elements, e.g.:
                // - fetches local scripts via XHR and evals them
                // - adds and immediately removes script elements from the document
                const rawDocument = $document[0];
                var script = rawDocument.createElement('script'), callback = null;
                script.type = "text/javascript";
                script.src = url;
                script.async = true;

                callback = function(event) {
                    removeEventListener(script, "load", callback);
                    removeEventListener(script, "error", callback);
                    rawDocument.body.removeChild(script);
                    script = null;
                    var status = -1;
                    var text = "unknown";

                    if (event) {
                        if (event.type === "load" && !callbacks[callbackId].called) {
                            event = { type: "error" };
                        }
                        text = event.type;
                        status = event.type === "error" ? 404 : 200;
                    }

                    if (done) {
                        done(status, text);
                    }
                };

                addEventListener(script, "load", callback);
                addEventListener(script, "error", callback);
                rawDocument.body.appendChild(script);
                return callback;
            }

            return doRequest;
        });
})();

